/***************************************************************************
 *  Drile.cpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "Drile.hpp"
#include "env/EnvManager.hpp"
#include "audio/AudioManager.hpp"
#include "input/InputManager.hpp"

#include "env/Reconstruct3D.hpp"
#include "env/Tunnel.hpp"
#include "env/Grid.hpp"
#include "env/Zone.hpp"
#include "env/Room.hpp"

using namespace std;

Drile::Drile():m_deviceIDCounter(0) {}

Drile::~Drile() {}

Drile* Drile::getInstance() {
    static Drile instance;
    return &instance;
}

void* inputThreadFunc(void* data) {
    Drile *drile = Drile::getInstance();
    while(1) {
        drile->updateInput();
        usleep(1);
    }
    pthread_exit(0);
}

void Drile::init(int argc,char* argv[]) {

    InputManager* inputManager = InputManager::getInstance();
    AudioManager* audioManager = AudioManager::getInstance();
    EnvManager* envManager = EnvManager::getInstance();

    //create ringbuffers from input thread to env thread
    m_rbToEnv = new RingBuffer(sizeof(Event));
    m_rbToInput = new RingBuffer(sizeof(Event));

    //start all managers
    audioManager->start();
    inputManager->init();
    envManager->start();

    //start the input thread
    if(pthread_create(&m_inputThread,NULL,inputThreadFunc,this)) {
        throw logic_error("Error creating the input thread");
    }

    //if a file given as argument, parse it
    if(argc>1) {
        try {
            parse(std::string(argv[1]));
        }
        catch(std::exception e) {
            cout<<"Drile: Error when parsing file "<<argv[1]
                <<e.what()<<endl;
        }
    }

    //main window
    m_mainWindow = new MainWindow(600,100,"Drile | Setup");
    m_mainWindow->show();

    m_mainWindow->resetElementsList();
}

void Drile::parse(std::string file) {
     xmlDocPtr doc= xmlReadFile((file).c_str(),NULL,0);
    xmlNodePtr rootNode = xmlDocGetRootElement(doc);
    if (doc == NULL || rootNode==NULL ) {
            throw logic_error("Unable to load file");
    }

    if (xmlStrcmp(rootNode->name, (const xmlChar *) "Drile")) {
            xmlFreeDoc(doc);
            throw logic_error("File is not a Drile file");
    }

    AudioManager* audioManager = AudioManager::getInstance();
    audioManager->reset();
    EnvManager* envManager = EnvManager::getInstance();
    envManager->reset();

    xmlNodePtr groupNode;
    for (groupNode = rootNode->children; groupNode; 
            groupNode = groupNode->next) {
        if (groupNode->type == XML_ELEMENT_NODE){
            if(!xmlStrcmp(groupNode->name, (const xmlChar *) "Cameras")) {
                parseCameras(groupNode);
            }
            else if(!xmlStrcmp(groupNode->name, (const xmlChar *) "Rooms")) {
                parseRooms(groupNode);
            }
        }
    }
}

void Drile::parseCameras(xmlNodePtr camerasNode) {
    xmlNodePtr camNode;
    for (camNode = camerasNode->children; camNode; camNode = camNode->next) {
        if (camNode->type == XML_ELEMENT_NODE && !xmlStrcmp(camNode->name, (const xmlChar *) "Musician")) {
            Musician* newCamera = EnvManager::getInstance()->addMusician();
            newCamera->parse(camNode);
            std::vector<MappingID> newMappings;
            newCamera->getMappingsIDs(newMappings);
            addMappings(newMappings);

            xmlNodePtr inputNode;
            for (inputNode = camNode->children; inputNode; inputNode = inputNode->next) {
                if (inputNode->type == XML_ELEMENT_NODE) {
                    #ifdef WITH_FREENECT 
                    if(!xmlStrcmp(inputNode->name, (const xmlChar *) "Reconstruction")) {
                        newCamera->addReconstruction(inputNode);
                    }
                    else 
                    #endif
                    if(!xmlStrcmp(inputNode->name, 
                                    (const xmlChar *) "Piivert")) {
                        Piivert* piiv=NULL;
                        PiivertDevice* piivDev = NULL;
                        addPiivert(newCamera, piiv, piivDev);
                        piiv->parse(inputNode);
                        piivDev->parse(inputNode);
                        newMappings.clear();
                        piivDev->getMappingsIDs(newMappings);
                        addMappings(newMappings);
                    }
                    else if(!xmlStrcmp(inputNode->name, (const xmlChar *) "Hands")) {
                        Piivert* piiv=NULL;
                        HandDevice* handDev = NULL;
                        addHands(newCamera, piiv, handDev);
                        piiv->parse(inputNode);
                        handDev->parse(inputNode);
                        newMappings.clear();
                        handDev->getMappingsIDs(newMappings);
                        addMappings(newMappings);
                    }
                }
            }
        }
        else if (camNode->type == XML_ELEMENT_NODE && 
                    !xmlStrcmp(camNode->name, (const xmlChar *) "Spectator")) {
            Spectator* newCamera = EnvManager::getInstance()->addSpectator();
            newCamera->parse(camNode);
            std::vector<MappingID> newMappings;
            newCamera->getMappingsIDs(newMappings);
            addMappings(newMappings);
        }
    }
}

void Drile::parseRooms(xmlNodePtr roomsNode) {
    xmlNodePtr roomNode;
    
    char* value=NULL;
    value = (char*)xmlGetProp(roomsNode,(xmlChar*)"worm_size");
    if(value!=NULL) {
        EnvManager::getInstance()->setWormSize(atof(value));
    }

    unsigned int roomCounter=0;
    for(roomNode = roomsNode->children; roomNode; roomNode = roomNode->next) {
        if (roomNode->type == XML_ELEMENT_NODE 
                && !xmlStrcmp(roomNode->name, (const xmlChar *) "Room")) {
            parseRoom(roomNode, roomCounter);
            roomCounter++;
        }
    }
}

void Drile::parseRoom(xmlNodePtr roomNode, const unsigned int& roomID) {
    //add new room and parse it
    EnvManager* envManager = EnvManager::getInstance();
    Room* newRoom = envManager->getCreateRoom(roomID);
    newRoom->parse(roomNode);

    //get zones
    unsigned int zoneCounter=0;
    xmlNodePtr roomSubNode;
    for (roomSubNode = roomNode->children; roomSubNode; 
            roomSubNode = roomSubNode->next) {
        if (roomSubNode->type == XML_ELEMENT_NODE){
            if(!xmlStrcmp(roomSubNode->name, (const xmlChar *) "Zone")) {
                Zone* newZone = newRoom->getCreateZone(zoneCounter);
                newZone->parse(roomSubNode);
                parseZone(newRoom, zoneCounter, roomSubNode);
                zoneCounter++;
            }
        }
    }
}

void Drile::parseZone(Room* newRoom, const unsigned int& zone, 
                        xmlNodePtr zoneNode) {

    //parse zone content: worms and tunnels
    xmlNodePtr roomSubNode;
    for(roomSubNode = zoneNode->children; roomSubNode; 
                                            roomSubNode = roomSubNode->next) {

        if (roomSubNode->type == XML_ELEMENT_NODE){
            if(!xmlStrcmp(roomSubNode->name, (const xmlChar *) "Tunnel")) {
                EnvManager* envManager = EnvManager::getInstance();
                Tunnel *newTunnel = envManager->addTunnel(newRoom->getID(),
                                                                        zone);
                newTunnel->parse(roomSubNode);
            }
            else if(!xmlStrcmp(roomSubNode->name, (const xmlChar *) "Grid")) {
                EnvManager* envManager = EnvManager::getInstance();
                Grid *newGrid = envManager->addGrid(newRoom->getID(), zone);
                newGrid->parse(roomSubNode);
            }
            else {
                Worm* newWorm;
                HllElement* newNode;
                NODE_TYPE type=LEAF_SAMPLE;
                if(!xmlStrcmp(roomSubNode->name, 
                                        (const xmlChar *) "SampleWorm")) {
                    type=LEAF_SAMPLE;
                }
                else if(!xmlStrcmp(roomSubNode->name, 
                                        (const xmlChar *) "DrileWorm")) {
                    type=LEAF_LAYERED_GRAIN;
                }
                else if(!xmlStrcmp(roomSubNode->name, 
                                        (const xmlChar *) "SasWorm")) {
                    type=LEAF_SAS;
                }
                addWorm(type, newWorm, newNode, newRoom->getID(), zone, 
                        -1, roomSubNode);
            }
        }
    }
}

void Drile::updateInput() {

    //get input events
    InputManager* inputManager = InputManager::getInstance();
    inputManager->update(m_inputEvents);
    //process them to produce audio and env events
    processInputEvents();
    
    //update all devices
    map<int, Device*>::iterator itDev = m_devices.begin();
    for(; itDev!=m_devices.end(); ++itDev) {
        itDev->second->update(m_eventsToAudio, m_eventsToEnvThread);
    }


    //get events from the ring from env
    try {
        Event newEvent;
        while(m_rbToInput->eventsAvailable()) {
            m_rbToInput->getEvent(newEvent);
            //piivert selection and speed events
            if(newEvent.type==EVENT_PIIVERT) { 
                switch(newEvent.parameter) {
                    case PIIVERT_SELECT_WORM: {
                        static_cast<PiivertDevice*>
                            (m_devices[newEvent.piivert])->
                                    setSelectedWorms(HAND(newEvent.hand),
                                                             newEvent.values);
                    }break;
                    case PIIVERT_SPEED: {
                        static_cast<PiivertDevice*>
                            (m_devices[newEvent.piivert])->
                                setSpeed(HAND(newEvent.hand), newEvent.values);
                    }break;
                    default:break;
                }
            }
            else if(newEvent.type==EVENT_MAPPING) { //mapping events
                MappingID &mappingId = newEvent.mapping;
                if(mappingId.inputType==INPUT_OSC) {
                    m_oscMappings[mappingId.strId]=mappingId.mapping;
                }
                else {
                    m_mappings[mappingId.inputType][mappingId.id]=
                                                        mappingId.mapping;
                }
            }
            else {
                m_eventsToAudio.push_back(newEvent);
            }
        }
    }
    catch (exception e) {
        cout<<"Error getting events from env thread "<<e.what()<<endl;
    }

    //process the audio
    AudioManager* audioManager = AudioManager::getInstance();
    audioManager->update(m_eventsToAudio, m_eventsToEnvThread);

    try {
        //add the events from input and audio to the ring to env
        vector<Event>::iterator itEv;
        for(itEv=m_eventsToEnvThread.begin(); 
                itEv!=m_eventsToEnvThread.end(); ++itEv) {    
            m_rbToEnv->addEvent(*itEv);
        }
    }
    catch (exception e) {
        cout<<"Error sending input events to env thread "<<e.what()<<endl;
    }
    m_eventsToAudio.clear();
    m_eventsToEnvThread.clear();
}


void Drile::update() {
    EnvManager* envManager = EnvManager::getInstance();
    try {
        Event newEvent;
        while(m_rbToEnv->eventsAvailable()) {
            m_rbToEnv->getEvent(newEvent);
            m_eventsToEnv.push_back(newEvent);
        }
    }
    catch (exception e) {
        cout<<"Error getting events from input thread "<<e.what()<<endl;
    }
    try {
        envManager->update(m_eventsToEnv, m_eventsToAudioThread);
    }
    catch (exception e) {
        cout<<"Error updating env"<<e.what()<<endl;
    }

    try {
        vector<Event>::iterator itEv;
        for(itEv=m_eventsToAudioThread.begin(); 
                itEv!=m_eventsToAudioThread.end(); ++itEv) {    
            m_rbToInput->addEvent(*itEv);
        }
    }
    catch (exception e) {
        cout<<"Error sending events to input thread "<<e.what()<<endl;
    }
    m_eventsToEnv.clear();
    m_eventsToAudioThread.clear();
}

void Drile::addWorm(NODE_TYPE type, Worm* &newWorm, HllElement* &newNode, 
                    int roomID, const unsigned int& zone, int parentWormID,
                    xmlNodePtr xmlNode){

    //add node the env manager
    EnvManager* envManager = EnvManager::getInstance();
    newWorm = envManager->addWorm(type, roomID, zone, parentWormID, xmlNode);

    //build a new node
    AudioManager* audioManager = AudioManager::getInstance();
    newNode = audioManager->createNode(type, parentWormID, xmlNode);

    //add node to the audio manager using the input thread
    Event newEvent;    
    newEvent.type=EVENT_TREE;
    newEvent.parameter=TREE_NODE_ADD;
    newEvent.nodePtr = newNode;
    m_rbToInput->addEvent(newEvent);
}

void Drile::addPiivert(Musician* cam, Piivert* &piiv, PiivertDevice* &piivDev) {
    //create a piivert
    piivDev=new PiivertDevice(m_deviceIDCounter);
    m_devices[m_deviceIDCounter]=piivDev;
    EnvManager* envManager = EnvManager::getInstance();
    piiv = envManager->addPiivert(m_deviceIDCounter, cam);
    m_deviceIDCounter++;
}

void Drile::addHands(Musician* cam, Piivert* &piiv, HandDevice* &handDev) {
    //create a hand device 
    handDev=new HandDevice(m_deviceIDCounter);
    m_devices[m_deviceIDCounter]=handDev;
    EnvManager* envManager = EnvManager::getInstance();
    piiv = envManager->addPiivert(m_deviceIDCounter, cam);
    m_deviceIDCounter++;
}


Room* Drile::addRoom() {
    EnvManager* envManager = EnvManager::getInstance();
    Room* newRoom = envManager->addRoom();
    return newRoom;
}


void Drile::addMapping(MappingID& mappingId) {
    Event newEvent;    
    newEvent.type=EVENT_MAPPING;
    newEvent.mapping = mappingId;
    m_rbToInput->addEvent(newEvent);
}

void Drile::addMappings(std::vector<MappingID>& mappings) {
    vector<MappingID>::iterator itMap;
    for(itMap=mappings.begin();itMap!=mappings.end();++itMap) {
        addMapping((*itMap));
    }
}

void Drile::processInputEvents() {
    vector<InputEvent>::iterator itEvent;
    for(itEvent=m_inputEvents.begin(); itEvent!=m_inputEvents.end(); 
            ++itEvent) {
        INPUT_TYPE &type = (*itEvent).type;
        switch((*itEvent).type) {
            case INPUT_OSC:{
                if(m_oscMappings.find((*itEvent).strId)!=m_oscMappings.end()) {
                    m_oscMappings[(*itEvent).strId]->process((*itEvent), 
                                                    m_eventsToAudio, 
                                                    m_eventsToEnvThread);
                }
                else if((*itEvent).strParameter.find("/audio/")!=string::npos) {
                    //parse audio event
                }
                else if((*itEvent).strParameter.find("/env/")!=string::npos) {
                    //parse env event
                }

            }break;
            default: {
                if(m_mappings[type].find((*itEvent).id)!=
                                    m_mappings[type].end()) {
                    m_mappings[type][(*itEvent).id]->process(
                                *itEvent, m_eventsToAudio, 
                                        m_eventsToEnvThread);
                }
            }break;
        }
    }
}

int main(int argc, char** argv) {
    Fl::gl_visual(FL_RGB);

    OSG::ChangeList::setReadWriteDefault(true);
    OSG::osgInit(0,0);

    Drile* drile = Drile::getInstance();
    drile->init(argc,argv);

    Fl::run();
    return EXIT_SUCCESS;
}

