/***************************************************************************
 *  DrileDisplay.cpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "DrileDisplay.hpp"

#include <Fl/gl.h>

OSG_USING_NAMESPACE
using namespace std;

DrileDisplay::DrileDisplay(int w, int h, const char* t): Fl_Gl_Window(w,h,t), 
                                                        m_fullscreen(false),
                                                        m_displayTooltip(false) {
    if(can_do(FL_STEREO)) {
        mode(FL_RGB|FL_DOUBLE|FL_STEREO|FL_ALPHA|FL_DEPTH|FL_MULTISAMPLE);
    }
    else {
        mode(FL_RGB|FL_DOUBLE|FL_ALPHA|FL_DEPTH|FL_MULTISAMPLE);
    }

    box(FL_DOWN_BOX);
    cursor(FL_CURSOR_NONE);
    resizable(this);

    m_ract=RenderAction::create();
    m_osgWin=PassiveWindow::create();
    m_osgWin->resize(w,h);

    Fl::add_idle(idle, this);
}

void DrileDisplay::idle(void* pnt) {
    DrileDisplay *win = (DrileDisplay*)pnt;
    win->redraw();
    usleep(1000);
}

DrileDisplay::~DrileDisplay(){}

void DrileDisplay::init(std::string name, int port) {
    m_name=name;
    m_port=port;
    if(m_port==-1) {
        m_server = new ClusterServer(m_osgWin, m_name, "Multicast", "");
    } else {
        m_server = new ClusterServer(m_osgWin, m_name, "StreamSock", "", m_port);
    }
    
    m_name = "Drile | Display : "+m_name;
    label(m_name.c_str());
    m_server->start();
}

void DrileDisplay::draw() {
    //init
    if(!valid()) {
        m_osgWin->init();
        valid(1);
    }

    //clear
    glClearColor(0, 0, 0, 0);
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    try {
        m_server->render(m_ract);
        OSG::Thread::getCurrentChangeList()->clear();
    }
    catch(OSG_STDEXCEPTION_NAMESPACE::exception &e) {
        SLOG << e.what() << endLog;
        OSG::Thread::getCurrentChangeList()->clear();
        m_server->stop();
        m_server->start();
    }

    if(m_displayTooltip && !m_fullscreen) {
        gl_color(FL_WHITE);
        gl_font(FL_HELVETICA, 24);
        //gl_draw("Press spacebar to enter fullscreen mode", 
        //        0, 0, 100, 30, FL_ALIGN_RIGHT);
        //gl_draw("Press spacebar to enter fullscreen mode ", 0, 0); 
    }
}

void DrileDisplay::resize(int x, int y, int w, int h) {
    m_osgWin->resize(w,h);
}

int DrileDisplay::handle(int event) {
    switch(event) {
        case FL_ENTER: {
            m_displayTooltip=true;
            return 1;
        }break;
        case FL_LEAVE: {
            m_displayTooltip=false;
            return 1;
        }break;
        case FL_KEYDOWN: {
            if(Fl::event_key()==32) {
                toggleFullscreen();
                return 1;
            }
        }break;
    }
    return Fl_Widget::handle(event);
}

void DrileDisplay::toggleFullscreen() {
    if(!m_fullscreen) {
        cursor(FL_CURSOR_NONE);
        int width=300;
        int height=300;
        if ( XineramaIsActive( fl_display ) ) {
                int nscreens;
                XineramaScreenInfo * info =
                        XineramaQueryScreens( fl_display, &nscreens );
                width=info[0].width;
                height=info[0].height;
        }
        else {
                width=WidthOfScreen(DefaultScreenOfDisplay(fl_display));
                height=HeightOfScreen(DefaultScreenOfDisplay(fl_display));
        }
        m_osgWin->resize(width,height);
    }
    else {
        m_osgWin->resize(300,300);
        cursor(FL_CURSOR_DEFAULT);
    }

    fl_open_display();
    XEvent xev;
    this->make_current();
    xev.xclient.type = ClientMessage;
    xev.xclient.serial = 0;
    xev.xclient.send_event = True;
    xev.xclient.message_type = XInternAtom(fl_display, "_NET_WM_STATE", False);
    xev.xclient.window = fl_xid(this);
    xev.xclient.format = 32;
    xev.xclient.data.l[0] = (m_fullscreen ? 0 : 1);
    xev.xclient.data.l[1] = 
        XInternAtom(fl_display, "_NET_WM_STATE_FULLSCREEN", False);
    xev.xclient.data.l[2] = 0;
    xev.xclient.data.l[3] = 0;
    xev.xclient.data.l[4] = 0;
    XSendEvent(fl_display, DefaultRootWindow(fl_display), False, 
                SubstructureRedirectMask | SubstructureNotifyMask, &xev);
    XSync(fl_display, False);

    m_fullscreen = ! m_fullscreen;
}


int main(int argc, char** argv) {
    Fl::gl_visual(FL_RGB);

    OSG::ChangeList::setReadWriteDefault(true);
    OSG::osgInit(0,0);
    DrileDisplay* display = new DrileDisplay(300,300,"Drile | Display");
    std::string name="cam";
    int port=-1;
    //first argument is the adress
    if(argc>1) {
        name=std::string(argv[1]);
    }
    //the second argument is the port
    if(argc>2) {
        port=atoi(argv[2]);
    }
    display->init(name,port);
    display->show();

    Fl::run();
    return EXIT_SUCCESS;
}
