/***************************************************************************
 *  AudioManager.cpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net 
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "AudioManager.hpp"
#include "../Drile.hpp"

using namespace std;

int audioCallback(jack_nframes_t nbFrames, void *arg) {
    AudioManager* audioManager = (AudioManager*) arg;
    //WE GET THE ports
    jack_default_audio_sample_t* out[2];
    out[0] = (jack_default_audio_sample_t *)
                jack_port_get_buffer(audioManager->m_inputPorts[0], nbFrames);    
    out[1] = (jack_default_audio_sample_t *)
                jack_port_get_buffer(audioManager->m_inputPorts[1], nbFrames);    

    audioManager->process(out);
    return 0;
}


int sRate (jack_nframes_t nframes, void *arg) {
    return 0;
}

void shutdown(void *arg) {
    AudioManager* proc = (AudioManager*) arg;
    proc->m_jackClient=NULL;
}

AudioManager::AudioManager(): m_started(false), m_sampleRate(44100), 
                                m_bufferSize(256), m_nodeIdCounter(0) {
    Event event;
    //ring buffers between the real-time audio process 
    //and the update/visu thread
    m_ringBufferFromProcess = new RingBuffer(sizeof(event));
    m_ringBufferFromUpdate = new RingBuffer(sizeof(event));
}

AudioManager::~AudioManager() {
    stop();
}

AudioManager* AudioManager::getInstance() {
    static AudioManager instance;
    return &instance;
}

void AudioManager::stop() {
    jack_deactivate(m_jackClient); 
    jack_port_unregister(m_jackClient, m_inputPorts[0]);
    jack_port_unregister(m_jackClient, m_inputPorts[1]);
    jack_client_close(m_jackClient);

    m_started=false;
}

void AudioManager::reset() {
    m_reset=true;
}

void AudioManager::start() {
    //jack client creation
    m_jackClient=NULL;
    m_jackClient = jack_client_open("Drile", JackNullOption, NULL);        
    if(!m_jackClient) {
        throw logic_error("Error creating Jack Client");
    }
    m_sampleRate = jack_get_sample_rate(m_jackClient);    
    m_bufferSize = jack_get_buffer_size(m_jackClient);    
    m_compressor = new Compressor(m_sampleRate);
    m_reverb= new Reverb(m_sampleRate);

    //enveloppes
    EnvelopeManager::getInstance()->init(64, 5000);

    //create the main output ports
    m_inputPorts[0] = jack_port_register (m_jackClient, 
                "Main-L", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
    m_inputPorts[1] = jack_port_register (m_jackClient, 
                "Main-R", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);

    //audio process callback
    jack_set_process_callback (m_jackClient, audioCallback, this);

    //shutdown callback
    jack_on_shutdown(m_jackClient,shutdown,this);    

    //Initialize the vectors: 
    //tempNode buffer has 3 channels: normal and synth/base sample channel
    //tempScene buffer has 5 channels: 2 normal, 2 compressed and 1 compressing
    std::vector<double> vec2(m_bufferSize,0.0);
    m_tempNodeBuffer.resize(3,vec2);
    m_tempSceneBuffer.resize(5,vec2);
    gettimeofday(&m_previousTime,NULL);

    //activate the client
    if(jack_activate(m_jackClient))  {
        throw logic_error("Cannot activate jack client, jackd not running ?");
    }

    //connect the main audio ports
    jack_connect(m_jackClient,"Drile:Main-L", "system:playback_1");
    jack_connect(m_jackClient,"Drile:Main-R", "system:playback_2");


    m_barkBands = new int[XTRACT_BARK_BANDS];

    try{
        xtract_init_fft(AudioManager::FFT_SIZE, XTRACT_SPECTRUM);
        xtract_init_bark(AudioManager::FFT_SIZE, m_sampleRate, m_barkBands);
    }
    catch(std::exception& e) {
        throw std::logic_error("Error initializing visualization"
                                        +std::string(e.what()));
    }

    m_started=true;
    m_reset=false;
}

jack_client_t* AudioManager::getClient() {
    return m_jackClient;
}

void AudioManager::process(jack_default_audio_sample_t *out[]) {
    Event newEvent;
    while(m_ringBufferFromUpdate->eventsAvailable()) {
        m_ringBufferFromUpdate->getEvent(newEvent);
        if(newEvent.type==EVENT_TREE) {
            switch(newEvent.parameter) {
                case TREE_NODE_ADD:{
                    m_nodesTrees.push_back(newEvent.nodePtr);
                    m_nodesMap[newEvent.nodePtr->getID()]=newEvent.nodePtr;
                }break;
                case TREE_BUILD_RECORD:{
                    m_nodesMap[newEvent.node]->loopStartRecord(newEvent.time);
                }break;
                case TREE_BUILD_CANCEL:{
                    m_nodesMap[newEvent.node]->loopErase();
                }break;
                case TREE_BUILD_CREATE: {
                    //cout<<"AudioManager: BUILD"<<endl;
                    m_nodesMap[newEvent.node]->loopStopRecord(newEvent.time);

                    //create a new parent
                    //if node does not already have a parent, 
                    //or if node has a sync parent
                    HllElement* childNode = m_nodesMap[newEvent.node];
                    if((!childNode->getParent() || 
                            (childNode->getParent() && 
                                childNode->getParent()->getNbChildren()>1))
                            && childNode->hasRecordedLoop()) {

                        //get previous parent
                        HllNode* oldParentNode = childNode->getParent();
                        int oldParentNodeId = 
                            (oldParentNode!=NULL)?oldParentNode->getID():-1;
                        //add a parent node
                        HllNodeLoop* parentNode = 
                            new HllNodeLoop(m_nodeIdCounter);
                        //transfer loops
                        parentNode->setExcitationLoop(
                                childNode->getRecordedExcitationLoopEvents());
                        for(unsigned int p=0;p<NODE_PARAMETERS_NB;++p) {
                            parentNode->setModulationLoop(p, 
                                childNode->getRecordedModulationLoopEvents(p));
                        }
                        //add the new worm to the map
                        m_nodesMap[m_nodeIdCounter]=parentNode;
                        //create a new tree root
                        m_nodesTrees.push_back(parentNode);
                        //move the child node
                        moveNode(childNode, childNode->getParent(), parentNode);
                        //move the parent node if needed 
                        if(oldParentNode!=NULL) {    
                            moveNode(parentNode, NULL, oldParentNode);
                        }
                        //start looping 
                        parentNode->loopPlay();//childNode->getLoopRecordStopTime());


                        //sending data to the update 
                        //create new node
                        Event event;
                        event.type=EVENT_TREE;
                        event.parameter=TREE_NODE_ADD;
                        event.values[0]=NODE_LOOP;
                        event.values[1]=parentNode->getID();
                        event.values[2]=childNode->getID();
                        m_ringBufferFromProcess->addEvent(event);
                        //move child node to parent
                        event.type=EVENT_TREE;
                        event.parameter=TREE_NODE_MOVE;
                        event.values[0]=childNode->getID();
                        event.values[1]=oldParentNodeId;
                        event.values[2]=parentNode->getID();
                        m_ringBufferFromProcess->addEvent(event);
                        //move the parent node if needed 
                        if(oldParentNode!=NULL) {    
                            event.type=EVENT_TREE;
                            event.parameter=TREE_NODE_MOVE;
                            event.values[0]=parentNode->getID();
                            event.values[1]=-1;
                            event.values[2]=oldParentNode->getID();
                            m_ringBufferFromProcess->addEvent(event);
                        }

                        //cout<<"AudioManager: add node in audio process"<<endl;

                        m_nodeIdCounter++;
                    }
                }break;
                case TREE_COLLIDE: {
                    //cout<<"AudioManager: MERGE"<<endl;
                    
                    //merge and synchronize second node to first node 
                    //or swap if one of the nodes has a parent

                    if(m_nodesMap.find(newEvent.values[0])!=m_nodesMap.end() && 
                        m_nodesMap.find(newEvent.values[1])!=m_nodesMap.end()) {

                        //cout<<"merge operation, valid nodes !!"<<endl;

                        HllElement* firstHllElement = 
                            m_nodesMap[newEvent.values[0]];
                        HllElement* secondHllElement = 
                            m_nodesMap[newEvent.values[1]];


                        //make sure they are not part of the same tree already
                        if(firstHllElement->
                                isElementInSubtree(secondHllElement) 
                            || secondHllElement->
                                isElementInSubtree(firstHllElement)) {

                            break;
                        }

                        //no parents , and either sync or loop nodes
                        if(!firstHllElement->getParent() 
                                && !secondHllElement->getParent()
                                && firstHllElement->getNbChildren()>0 
                                && secondHllElement->getNbChildren()>0 ) {

                            //cout<<"merge operation  possible !!"<<endl;

                            Event event;
                            
                            //define the merging sync node    
                            HllNodeSync* mergingNode=NULL;    
                            HllNodeSync* removedNode=NULL;    
                            //and the loop nodes which will be added to it
                            std::vector<HllElement*> addedLoopNodes;
                            HllNodeLoop* syncNode=NULL;

                            //if only one child -> loop node
                            //if at least two children -> sync node
                            if(firstHllElement->getNbChildren()==1 
                                    && secondHllElement->getNbChildren()==1) {

                                //two loop nodes, create a sync node
                                mergingNode = new HllNodeSync(m_nodeIdCounter);
                                m_nodesMap[m_nodeIdCounter]=mergingNode;
                                m_nodesTrees.push_back(mergingNode);
                                m_nodeIdCounter++;
                            
                                //set the children
                                addedLoopNodes.push_back(firstHllElement);
                                addedLoopNodes.push_back(secondHllElement);

                                //and the sync node
                                syncNode = 
                                    static_cast<HllNodeLoop*>(firstHllElement);

                                event.type=EVENT_TREE;
                                event.parameter=TREE_NODE_ADD;
                                event.values[0]=NODE_SYNC;
                                event.values[1]=mergingNode->getID();
                                event.values[2]=syncNode->getID();
                                m_ringBufferFromProcess->addEvent(event);
                            }
                            else if(firstHllElement->getNbChildren()>1 &&
                                    secondHllElement->getNbChildren()==1) {
                                //one loop and one sync
                                mergingNode = 
                                    static_cast<HllNodeSync*>(firstHllElement);
                                addedLoopNodes.push_back(secondHllElement);
                            }
                            else if(firstHllElement->getNbChildren()==1 &&
                                    secondHllElement->getNbChildren()>1) {
                                //one loop and one sync
                                mergingNode = 
                                    static_cast<HllNodeSync*>(secondHllElement);
                                addedLoopNodes.push_back(firstHllElement);
                                syncNode=
                                    static_cast<HllNodeLoop*>(firstHllElement);
                            }
                            else if(firstHllElement->getNbChildren()>1 &&
                                    secondHllElement->getNbChildren()>1) {
                                //two syncs
                                mergingNode = 
                                    static_cast<HllNodeSync*>(firstHllElement);
                                for(unsigned int c=0; 
                                    c<secondHllElement->getNbChildren(); ++c) {

                                    addedLoopNodes.push_back(
                                        secondHllElement->getChild(c));
                                }
                                removedNode=
                                    static_cast<HllNodeSync*>(secondHllElement);
                            }


                            if(mergingNode) {
                                vector<HllElement*>::iterator itNode;
                                for(itNode=addedLoopNodes.begin(); 
                                    itNode!=addedLoopNodes.end(); ++itNode){

                                    //move the children
                                    moveNode(*itNode, 
                                    static_cast<HllNode*>(*itNode)->getParent(),
                                    static_cast<HllNode*>(mergingNode));

                                    event.type=EVENT_TREE;
                                    event.parameter=TREE_NODE_MOVE;
                                    event.values[0]=(*itNode)->getID();
                                    event.values[1]=-1;
                                    event.values[2]=mergingNode->getID();
                                    m_ringBufferFromProcess->addEvent(event);
                                }

                                if(syncNode) {
                                    mergingNode->setPulseNode(syncNode);
                                }
                            }

                            if(removedNode) {
                                m_nodesMap.erase(removedNode->getID());
                                vector<HllElement*>::iterator itNode;
                                for(itNode=m_nodesTrees.begin();
                                        itNode!=m_nodesTrees.end();++itNode) {
                                    if((*itNode)->getID()==removedNode->getID()) {
                                        m_nodesTrees.erase(itNode);
                                        break;
                                    }
                                }
                                event.type=EVENT_TREE;
                                event.parameter=TREE_NODE_REMOVE;
                                event.values[0]=removedNode->getID();
                                event.values[1]=-1;
                                event.values[2]=-1;
                                m_ringBufferFromProcess->addEvent(event);
                                delete removedNode;
                            }
                        }
                        else if(firstHllElement->getParent() 
                                    || secondHllElement->getParent()) {
                            HllElement* swapElement1 = 
                                m_nodesMap[newEvent.values[0]];
                            HllElement* swapElement2 = 
                                m_nodesMap[newEvent.values[1]];

                            bool swap=true;    
                            int parent1ID = -1;
                            int parent2ID = -1;
                            //if one of the swap is a loop node below a sync, 
                            //the other should be a loop node too
                            if(swapElement1->getParent()) {
                                if(swapElement1->getParent()->getNbChildren()>1) {
                                    if(swapElement2->getNbChildren()!=1) {
                                        swap=false;
                                    }
                                }
                                parent1ID=swapElement1->getParent()->getID();
                            }
                            if(swapElement2->getParent()) {
                                if(swapElement2->getParent()->getNbChildren()>1) {
                                    if(swapElement1->getNbChildren()!=1) {
                                        swap=false;
                                    }
                                }
                                parent2ID=swapElement2->getParent()->getID();
                            }

                            if(swap) {

                                //cout<<"SWAP NODES !!"<<endl;
                                HllNode* parent1 = swapElement1->getParent();
                                HllNode* parent2 = swapElement2->getParent();
                                moveNode(swapElement1, parent1,parent2);        
                                moveNode(swapElement2, parent2, parent1);



                                Event event;
                                event.type=EVENT_TREE;
                                event.parameter=TREE_NODE_MOVE;
                                event.values[0]=swapElement1->getID();
                                event.values[1]=parent1ID;
                                event.values[2]=parent2ID;
                                m_ringBufferFromProcess->addEvent(event);
                                event.type=EVENT_TREE;
                                event.parameter=TREE_NODE_MOVE;
                                event.values[0]=swapElement2->getID();
                                event.values[1]=parent2ID;
                                event.values[2]=parent1ID;
                                m_ringBufferFromProcess->addEvent(event);
                            }
                        }
                    }
                }break;
                case TREE_DUPLICATE: {
                    if(m_nodesMap.find(newEvent.values[0])!=m_nodesMap.end()) {
                        HllElement* duplicatedElement = 
                            m_nodesMap[newEvent.values[0]];
                        std::vector<HllElement*> createdElements;
                        //call duplicate on the given element
                        duplicatedElement->duplicate(createdElements, 
                                                        m_nodeIdCounter);

                        Event event;
                        if(createdElements.size()>0) {
                            //add the last (top) element to roots
                            m_nodesTrees.push_back(createdElements.back());

                            vector<HllElement*>::reverse_iterator itEl;
                            for(itEl=createdElements.rbegin();
                                    itEl!=createdElements.rend(); ++itEl) {
                                HllElement* newEl = (*itEl);
                                //for each created element, add to map
                                m_nodesMap[newEl->getID()]=newEl;

                                //and send create and move worm events
                                event.type=EVENT_TREE;
                                event.parameter=TREE_NODE_ADD;
                                if(newEl->getNbChildren()>1) {
                                    event.values[0]=NODE_SYNC;
                                } else if(newEl->getNbChildren()==1) {
                                    event.values[0]=NODE_LOOP;
                                } else {
                                    event.values[0]=LEAF_LAYERED_GRAIN;
                                }
                                event.values[1]=newEl->getID();
                                event.values[2]=duplicatedElement->getID();
                                m_ringBufferFromProcess->addEvent(event);

                                event.type=EVENT_TREE;
                                event.parameter=TREE_NODE_MOVE;
                                event.values[0]=newEl->getID();
                                event.values[1]=-1;
                                if(newEl->getParent()) {
                                    event.values[2]=
                                        newEl->getParent()->getID();
                                }
                                else {
                                    event.values[2]=-1;
                                }
                                m_ringBufferFromProcess->addEvent(event);


                            }
                        }
                    }
                }break;
                case TREE_EXTRACT: {
                    //cout<<"asked for extract"<<endl;
                    if(m_nodesMap.find(newEvent.values[0])!=m_nodesMap.end()) {
                        //cout<<" ... valid node: ";
                        HllElement* extractedNode = 
                            m_nodesMap[newEvent.values[0]];
                        //cout<<extractedNode->getID();
                        //if has a parent
                        if(extractedNode->getParent()) {
                            HllNode* parent=extractedNode->getParent();
                            //cout<<" has a parent, ";
                            //remove from parent    
                            moveNode(extractedNode, parent, NULL);
                            Event event;
                            event.type=EVENT_TREE;
                            event.parameter=TREE_NODE_MOVE;
                            event.values[0]=extractedNode->getID();
                            event.values[1]=parent->getID();
                            event.values[2]=-1;
                            m_ringBufferFromProcess->addEvent(event);

                            //test parents
                            while(parent) {
                                //cout<<" parent has "<<parent->getNbChildren()<<" children";
                                //    if parent has only one or zero children left
                                //    erase the parent
                                if(parent->getNbChildren()<=1) {
                                    //cout<<" so remove a parent, ";
                                    //get its parent
                                    HllNode* oldParent = parent->getParent();
                                    //remove from its parent, if any
                                    moveNode(parent, parent->getParent(), NULL);

                                    //move the child if any
                                    if(parent->getNbChildren()==1) {
                                        moveNode(parent->getChild(0),
                                                    parent, NULL);
                                        Event event;
                                        event.type=EVENT_TREE;
                                        event.parameter=TREE_NODE_MOVE;
                                        event.values[0]=parent->
                                                        getChild(0)->getID();
                                        event.values[1]=parent->getID();
                                        event.values[2]=-1;
                                        m_ringBufferFromProcess->
                                                            addEvent(event);
                                    }

                                    //remove the parent from the tree 
                                    //and from the map
                                    m_nodesMap.erase(parent->getID());
                                    for(unsigned int i=0;i<m_nodesTrees.size();
                                                                    ++i) {
                                        if(m_nodesTrees[i]->getID()==
                                                    parent->getID()) {
                                            m_nodesTrees.erase(
                                                m_nodesTrees.begin()+i);
                                            break;
                                        }
                                    }
                                    Event event;
                                    event.type=EVENT_TREE;
                                    event.parameter=TREE_NODE_REMOVE;
                                    event.values[0]=parent->getID();
                                    m_ringBufferFromProcess->addEvent(event);

                                    //delete the parent
                                    delete parent;
                                    parent=oldParent;
                                }
                                else {
                                    parent=NULL;
                                }
                            }

                        }
                    }

                    //cout<<endl;
                }break;
                default:break;
            }
        }
        else if(newEvent.type==EVENT_NODE) {
            if(m_nodesMap.find(newEvent.node)!=m_nodesMap.end()) {
                m_nodesMap[newEvent.node]->processEvent(newEvent);
            }
        }
    }

    //zero out the buffer
    for(unsigned int c=0; c<m_tempSceneBuffer.size(); ++c) {
        m_tempSceneBuffer[c].assign(m_tempSceneBuffer[c].size(), 0);
    }

    //get elapsed time since previous process
    gettimeofday(&m_currentTime,NULL);
    timersub(&m_currentTime, &m_previousTime, &m_timeDiff);
    m_timeDiffD=m_timeDiff.tv_sec*1000000+m_timeDiff.tv_usec;

    //process nodes of all trees
    m_outputEventsFromProcess.clear();
    for(unsigned int n=0; n<m_nodesTrees.size(); n++) {
        m_nodesTrees[n]->process(m_timeDiffD, m_tempNodeBuffer, 
            m_tempSceneBuffer, m_outputEventsFromProcess);
    }

    //send output Events
    vector<Event>::iterator itEv;
    for(itEv=m_outputEventsFromProcess.begin(); 
            itEv!=m_outputEventsFromProcess.end(); ++itEv) {
        m_ringBufferFromProcess->addEvent((*itEv));
    }

    //compress buffer 2-3 with buffer 4 
    m_compressor->process(m_tempSceneBuffer[2], m_tempSceneBuffer[3], 
                            m_tempSceneBuffer[4]);

    //accumulate buffers and apply reverb
    for(int i=0;i<m_bufferSize;++i) {
        m_tempSceneBuffer[0][i]+=m_tempSceneBuffer[2][i];
        m_tempSceneBuffer[1][i]+=m_tempSceneBuffer[3][i];
    }
    m_reverb->process(m_tempSceneBuffer[0], m_tempSceneBuffer[1]);

    //fill the audio buffer
    for(int i=0; i<m_bufferSize; ++i) {
        out[0][i]=m_tempSceneBuffer[0][i];
        out[1][i]=m_tempSceneBuffer[1][i];
    }

    m_previousTime=m_currentTime;
}


void AudioManager::update(std::vector<Event>& inputEvents,
                            std::vector<Event>& outputEvents) {
    
    try {
        //send new events to the process thread
        vector<Event>::iterator itEv;
        for(itEv=inputEvents.begin(); itEv!=inputEvents.end(); ++itEv) {    
            m_ringBufferFromUpdate->addEvent(*itEv);
        }
    }
    catch(exception e) {
        cout<<"Drile: error while adding events to jack process"<<endl;
    }

    try {
        //get events from the process thread
        Event newEvent;
        while(m_ringBufferFromProcess->eventsAvailable()) {
            m_ringBufferFromProcess->getEvent(newEvent);
            outputEvents.push_back(newEvent);
        }
    }
    catch(exception e) {
        cout<<"Drile: error while getting events from jack process"<<endl;
    }

/*
    //analyse nodes and get visu events
    try {
        timeval curTime;
        gettimeofday(&curTime,NULL);
        vector<HllElement*>::iterator itHllElement = m_nodesTrees.begin();
        for(;itHllElement!=m_nodesTrees.end();++itHllElement) {
            (*itHllElement)->analyse(curTime, outputEvents);
        }
    }
    catch(exception e) {
        cout<<"Error while visualizing the nodes"<<endl;
    }
*/
}

HllElement* AudioManager::createNode(NODE_TYPE type, int parent, 
                                    xmlNodePtr xmlNode) {
    HllElement* newNode; 
    switch(type) {
        case NODE_LOOP:{
            newNode = new HllNodeLoop(m_nodeIdCounter);
        }break;
        case NODE_SYNC:{
            newNode = new HllNodeSync(m_nodeIdCounter);
        }break;
        case LEAF_SAMPLE:{
            newNode = new HllLeafLayeredGrain(m_nodeIdCounter, m_sampleRate, 
                                                m_bufferSize);
        }break;
        case LEAF_LAYERED_GRAIN:{
            newNode = new HllLeafLayeredGrain(m_nodeIdCounter, m_sampleRate, 
                                                m_bufferSize);
        }break;
        case LEAF_SAS:{
            newNode = new HllLeafLayeredGrain(m_nodeIdCounter, m_sampleRate, 
                                                m_bufferSize);
        }break;
        default:{
            newNode = new HllLeafLayeredGrain(m_nodeIdCounter, m_sampleRate, 
                                                m_bufferSize);
        }break;
    }

    if(xmlNode) {
        newNode->parse(xmlNode);
    }

    m_nodeIdCounter++;

    return newNode;
}

void AudioManager::moveNode(HllElement* movedNode, 
                                HllNode* fromParent, 
                                    HllNode* toParent) {
    //if root
    if(fromParent==NULL) {
        //remove from roots vector
        for(unsigned int i=0;i<m_nodesTrees.size();++i) {
            if(m_nodesTrees[i]->getID()==movedNode->getID()) {
                m_nodesTrees.erase(m_nodesTrees.begin()+i);
                break;
            }
        }
    }
    else { //remove from parent
        fromParent->removeChild(movedNode);    
    }


    if(toParent==NULL) {
        //add to roots vector
        m_nodesTrees.push_back(movedNode);
        movedNode->setParent(NULL);
    }
    else {
        toParent->addChild(movedNode);        
        movedNode->setParent(toParent);
    }
}

void AudioManager::printTrees() {
    vector<HllElement*>::iterator itHllElement = m_nodesTrees.begin();
    for(;itHllElement!=m_nodesTrees.end();++itHllElement) {
        (*itHllElement)->print("");
    }
}

