/***************************************************************************
 *  Effect.cpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "Effect.hpp"
#include <math.h>
#include <sstream>

#include "AudioManager.hpp"

using namespace std;

Effect::Effect(EFFECT_TYPE type, int effect): m_type(type), m_effect(effect), m_active(false), m_newActive(false), m_parametersChanged(false), m_duration(0)
{
    if(type==EFFECT_AUDIO) {
        switch(effect) {
            case VOLUME_EFFECT: {
                m_parameters.resize(1, 1.0);
                m_newParameters.resize(1, 1.0);
            }break;
            case MUTE_EFFECT: {
                m_parameters.resize(1, 0.0);
                m_newParameters.resize(1, 0.0);
            }break;
            case PAN_EFFECT: {
                m_parameters.resize(2, 1.0);
                m_newParameters.resize(2, 1.0);
            }break;
            case FILTER_EFFECT: {
                m_parameters.resize(2, 0.0);
                m_newParameters.resize(2, 0.0);
            }break;
            case DISTORTION_EFFECT: {
                m_parameters.resize(1, 0.0);
                m_newParameters.resize(1, 0.0);
            }break;
            case REVERB_EFFECT: {
                m_parameters.resize(1, 0.0);
                m_newParameters.resize(1, 0.0);
                //m_gverb = gverb_new(AudioManager::SAMPLE_RATE, 300.0f, 50.0f, 7.0f, 0.5f, 15.0f, 0.5f, 0.5f, 0.5f);
            }break;
            case DIODE_EFFECT: {
                m_parameters.resize(1, 0.0);
                m_newParameters.resize(1, 0.0);
            }break;
        }
    }
    else { //CONTROL EFFECT
        m_parameters.resize(CONTROL_EFFECT_PARAMS_NB, 0.0);
        m_newParameters.resize(CONTROL_EFFECT_PARAMS_NB, 0.0);
    }
    

    BEAT_DIVISIONS_ARRAY[0] = 0.125;
    BEAT_DIVISIONS_ARRAY[1] = 0.16;
    BEAT_DIVISIONS_ARRAY[2] = 0.25;
    BEAT_DIVISIONS_ARRAY[3] = 0.33;
    BEAT_DIVISIONS_ARRAY[4] = 0.5;
    BEAT_DIVISIONS_ARRAY[5] = 1.0;
    BEAT_DIVISIONS_ARRAY[6] = 2.0;
}

Effect::~Effect()
{
    //gverb_free(m_gverb);
}

void Effect::processAudio(std::vector<std::vector<float> >& outputBuffer, int framesPerBuffer)
{
    if(m_parametersChanged) {
        std::vector<std::vector<float> >fadeOutVec = outputBuffer;
        std::vector<std::vector<float> >fadeInVec = outputBuffer;
        float fadeStep = 1.0/float(framesPerBuffer);
        //fade out with previous params
        processFadedAudio(fadeOutVec, framesPerBuffer, -1.0*fadeStep);
        //change params
        m_active=m_newActive;
        m_parameters=m_newParameters;
        //fade in with new params
        processFadedAudio(fadeInVec, framesPerBuffer, fadeStep);
        for(int i=0;i<framesPerBuffer;++i) {
            outputBuffer[i][0]=fadeOutVec[i][0] + fadeInVec[i][0];
            outputBuffer[i][1]=fadeOutVec[i][1] + fadeInVec[i][1];
        }
        m_parametersChanged=false;
    }
    else {
        processFadedAudio(outputBuffer, framesPerBuffer, 0);
    }
}


void Effect::processFadedAudio(std::vector<std::vector<float> >& outputBuffer, int framesPerBuffer, float fadeStep)
{
    if(m_active) {
        //process the effect
        switch(m_effect) {
            case MUTE_EFFECT:{
                float fade=1.0;
                if(fadeStep>0) {
                    fade=0;
                }
                for(int i=0;i<framesPerBuffer;++i, fade+=fadeStep) {
                    outputBuffer[i][0]*=(1.0-m_parameters[0])*fade;
                    outputBuffer[i][1]*=(1.0-m_parameters[0])*fade;
                }
            }break;
            case VOLUME_EFFECT:{
                float fade=1.0;
                if(fadeStep>0) {
                    fade=0;
                }
                for(int i=0;i<framesPerBuffer;++i, fade+=fadeStep) {
                    outputBuffer[i][0]*=m_parameters[0]*fade;
                    outputBuffer[i][1]*=m_parameters[0]*fade;
                }
            }break;
            case PAN_EFFECT:{
                float fade=1.0;
                if(fadeStep>0) {
                    fade=0;
                }
                float panLeft = sqrt(2.0)/2.0*(cos(m_parameters[0]*(M_PI/4.0)) - sin(m_parameters[0]*(M_PI/4.0)));
                float panRight=sqrt(2.0)/2.0*(cos(m_parameters[0]*(M_PI/4.0)) + sin(m_parameters[0]*(M_PI/4.0)));
                for(int i=0;i<framesPerBuffer;++i, fade+=fadeStep) {
                    outputBuffer[i][0]*=panLeft*fade;
                    outputBuffer[i][1]*=panRight*fade;
                }
            }break;
            case DIODE_EFFECT: { //modified from Steve Harris SWH Diode Processor plugin
                float fade=1.0;
                if(fadeStep>0) {
                    fade=0;
                }
                long pos;
                if (m_parameters[0] > 0.0f && m_parameters[0] < 1.0f) {
                    for (pos = 0; pos < framesPerBuffer; pos++, fade+=fadeStep) {
                        outputBuffer[pos][0] = (((1.0f-m_parameters[0]) * outputBuffer[pos][0]) + (m_parameters[0] * (outputBuffer[pos][0] > 0.0f ? outputBuffer[pos][0] : 0.0f)))*fade;
                        outputBuffer[pos][1] = (((1.0f-m_parameters[1]) * outputBuffer[pos][1]) + (m_parameters[0] * (outputBuffer[pos][1] > 0.0f ? outputBuffer[pos][1] : 0.0f)))*fade;
                    }
                } else if (m_parameters[0] >= 1.0f && m_parameters[0] < 2.0f) {
                    float fac = m_parameters[0] - 1.0f;
                    for (pos = 0; pos < framesPerBuffer; pos++, fade+=fadeStep) {
                        outputBuffer[pos][0] = (((1.0f-fac) * (outputBuffer[pos][0] > 0 ? outputBuffer[pos][0] : 0.0)) + (fac * fabs(outputBuffer[pos][0])))*fade;
                        outputBuffer[pos][1] =(((1.0f-fac) * (outputBuffer[pos][1] > 0 ? outputBuffer[pos][1] : 0.0)) + (fac * fabs(outputBuffer[pos][1])))*fade;
                    }
                } else if (m_parameters[0] >= 2) {
                    float fac = m_parameters[0] < 3 ? m_parameters[0] - 2 : 1.0;
                    for (pos = 0; pos < framesPerBuffer; pos++, fade+=fadeStep) {
                        outputBuffer[pos][0] = (1.0-fac) * fabs(outputBuffer[pos][0]) * fade;
                        outputBuffer[pos][1] = (1.0-fac) * fabs(outputBuffer[pos][1]) * fade;
                    }
                }
            }break;
            case REVERB_EFFECT: { 
                /*
                float fade=1.0;
                if(fadeStep>0) {
                    fade=0;
                }
                float l1,l2,r1,r2;
                for (unsigned int i=0; i<framesPerBuffer; i++, fade+=fadeStep) {
                    gverb_do(m_gverb, outputBuffer[i][0], &l1, &r1);
                    gverb_do(m_gverb, outputBuffer[i][1], &l2, &r2);
                    outputBuffer[i][0] = (l1+l2)/2.0 * fade;
                    outputBuffer[i][1] = (r1+r2)/2.0 * fade;
                }
                */
            }break;
        }
    }
    else {
        float fade=1.0;
        if(fadeStep>0) {
            fade=0;
        }
        for(int i=0;i<framesPerBuffer;++i, fade+=fadeStep) {
            outputBuffer[i][0]*=fade;
            outputBuffer[i][1]*=fade;
        }
    }
}


/*
void Effect::processControl(std::vector<PatternEvent> &events, timeval curTime) 
{
    //INITIALIZE
    if(!m_initialized) {
        if(events.size()>0) {
            switch(m_effect) {
                case REPEAT_EFFECT: {

                    m_repeatStartTime=curTime;

                    timeval repeatCurrentTime;
                    timersub(&curTime,&m_repeatStartTime,&repeatCurrentTime);
                    long repeatCurrentTimeMs = repeatCurrentTime.tv_sec*1000 + (long)((float)repeatCurrentTime.tv_usec/1000.0);
                    
                    m_eventsBuffer.clear();
                    m_variationPitch=events[0].pitch;
                    m_variationComponent=events[0].componentID;
                    m_initialized=true;

                }break;
                case VELOCITY_EFFECT: {
                    m_eventsBuffer.clear();
                    m_variationPitch=events[0].pitch;
                    m_variationComponent=events[0].componentID;
                    m_initialized=true;
                }break;
                case PITCH_EFFECT: {
                    m_eventsBuffer.clear();
                    m_variationPitch=events[0].pitch;
                    m_variationComponent=events[0].componentID;
                    m_initialized=true;
                }break;
                case DELAY_EFFECT: {
                    m_eventsBuffer.clear();
                    m_repeatStartTime=curTime;
                    m_variationPitch=events[0].pitch;
                    m_variationComponent=events[0].componentID;
                    m_initialized=true;
                }break;
            }
        }
    }

    //PROCESS
    if(m_initialized) {
        switch(m_effect) {
            case REPEAT_EFFECT: {
                    timeval repeatCurrentTime;
                    timersub(&curTime,&m_repeatStartTime,&repeatCurrentTime);
                    long repeatCurrentTimeMs = repeatCurrentTime.tv_sec*1000 + (long)((float)repeatCurrentTime.tv_usec/1000.0);

                    if(m_parameters[CONTROL_EFFECT_SPREAD]==0) {
                        //Record only the events with same component and pitch
                        for(unsigned int e=0;e<events.size();++e) {
                            if(events[e].componentID == m_variationComponent && events[e].pitch == m_variationPitch) {
                                PatternEvent ev = events[e];
                                ev.time=repeatCurrentTimeMs;
                                for(int i=0;i<8;++i) {
                                    ev.time += m_repeatSizeMs;
                                    m_eventsBuffer.push_back(ev);
                                }
                            }
                        }
                    }
                    else {
                        //Record all the new events
                        for(unsigned int e=0;e<events.size();++e) {
                            PatternEvent ev = events[e];
                            for(int i=0;i<8;++i) {
                                ev.time += m_repeatSizeMs;
                                m_eventsBuffer.push_back(ev);
                            }
                        }
                    }

                    //Eventually put the recorded events in the buffer
                    for(unsigned int e=0;e<m_eventsBuffer.size();) {
                        if(m_eventsBuffer[e].time<repeatCurrentTimeMs) {
                            events.push_back(m_eventsBuffer[e]);    
                            m_eventsBuffer.erase(m_eventsBuffer.begin()+e);
                        }
                        else {
                            ++e;
                        }
                    }

            }break;
            case DELAY_EFFECT: {
                timeval repeatCurrentTime;
                timersub(&curTime,&m_repeatStartTime,&repeatCurrentTime);
                long repeatCurrentTimeMs = repeatCurrentTime.tv_sec*1000 + (long)((float)repeatCurrentTime.tv_usec/1000.0);

                if(m_parameters[CONTROL_EFFECT_SPREAD]==0) {
                    //add and clear only the events with same component and pitch
                    for(unsigned int e=0;e<events.size();) {
                        if(events[e].componentID == m_variationComponent && events[e].pitch == m_variationPitch) {
                            PatternEvent ev = events[e];
                            ev.time = repeatCurrentTimeMs + m_repeatSizeMs;
                            m_eventsBuffer.push_back(ev);
                            events.erase(events.begin()+e);
                        }
                        else {
                            ++e;
                        }
                    }
                }
                else {
                    //Record the all the new events
                    for(unsigned int e=0;e<events.size();++e) {
                        PatternEvent ev = events[e];
                        ev.time = repeatCurrentTimeMs + m_repeatSizeMs;
                        m_eventsBuffer.push_back(ev);
                    }

                    //clear all the events
                    events.clear();

                }

                //Eventually put the recorded events in the buffer
                for(unsigned int e=0;e<m_eventsBuffer.size();) {
                    if(m_eventsBuffer[e].time<repeatCurrentTimeMs) {
                        events.push_back(m_eventsBuffer[e]);    
                        m_eventsBuffer.erase(m_eventsBuffer.begin()+e);
                    }
                    else {
                        ++e;
                    }
                }
            }break;
            case VELOCITY_EFFECT: {
                if(m_parameters[CONTROL_EFFECT_SPREAD]==0) { // apply only to same component / note
                    for(unsigned int e=0; e<events.size(); ++e) {
                        if(events[e].componentID == m_variationComponent && events[e].pitch == m_variationPitch) {
                            events[e].velocity=m_parameters[CONTROL_EFFECT_VALUE];
                        }
                    }
                }
                else { //apply on all events
                    for(unsigned int e=0; e<events.size(); ++e) {
                        events[e].velocity=m_parameters[CONTROL_EFFECT_VALUE];
                    }
                }
            }break;
            case PITCH_EFFECT: {
                for(unsigned int e=0; e<events.size(); ++e) {
                    //events[e].pitch+=((m_parameters[CONTROL_EFFECT_VALUE]*10.0)*(m_parameters[SHOOT_MODE]*2.0-1.0));
                    events[e].pitch+=((m_parameters[CONTROL_EFFECT_VALUE]*20.0)-10.0);
                }
            }break;
        }
    }

}
*/

void Effect::setParameters(std::vector<float> params)
{
    switch(m_effect) {
        case PAN_EFFECT: {
            m_newParameters=params;
        }break;
        case DIODE_EFFECT: {
            m_newParameters[0]=params[0]*2.0;
        }break;
        default: {
            m_newParameters=params;
        }break;
    }

    m_parametersChanged=true;
}


void Effect::setActive(bool active)
{
    if(m_type==EFFECT_AUDIO && m_active!=active) {
        m_newActive=active;
        m_parametersChanged=true;
    }
    else {
        m_active=active;
    }
}

void Effect::setParameters(std::vector<float> params, long beatDurationMs, int beatsPerBar)
{


    switch(m_effect) {
        case REPEAT_EFFECT: {
            gettimeofday(&m_startTime,NULL);
            m_parameters=params;
            m_repeatSizeMs=int(float(beatDurationMs)*BEAT_DIVISIONS_ARRAY[int(floor(m_parameters[CONTROL_EFFECT_VALUE]*float(BEAT_DIVISIONS_NB-1)))]);
            m_duration = int(float(beatDurationMs * beatsPerBar) * m_parameters[CONTROL_EFFECT_VALUE]*2.0* m_parameters[CONTROL_EFFECT_LENGTH]);
            m_initialized=false;
        }break;
        case VELOCITY_EFFECT: {
            gettimeofday(&m_startTime,NULL);
            m_parameters=params;
            m_duration = int(float(beatDurationMs * beatsPerBar) * m_parameters[CONTROL_EFFECT_LENGTH]);
            m_initialized=false;
        }break;
        case PITCH_EFFECT: {
            gettimeofday(&m_startTime,NULL);
            m_parameters=params;
            m_duration = int(float(beatDurationMs * beatsPerBar) * m_parameters[CONTROL_EFFECT_LENGTH]);
            m_initialized=false;
        }break;
        case DELAY_EFFECT: {
            gettimeofday(&m_startTime,NULL);
            m_parameters=params;
            m_repeatSizeMs=int(float(beatDurationMs)*BEAT_DIVISIONS_ARRAY[int(floor(m_parameters[CONTROL_EFFECT_VALUE]*float(BEAT_DIVISIONS_NB-1)))]);
            m_duration = int(float(beatDurationMs * beatsPerBar) * m_parameters[CONTROL_EFFECT_LENGTH]);
            m_initialized=false;
        }break;
    }
    //cout<<"mode "<<m_parameters[SHOOT_MODE]<<" vel:"<<m_parameters[SHOOT_VELOCITY]<<endl;
}

