/***************************************************************************
 *  HllElement.cpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net 
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "HllElement.hpp"
#include "HllNode.hpp"
#include "AudioManager.hpp"

using namespace std;

HllElement::HllElement(const unsigned int& id): m_id(id), m_parent(NULL), 
            m_dynamicAnalysisCounter(0), m_dynamicAnalysisPeriod(5), 
            m_staticAnalysisSent(true), m_newEventsWaiting(false),
            m_parametersChanged(false), m_parametersChangedFromLoop(false), 
            m_loopState(LOOP_NONE), m_hasRecordedExcitationEvents(false) {

    for(int p=0; p<NODE_PARAMETERS_NB; ++p) {
        m_parametersValues[p]=0;
        m_parametersNewValues[p]=0;
        m_dryParametersValues[p]=0;
        m_parentParametersValues[p]=0.5;
        m_parentParametersNewValues[p]=0.5;
        m_hasRecordedModulationEvents[p]=false;
    }

    m_parametersValues[NODE_IRREGULARITY]=0;
    m_parametersNewValues[NODE_IRREGULARITY]=0;
    m_dryParametersValues[NODE_IRREGULARITY]=0;
    m_parentParametersValues[NODE_IRREGULARITY]=0;
    m_parentParametersNewValues[NODE_IRREGULARITY]=0;
    m_parametersValues[NODE_RESONANCE]=0;
    m_parametersNewValues[NODE_RESONANCE]=0;
    m_dryParametersValues[NODE_RESONANCE]=0;
    m_parentParametersValues[NODE_RESONANCE]=0;
    m_parentParametersNewValues[NODE_RESONANCE]=0;
    m_parametersValues[NODE_PITCH]=0.5;
    m_parametersNewValues[NODE_PITCH]=0.5;
    m_dryParametersValues[NODE_PITCH]=0.5;
    m_parametersValues[NODE_VOLUME]=0.5;
    m_parametersNewValues[NODE_VOLUME]=0.5;
    m_dryParametersValues[NODE_VOLUME]=0.5;
    m_parametersValues[NODE_SPEED]=0.5;
    m_parametersNewValues[NODE_SPEED]=0.5;
    m_dryParametersValues[NODE_SPEED]=0.5;
    m_parametersChanged=true;
}

HllElement::~HllElement() {}

void HllElement::resetParentParameters() {
    for(int p=0; p<NODE_PARAMETERS_NB; ++p) {
        m_parentParametersNewValues[p]=0.5;
    }
    m_parentParametersNewValues[NODE_RESONANCE]=0;
    m_parentParametersNewValues[NODE_IRREGULARITY]=0;
    m_parametersChanged=true;
}

void HllElement::processParentNewValues() {
    memcpy(m_parentParametersValues, m_parentParametersNewValues, 
            sizeof(double)*NODE_PARAMETERS_NB);
    //update audio parameters using parent parameters
    for(int p=0; p<NODE_PARAMETERS_NB; ++p) {
        m_dryParametersValues[p]=m_parametersNewValues[p];
        m_parametersValues[p]=m_parametersNewValues[p];
    }
    m_parametersValues[NODE_PITCH]=(m_parametersNewValues[NODE_PITCH])
                                + (m_parentParametersValues[NODE_PITCH]-0.5);
    m_parametersValues[NODE_VOLUME]
                        = (m_parametersNewValues[NODE_VOLUME])
                            * (2.0*m_parentParametersValues[NODE_VOLUME]);
    m_parametersValues[NODE_RESONANCE]
                        = (m_parametersNewValues[NODE_RESONANCE])
                            + m_parentParametersValues[NODE_RESONANCE];
    m_parametersValues[NODE_IRREGULARITY]
                         = (m_parametersNewValues[NODE_IRREGULARITY])
                            + m_parentParametersValues[NODE_IRREGULARITY];
}

void HllElement::processEvent(const Event& event, bool eventFromLoop) {
    switch(event.parameter) {    
        case NODE_EXCITE_INST: 
        case NODE_EXCITE_CONT: 
        case NODE_EXCITE_CONT_RELEASE: 
        case NODE_STOP: {
            if(!m_newEventsWaiting) {
                m_newEvents.clear();
            }
            //if recording loop and not an event coming from a loop
            //record the event in the corresponding loop
            if(m_loopState==LOOP_RECORDING && !eventFromLoop) {
                m_recordedExcitationLoopEvents.push_back(event);
                Event &newLoopEvent = 
                        m_recordedExcitationLoopEvents.back();
                timeval curTime, timeDif;
                gettimeofday(&curTime, NULL);
                timersub(&curTime, &m_loopRecordTime, &timeDif);
                newLoopEvent.timeD=timeDif.tv_sec*1000000+timeDif.tv_usec;


                //if second event recorded
                if(m_parent) {
                    if(m_recordedExcitationLoopEvents.size()>10) {
                        m_parent->loopStop(NODE_EXCITE_INST);
                    }
                }
            }
            m_newEvents.push_back(event);
            m_newEventsWaiting=true;
        }break; 
        default:{
            //if recording loop and not an event coming from a loop
            //record the event in the corresponding loop
            if(m_loopState==LOOP_RECORDING && !eventFromLoop) {
                m_recordedModulationLoopEvents[event.parameter]
                                                            .push_back(event);
                Event &newLoopEvent = 
                    m_recordedModulationLoopEvents[event.parameter].back();
                timeval curTime, timeDif;
                gettimeofday(&curTime, NULL);
                timersub(&curTime, &m_loopRecordTime, &timeDif);
                newLoopEvent.timeD=timeDif.tv_sec*1000000+timeDif.tv_usec;

                //stop the parent loop if rerecording
                if(m_parent) {
                    m_parent->loopStop(event.parameter);
                }
            }
            m_parametersNewValues[event.parameter]=event.values[0];
            m_parametersChanged=true;
        }break;
    }
    
    if(eventFromLoop) {
        m_parametersChangedFromLoop=true;
    }
}

void HllElement::process(const double &timeDiff,
                std::vector<std::vector<double> >& tempBuffer, 
                std::vector<std::vector<double> >& outputBuffer, 
                std::vector<Event>& outputEvents) {

    //one if parameters have been modified from loops
    if(m_parametersChangedFromLoop) {    
        outputEvents.push_back(Event());
        Event& newVisuEvent = outputEvents.back();
        newVisuEvent.type=EVENT_NODE;
        newVisuEvent.node=m_id;
        newVisuEvent.parameter=NODE_MODULATION;
        for(unsigned int i=0;i<NODE_PARAMETERS_NB;++i) {
            newVisuEvent.values[i]=m_dryParametersValues[i];
        }
        m_parametersChangedFromLoop=false;
    }

    //send analysis/visu
    if(m_dynamicAnalysisCounter>=m_dynamicAnalysisPeriod) {

        //dynamic 
        outputEvents.push_back(Event());
        Event& newEvent = outputEvents.back();
        newEvent.type=EVENT_NODE;
        newEvent.node=m_id;
        newEvent.parameter=NODE_ANALYSIS_DYNAMIC;
        newEvent.values[0]=m_dynamicAnalysis;

        //static
        if(!m_staticAnalysisSent) {
            //cout<<"send static analysis values"<<endl;
            outputEvents.push_back(Event());
            Event& newEvent = outputEvents.back();
            newEvent.type=EVENT_NODE;
            newEvent.node=m_id;
            newEvent.parameter=NODE_ANALYSIS_STATIC;
            for(unsigned int i=0;i<m_staticAnalysisValues.size();++i) {
                newEvent.values[i]=m_staticAnalysisValues[i];
            }
            m_staticAnalysisSent=true;
        }
        m_dynamicAnalysisCounter=0;
    }
    else {
        m_dynamicAnalysisCounter++;
    }
}

void HllElement::loopStartRecord(const timeval& loopStartTime) {
    if(m_loopState!=LOOP_RECORDING) {

        //clear previous recorded events
        clearRecordedLoopEvents();

        //store record time
        m_loopRecordTime=loopStartTime;

        //add a first fake event to all loops to get the correct starting time
        Event newLoopEvent;
        newLoopEvent.timeD=0;
        newLoopEvent.parameter=NODE_STOP;
        newLoopEvent.values[0]=0;
        m_recordedExcitationLoopEvents.push_back(newLoopEvent);
        for(unsigned int p=0;p<NODE_PARAMETERS_NB;++p) {
            newLoopEvent.parameter=p;
            newLoopEvent.values[0]=m_dryParametersValues[p];
            m_recordedModulationLoopEvents[p].push_back(newLoopEvent);        
        }

        //wait for first events    
        m_loopState=LOOP_RECORDING;
    }
}

void HllElement::loopStopRecord(const timeval& loopStopTime) {
    //record stop time
    m_loopRecordStopTime=loopStopTime;
    m_loopState=LOOP_NONE;

    //count excitation events
    double excitationEventsCount=0;
    for(unsigned int e=0;e<m_recordedExcitationLoopEvents.size();++e) {
        if(m_recordedExcitationLoopEvents[e].parameter==NODE_EXCITE_INST) {
            excitationEventsCount++;
        }
        else if(m_recordedExcitationLoopEvents[e].parameter==NODE_EXCITE_CONT) {
            excitationEventsCount+=0.1;
        }
    }
    if(excitationEventsCount>2) {
        //cout<<"recorded excitation loop !!!"<<endl;
        m_hasRecordedExcitationEvents=true;    
        //copy the last event with the current Time
        m_recordedExcitationLoopEvents.push_back(
                                    m_recordedExcitationLoopEvents.back());
        timeval timeDif;
        timersub(&loopStopTime, &m_loopRecordTime, &timeDif);
        m_recordedExcitationLoopEvents.back().timeD= 
                                timeDif.tv_sec*1000000+timeDif.tv_usec;    
        m_loopState=LOOP_RECORDED;
    }
    else {
        m_recordedExcitationLoopEvents.clear();
        m_hasRecordedExcitationEvents=false;
    }
    

    //count modulation events
    for(unsigned int p=0; p < NODE_PARAMETERS_NB; ++p) { 
        if(m_recordedModulationLoopEvents[p].size()>1) {
            m_hasRecordedModulationEvents[p]=true;
            //copy the last event with the current Time
            m_recordedModulationLoopEvents[p].push_back(
                                    m_recordedModulationLoopEvents[p].back());
            timeval timeDif;
            timersub(&loopStopTime, &m_loopRecordTime, 
                            &timeDif);
            m_recordedModulationLoopEvents[p].back().timeD=
                                timeDif.tv_sec*1000000+timeDif.tv_usec;    
            m_loopState=LOOP_RECORDED;
        }
        else {
            m_recordedModulationLoopEvents[p].clear();
            m_hasRecordedModulationEvents[p]=false;
        }
    }

    //if has parent, replace recorded loops in it
    if(m_parent) {
        if(m_hasRecordedExcitationEvents) { //if excitation loop was recorded
            //set the new loop
            m_parent->setExcitationLoop(m_recordedExcitationLoopEvents);
            //reset all other loop
            for(unsigned int p=0; p < NODE_PARAMETERS_NB; ++p) { 
                m_parent->setModulationLoop(p, 
                            m_recordedModulationLoopEvents[p]);
            }
            //play loop
            m_parent->loopPlay(0);
        }
        else {
            //set loops which have new events
            for(unsigned int p=0; p < NODE_PARAMETERS_NB; ++p) { 
                if(m_hasRecordedModulationEvents[p]) {
                    m_parent->setModulationLoop(p, 
                            m_recordedModulationLoopEvents[p]);
                }
            }
        }
    }
}

void HllElement::loopErase() {
    m_loopState=LOOP_NONE;
}

void HllElement::computeStaticAnalysis() {

}

bool HllElement::isElementInSubtree(HllElement* el) {
    if(m_id==el->getID()) {
        return true;
    }
    else {
        for(unsigned int c=0; c<m_children.size();++c) {
            if(m_children[c]->isElementInSubtree(el)) {
                return true;
            }
        }
    }
    return false;
}

void HllElement::print(const std::string& prefix) {
    cout<<prefix<<m_id<<endl;
}

