/***************************************************************************
 *  HllLeafLayeredGrain.cpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "HllLeafLayeredGrain.hpp"
#include "AudioManager.hpp"

using namespace std;

HllLeafLayeredGrain::HllLeafLayeredGrain(const int& id, const int& sampleRate,
                                        const int& bufferSize):
            HllLeaf(id, sampleRate, bufferSize), m_typeStr("lead"),
            m_baseSample(NULL), m_currentSample(NULL), m_filterL(NULL), 
            m_filterR(NULL), m_filterFrequency(400), 
            m_baseVolume(1.0), m_polyphony(3), m_baseSynth(false), 
            m_grainVolume(1.0), m_grainDensity(500), 
            m_grainNumber(0), m_grainNumberMax(20), m_grainSize(4096), 
            m_baseBassFrequency(65.0), m_spatialized(false) {

    timerclear(&m_lastParameterFadeTime);
    srand ( time(NULL) );    

    setVolume(0.5);
    setResonance(0);
    setPitch(0.5);
    setIrregularity(0);

    setType("lead");

    for(unsigned int i=0; i<10; ++i) {
       m_layeredGrainNodeEventReserve.push_back(new HllLeafLayeredGrainEvent());
    }
}

HllLeafLayeredGrain::~HllLeafLayeredGrain() {}

void HllLeafLayeredGrain::duplicate(std::vector<HllElement*>& createdElements, 
                                int& elementsIdCounter) {

    //duplicate ourself
    HllLeafLayeredGrain* duplicatedNode = 
                new HllLeafLayeredGrain(elementsIdCounter, m_sampleRate, 
                                        m_bufferSize);
    elementsIdCounter++;

    duplicatedNode->m_samples = m_samples;
    duplicatedNode->m_samplesDir = m_samplesDir;
    duplicatedNode->m_currentSample = m_currentSample;
    duplicatedNode->m_baseVolume = m_baseVolume;
    duplicatedNode->m_volume = m_volume;
    duplicatedNode->m_polyphony = m_polyphony;
    duplicatedNode->m_grainVolume = m_grainVolume;
    duplicatedNode->m_grainDensity = m_grainDensity;
    duplicatedNode->m_grainDistance = m_grainDistance;
    duplicatedNode->m_grainSize = m_grainSize;
    duplicatedNode->m_pitch = m_pitch;
    duplicatedNode->m_resonance = m_resonance;
    duplicatedNode->m_phaseStep = m_phaseStep;
    duplicatedNode->m_staticAnalysisValues = m_staticAnalysisValues;
    duplicatedNode->m_staticAnalysisSent=false;

    duplicatedNode->setType(m_typeStr);

    //finally add ourself to the vector
    createdElements.push_back(duplicatedNode);
}

void HllLeafLayeredGrain::parse(xmlNodePtr node) {
    char* value=NULL;
    //get samples dir
    value=NULL;
    value = (char*)xmlGetProp(node,(xmlChar*)"samples_directory");
    if(value!=NULL) {
        setSamplesDir(std::string(value));
    }
    value=NULL;
    value = (char*)xmlGetProp(node,(xmlChar*)"type");
    if(value!=NULL) {
        std::string typeStr = std::string(value);
        setType(typeStr);
    }
    value=NULL;
    value = (char*)xmlGetProp(node,(xmlChar*)"volume");
    if(value!=NULL) {
        setBaseVolume(atof(value));
        setVolume(atof(value));
    }
}

void HllLeafLayeredGrain::setType(const std::string& typeStr) {

    if(m_baseSample) {
        delete m_baseSample;
        m_baseSample=NULL;
    }
    m_spatialized=false;
    m_baseSynth=false;
    m_compressed=false;
    m_compressing=false;

    if(typeStr.compare("bass")==0) {
        m_baseSynth=true;
        m_filterFrequency=80.0;
        m_typeStr=typeStr;
        m_compressed=true;
    }
    else if(typeStr.compare("layer")==0) {
        m_baseSample=new Sample(m_samplesDir+"/base_layer.wav");
        try{
            m_baseSample->load();
        }
        catch(exception e) {
            cout<<"Error loading base_layer.wav in "
                <<m_samplesDir<<" "<<e.what()<<endl;
            delete m_baseSample;
            m_baseSample=NULL;
        }
        m_filterFrequency=300.0;
        m_typeStr=typeStr;
        m_compressing=true;
        m_compressed=true;
    }
    else if(typeStr.compare("cymbal")==0) {
        m_filterFrequency=2000.0;
        m_typeStr=typeStr;
        m_compressed=true;
    }
    else if(typeStr.compare("lead")==0) {
        m_filterFrequency=400.0;
        m_typeStr="lead";
        m_spatialized=true;
        m_compressed=true;
    }
    else {
        m_filterFrequency=400.0;
        m_typeStr="other";
    }

    if(m_filterL) {
        delete m_filterL;
    }
    if(m_filterR) {
        delete m_filterR;
    }
    
    //recreate filter
    m_filterL = new Filter(m_filterFrequency, m_sampleRate);
    m_filterR = new Filter(m_filterFrequency, m_sampleRate);
}

void HllLeafLayeredGrain::setBaseVolume(const float& vol) {
    m_baseVolume=vol;
}

void HllLeafLayeredGrain::setSamplesDir(const std::string& dir) {
    m_samplesDir=dir;
    m_samples.clear();

    DIR *dp2;
    struct dirent *dirp2;
    if((dp2  = opendir(m_samplesDir.c_str())) != NULL) {
        while ((dirp2 = readdir(dp2)) != NULL) {
            std::string sampNameStr = std::string(dirp2->d_name);
            //only wav files for now
            if(sampNameStr.find(".wav")!=sampNameStr.npos 
                    && sampNameStr.compare("base_layer.wav")!=0) {
                Sample* newSample =  new Sample(m_samplesDir+"/"+sampNameStr);
                try { 
                    newSample->load();
                } catch(exception e) {
                    delete newSample;
                }
                if(newSample) {
                    m_samples.push_back(newSample);
                }
            }
        }
    }
    if(m_samples.size()==0) {
        throw std::logic_error("Drile Node: No samples defined");
    }
    else {
        m_currentSample=m_samples[0];
    }
    computeStaticAnalysis();
}

void HllLeafLayeredGrain::setNoisiness(const double& noisiness) {
/*
    float noise = (noisiness<0)?0:(noisiness>1)?1:noisiness;
    //m_grainSize=4096.0*(1.0-noise)+128;
    m_grainDensity=double(m_grainAttack+m_grainSize+m_grainRelease)*
                        (noise*0.6+0.1);
    m_grainVolume=noise*0.2+0.2;
*/
}

void HllLeafLayeredGrain::setIrregularity(const double& irreg) {
    double nois = (irreg<0)?0:(irreg>1)?1:irreg;
	m_grainSize=4000.0*(1.0-((nois<0.5)?0:nois*2.0-1.0))+256;
	m_grainDensity=float(m_grainSize)*(0.2+0.6*((nois<0.5)?0:nois*2.0-1.0));
	m_grainVolume=1;
	m_grainDistance=((nois>0.5)?1:nois*2.0)*0.9+0.1;
}

void HllLeafLayeredGrain::setResonance(const double& resonance) {
    m_resonance = (resonance<0)?0:(resonance>1)?1:resonance;
    m_resonance=1.0/((m_resonance)*50000.0+100.0);
}

void HllLeafLayeredGrain::setPitch(const double& pitch) {
    m_pitch = (pitch<0.0001)?0.0001:(pitch>1)?1:pitch;
    m_pitch=pow(2.0,((m_pitch*128.0-64.0)/12.0));
}

void HllLeafLayeredGrain::setVolume(const double& volume) {
    m_volume = (volume<0)?0:(volume>1)?1:volume;
    m_volume=m_baseVolume*m_volume;
}

void HllLeafLayeredGrain::process(const double &timeDiff, 
                std::vector<std::vector<double> >& tempBuffer, 
                std::vector<std::vector<double> >& outputBuffer, 
                std::vector<Event>& outputEvents) {

    //FIXME do this in update, and create evtns for the audio thread
    if(m_parametersChanged) { //get new parameters values
            m_parametersChanged=false;
            processParentNewValues();
            //set new values
            setResonance(m_parametersValues[NODE_RESONANCE]);
            setPitch(m_parametersValues[NODE_PITCH]);
            setNoisiness(m_parametersValues[NODE_NOISINESS]);
            setVolume(m_parametersValues[NODE_VOLUME]);
            setIrregularity(m_parametersValues[NODE_IRREGULARITY]);
    }

    //get new events
    if(m_newEventsWaiting) {
        m_events = m_newEvents;
        m_newEventsWaiting = false;
        vector<Event>::iterator itAudioEvent = m_events.begin();
        for(;itAudioEvent!=m_events.end(); ++itAudioEvent) {
            //if it is a release event
            if((*itAudioEvent).parameter==NODE_EXCITE_CONT_RELEASE) {
                list<HllLeafLayeredGrainEvent*>::iterator itEvent 
                                    = m_layeredGrainNodeEvents.begin();
                for(;itEvent!=m_layeredGrainNodeEvents.end(); ++itEvent) {
                    if((*itEvent)->id==(*itAudioEvent).finger) {
                        (*itEvent)->fading=true;
                    }
                }
            } 
            else if((*itAudioEvent).parameter==NODE_STOP) {
                list<HllLeafLayeredGrainEvent*>::iterator itEvent 
                                    = m_layeredGrainNodeEvents.begin();
                for(;itEvent!=m_layeredGrainNodeEvents.end(); ++itEvent) {
                    if((*itEvent)->id==(*itAudioEvent).finger) {
                        (*itEvent)->fadeStep=0.01;
                        (*itEvent)->fading=true;
                    }
                }
            } 
            else { //if excitation event
                bool createEvent=true;
                //if continuous, test first if existing event and modify it
                if((*itAudioEvent).parameter==NODE_EXCITE_CONT) {  
                    list<HllLeafLayeredGrainEvent*>::iterator itEvent = 
                        m_layeredGrainNodeEvents.begin();
                    for(;itEvent!=m_layeredGrainNodeEvents.end(); ++itEvent) {
                        if((*itEvent)->type==NODE_EXCITE_CONT 
                                && (*itEvent)->fade > 0
                                && (*itEvent)->id==(*itAudioEvent).finger) {
                            createEvent=false;
                            (*itEvent)->fading=false;
                            (*itEvent)->fade=1.0;
                            (*itEvent)->velocity=(*itAudioEvent).values[0]/2.0;
                        }
                    }
                }
                else if(m_baseSynth) { //if instant and bass -> monophony
                    list<HllLeafLayeredGrainEvent*>::iterator itEvent = 
                        m_layeredGrainNodeEvents.begin();
                    for(;itEvent!=m_layeredGrainNodeEvents.end(); ++itEvent) {
                        if((*itEvent)->type==NODE_EXCITE_INST) {
                            (*itEvent)->fadeStep=0.01;
                            (*itEvent)->fading=true;
                        }
                    }
                }

                //create a new event
                if(createEvent && m_layeredGrainNodeEventReserve.size()>0) { 
                    m_layeredGrainNodeEvents.push_back(
                                    m_layeredGrainNodeEventReserve.back());
                    m_layeredGrainNodeEventReserve.pop_back();
                    HllLeafLayeredGrainEvent *newEvent = 
                                        m_layeredGrainNodeEvents.back();
                    newEvent->fade=1.0;
                    newEvent->fadeStep=m_resonance;
                    newEvent->firstPitch=m_pitch;
                    newEvent->id=(*itAudioEvent).finger;
                    newEvent->type=(*itAudioEvent).parameter;
                    if(newEvent->type==NODE_EXCITE_INST) { //if instant excit
                        newEvent->fading=true;
                        newEvent->velocity=(*itAudioEvent).values[0];
                        if((*itAudioEvent).offset>0) {
                            newEvent->fade=1.0-
                                    min<float>(1, 
                                        newEvent->fadeStep
                                            *(*itAudioEvent).offset
                                            *float(m_sampleRate)/1000000.0);
                        }
                    }
                    else {
                        newEvent->fading=false;
                        newEvent->velocity=(*itAudioEvent).values[0]/2.0;
                    }
                    newEvent->spatializeSide=true;

                    //add a first grain for the attack, 
                    //according to the velocity
                    newEvent->grains.push_back(newEvent->grainReserve.back());
                    newEvent->grainReserve.pop_back();
                    HllLeafLayeredGrainGrain *newGrain=newEvent->grains.back();
                    newGrain->sample = m_currentSample;
                    newGrain->type=HllLeafLayeredGrainGrain::GRAIN_NORMAL;
                    if(newEvent->type==NODE_EXCITE_INST) {
                        newGrain->counter=0;
                        if((*itAudioEvent).offset>0) {
                            newGrain->offset= 
                                rand()%int(m_grainDistance
                                *newGrain->sample->getFloatFramesCount())+1024;
                        }
                        else {
                            newGrain->offset=0;
                        }
                        newGrain->size=m_grainSize;
                    }
                    else {
                        newGrain->counter=0;
                        newGrain->offset= 
                            rand()%int(m_grainDistance
                                *newGrain->sample->getFloatFramesCount())+1024;
                        newGrain->size=m_grainSize;
                    }
                    newGrain->step=m_pitch;
                    newGrain->approxOffset=newGrain->offset;
                    newGrain->env=0;
                    newGrain->volumes[0]=
                       m_grainVolume*m_volume*newEvent->velocity*newEvent->fade;
                    newGrain->volumes[1]=newGrain->volumes[0];
                    if(m_spatialized) {
                       newGrain->volumes[0]*=(newEvent->spatializeSide)?0.01:1;
                       newGrain->volumes[1]*=(!newEvent->spatializeSide)?0.01:1;
                       newEvent->spatializeSide=!newEvent->spatializeSide;
                    }
                    newEvent->firstGrain->copy(newEvent->grains.back());

                    //if instant and base
                    if(m_baseSample && newEvent->type==NODE_EXCITE_INST) {
                        if((*itAudioEvent).offset==0) {
                            //add a grain for the base sample
                            newEvent->grains.push_back(
                                                newEvent->grainReserve.back());
                            newEvent->grainReserve.pop_back();
                            HllLeafLayeredGrainGrain *baseGrain
                                                    =newEvent->grains.back();
                            baseGrain->sample=m_baseSample;
                            baseGrain->offset=0;
                            baseGrain->counter=0;
                            baseGrain->step=1;
                            baseGrain->size=m_baseSample->getFramesCount();
                            baseGrain->type=
                                HllLeafLayeredGrainGrain::GRAIN_BASE_SAMPLE;
                            baseGrain->volumes[0]=
                                            m_grainVolume*m_volume*0.5
                                            *newEvent->velocity*newEvent->fade;
                        }
                    }
                    else if(m_baseSynth) {// && newEvent->type==NODE_EXCITE_CONT) { 
                        //add a grain for the base synth
                        newEvent->grains.push_back(
                                            newEvent->grainReserve.back());
                        newEvent->grainReserve.pop_back();
                        HllLeafLayeredGrainGrain* baseGrain
                                                =newEvent->grains.back();
                        baseGrain->offset=0;
                        baseGrain->synthStopped=false;
                        baseGrain->size=double(m_sampleRate)
                                        /(m_baseBassFrequency*(m_pitch));
                        baseGrain->env=0;
                        baseGrain->type=
                            HllLeafLayeredGrainGrain::GRAIN_BASE_SYNTH;
                        baseGrain->volumes[0]=m_grainVolume*m_volume
                                            *newEvent->velocity*newEvent->fade;
                        //add a grain for the sinusoid
                        m_phaseStep= 2.0*M_PI*m_baseBassFrequency
                                    *(m_pitch)/double(m_sampleRate);
                        baseGrain->envC=2.0*cos(m_phaseStep);
                        baseGrain->env1=0;
                        baseGrain->env2=cos(M_PI/2.0 + m_phaseStep);
                    }
                }
            }
        }
    }

    //zero out the stereo buffers
    tempBuffer[0].assign(tempBuffer[0].size(), 0);
    tempBuffer[1].assign(tempBuffer[1].size(), 0);
    tempBuffer[2].assign(tempBuffer[2].size(), 0);

    if(m_layeredGrainNodeEvents.size()>0) {
        //process the instrument events
        std::list<HllLeafLayeredGrainEvent*>::iterator itEvent 
                                        = m_layeredGrainNodeEvents.begin();    
        for(;itEvent!=m_layeredGrainNodeEvents.end();) {
            HllLeafLayeredGrainEvent *event = *itEvent;
            if(event->fade>0) {
                //if we should start a new grain
                if(event->grainDensityCounter>=m_grainDensity  
                                    && event->grainReserve.size()>0) {
                    //get the grain
                    event->grains.push_back(event->grainReserve.back());
                    event->grainReserve.pop_back();
                    HllLeafLayeredGrainGrain* newGrain=event->grains.back();
                    if(event->type==NODE_EXCITE_INST) {
                        //do not change pitch for inst events
                        newGrain->step=event->firstGrain->step;
                    }
                    else {
                        newGrain->step=m_pitch;
                    }
                    newGrain->size=m_grainSize;
                    newGrain->sample=m_currentSample;
                    newGrain->offset=
                                rand()%max<long>(128,long(m_grainDistance
                                    *(newGrain->sample->getFloatFramesCount()
                                    -m_grainSize*m_pitch-128)));
                    newGrain->approxOffset=newGrain->offset;
                    newGrain->counter=0;
                    newGrain->type=HllLeafLayeredGrainGrain::GRAIN_NORMAL;
                    newGrain->volumes[0]=
                        m_grainVolume*m_volume*event->velocity*event->fade;
                    newGrain->volumes[1]=newGrain->volumes[0];

                    if(m_spatialized) {
                        newGrain->volumes[0]*=(event->spatializeSide)?0.01:1;
                        newGrain->volumes[1]*=(!event->spatializeSide)?0.01:1;
                        event->spatializeSide=!event->spatializeSide;
                    }
                    event->grainDensityCounter-=m_grainDensity;
                }
                event->grainDensityCounter+=outputBuffer[0].size();
            }


            //process existing grains: 
            //put normal in temp[1] and temp[2] because 
            //they will be filtered, and base/synth in temp[0]
            std::list<HllLeafLayeredGrainGrain*>::iterator itGrain
                                                    = event->grains.begin();
            for(; itGrain!=event->grains.end();) {
                HllLeafLayeredGrainGrain* grain = *itGrain;
                switch(grain->type) {
                    case HllLeafLayeredGrainGrain::GRAIN_BASE_SAMPLE:{
                        std::vector<double>::iterator itFrame 
                                                = tempBuffer[0].begin();
                        for(;itFrame!=tempBuffer[0].end() 
                            && grain->offset<grain->sample->getFramesCount(); 
                                itFrame++, grain->offset++) {

                            *itFrame+=m_baseSample->getFrame(grain->offset)
                                            *grain->volumes[0];
                        }
                        if(grain->offset>=grain->sample->getFramesCount()) {
                            event->grainReserve.push_back(*itGrain);
                            itGrain=event->grains.erase(itGrain);
                        }
                        else {
                            ++itGrain;
                        }
                    }break;
                    case HllLeafLayeredGrainGrain::GRAIN_BASE_SYNTH:{
                        std::vector<double>::iterator itFrame 
                                                = tempBuffer[0].begin();
                        for(;itFrame!=tempBuffer[0].end() 
                                && !grain->synthStopped; itFrame++) {
                            *itFrame+=grain->env*grain->volumes[0];
                            grain->env=grain->envC*grain->env1 - grain->env2;
                            grain->env2=grain->env1;
                            grain->env1=grain->env;
                            grain->offset++;

                            //if reached the end of a period, 
                            //create a new synth grain with updated freq
                            if(grain->offset>=grain->size) {
                                if(event->fading && event->fade<=0) {
                                    grain->synthStopped=true;
                                }
                                else {
                                    grain->offset=0;
                                    grain->env=0;
                                    grain->volumes[0]=m_grainVolume
                                                    *m_volume*event->velocity
                                                    *event->fade;
                                    //add a grain for the sinusoid
                                    if(event->type==NODE_EXCITE_INST) {
                                        m_phaseStep=2.0*M_PI*m_baseBassFrequency
                                                    *event->firstPitch
                                                    /double(m_sampleRate);
                                        grain->size=double(m_sampleRate)
                                                    /(m_baseBassFrequency
                                                    *event->firstPitch);
                                    }
                                    else {
                                        m_phaseStep=2.0*M_PI*m_baseBassFrequency
                                                *m_pitch/double(m_sampleRate);
                                        grain->size=double(m_sampleRate)
                                                /(m_baseBassFrequency*m_pitch);
                                    }
                                    grain->envC=2.0*cos(m_phaseStep);
                                    grain->env1=0;
                                    grain->env2=cos(M_PI/2.0 + m_phaseStep);
                                }
                            }
                        }
                        if(grain->synthStopped) {
                            event->grainReserve.push_back(*itGrain);
                            itGrain=event->grains.erase(itGrain);
                        }
                        else {
                            ++itGrain;
                        }
                    }break;
                    default: { //NORMAL grain
                        std::vector<double>::iterator itFrameL 
                            = tempBuffer[1].begin();
                        std::vector<double>::iterator itFrameR 
                            = tempBuffer[2].begin();
                        while(grain->counter<grain->size
                                && itFrameL!=tempBuffer[1].end()
                                && itFrameR!=tempBuffer[2].end()) {
                            //get frames
                            m_tempVal=
                                grain->sample->getLinInterpFrame(
                                                    grain->approxOffset)
                                *EnvelopeManager::getInstance()
                                    ->getEnvelope(grain->size)[grain->counter];
                            (*itFrameL)+=m_tempVal*grain->volumes[0];
                            (*itFrameR)+=m_tempVal*grain->volumes[1];
                            //increment positions
                            itFrameL++;
                            itFrameR++;
                            grain->counter++;
                            grain->approxOffset=fmod(
                                grain->approxOffset+grain->step, 
                                grain->sample->getFloatFramesCount());
                            grain->offset=int(grain->approxOffset);
                        }
                        //if done with the grain, remove it
                        if(grain->counter>=grain->size) {
                            event->grainReserve.push_back(*itGrain);
                            itGrain=event->grains.erase(itGrain);
                        }
                        else {
                            ++itGrain;
                        }
                    }break;
                }
            }

            //fade
            if(event->fading && event->fade>0) {
                event->fade=max<float>(0, 
                    event->fade-event->fadeStep*float(tempBuffer[0].size()));
            }

            //if all grains have stopped, erase the event
            if(event->fading && event->fade<=0 && event->grains.size()==0) {
                m_layeredGrainNodeEventReserve.push_back(*itEvent);
                itEvent=m_layeredGrainNodeEvents.erase(itEvent);
            }
            else {
                ++itEvent;
            }
        }    

        //filter normal grains
        m_filterL->process(tempBuffer[1]);
        m_filterR->process(tempBuffer[2]);

        //add base samples or sines
        if(m_baseSample || m_baseSynth) {
            vector<double>::iterator itFrameBase = tempBuffer[0].begin();
            vector<double>::iterator itFrameNormal = tempBuffer[1].begin();
            vector<double>::iterator itFrameNormalR = tempBuffer[2].begin();
            for(; itFrameBase!=tempBuffer[0].end() 
                        && itFrameNormal!=tempBuffer[1].end() 
                        && itFrameNormalR!=tempBuffer[2].end(); 
                    itFrameBase++, itFrameNormal++, itFrameNormalR++) {

                *itFrameNormal+=*itFrameBase;
                *itFrameNormalR+=*itFrameBase;
            }
        }
    }

    //dynamic analysis
    if(m_dynamicAnalysisCounter>=m_dynamicAnalysisPeriod) {
        for(unsigned int i=0;i<tempBuffer[1].size();++i) {
            m_analysisBuffer[i]=tempBuffer[1][i]+tempBuffer[2][i];
        }
        xtract_rms_amplitude(&(m_analysisBuffer[0]), m_analysisBuffer.size(), 
                                NULL, &m_dynamicAnalysis);
        if(m_dynamicAnalysis>1.0) {
            m_dynamicAnalysis=1.0;
        }
    }

    //copy depending on type: normal, compressing, compressed
    vector<double>::iterator itFrameTempL = tempBuffer[1].begin();
    vector<double>::iterator itFrameTempR = tempBuffer[2].begin();
    vector<double>::iterator itFrameOutL;
    vector<double>::iterator itFrameOutR;
    if(m_compressed) {
        itFrameOutL = outputBuffer[2].begin();
        itFrameOutR = outputBuffer[3].begin();
        for(; itFrameTempL!=tempBuffer[1].end() 
                && itFrameTempR!=tempBuffer[2].end()
                && itFrameOutL!=outputBuffer[2].end() 
                && itFrameOutR!=outputBuffer[3].end(); 
                itFrameOutL++, itFrameOutR++, itFrameTempR++, itFrameTempL++) {

            *itFrameOutL+=*itFrameTempL;
            *itFrameOutR+=*itFrameTempR;
        }
    }
    else {
        itFrameOutL = outputBuffer[0].begin();
        itFrameOutR = outputBuffer[1].begin();
        for(; itFrameTempL!=tempBuffer[1].end() 
                && itFrameTempR!=tempBuffer[2].end()
                && itFrameOutL!=outputBuffer[0].end() 
                && itFrameOutR!=outputBuffer[1].end(); 
                itFrameOutL++, itFrameOutR++, itFrameTempR++, itFrameTempL++) {

            *itFrameOutL+=*itFrameTempL;
            *itFrameOutR+=*itFrameTempR;
        }
        if(m_compressing) {
            itFrameTempL = tempBuffer[1].begin();
            itFrameTempR = tempBuffer[2].begin();
            itFrameOutL = outputBuffer[4].begin();
            for(; itFrameTempL!=tempBuffer[1].end() 
                    && itFrameTempR!=tempBuffer[2].end()
                    && itFrameOutL!=outputBuffer[4].end();
                    itFrameOutL++, itFrameOutR++, 
                    itFrameTempR++, itFrameTempL++) {

                *itFrameOutL+=(*itFrameTempL+*itFrameTempR);
            }
        }
    }

    //element process
    HllLeaf::process(timeDiff, tempBuffer, outputBuffer, outputEvents);
}

void HllLeafLayeredGrain::computeStaticAnalysis() {

    AudioManager* audioManager = audioManager->getInstance();

    float argf[4];
    argf[0] = (double)m_sampleRate/(double)AudioManager::FFT_SIZE;
    argf[1] = XTRACT_MAGNITUDE_SPECTRUM;
    argf[2] = 0.f;
    argf[3] = 0.f;
    double tempRms=0,rms=0;

    vector<double> tempBuffer(AudioManager::FFT_SIZE,0);
    vector<double> spectrumBuffer(AudioManager::FFT_SIZE,0);;
    vector<double> tempBarkBuffer(XTRACT_BARK_BANDS,0);
    std::vector<double> barkBuffer(XTRACT_BARK_BANDS, 0);

    //go through all the samples and add all the bark spectrums
    double spectrumBuffersCount=0;
    //add all the loaded samples to a vector
    std::vector<Sample*> analysedSamples;
    vector<Sample*>::iterator itSample = m_samples.begin();
    for(;itSample!=m_samples.end();++itSample) {
        analysedSamples.push_back(*itSample);
    }


    //analyse each sample of the vector
    vector<Sample*>::iterator itAnalysedSample = analysedSamples.begin();
    for(;itAnalysedSample!=analysedSamples.end();++itAnalysedSample) {
        if((*itAnalysedSample)->isLoaded()) {
            bool sampleAnalysed=false;
            long offset=0;
            while(!sampleAnalysed) {
                int length = AudioManager::FFT_SIZE;
                if((*itAnalysedSample)->getFramesCount() - offset < length) {
                    length = (*itAnalysedSample)->getFramesCount() - offset;
                }
                for(int i=0;i<length;++i) {
                    tempBuffer[i] = (*itAnalysedSample)->getFrame(offset);
                    offset++;
                }
                //if finished
                if(length<AudioManager::FFT_SIZE) {
                    sampleAnalysed=true;
                }
                //analyse the buffer
                xtract_spectrum(&tempBuffer[0], AudioManager::FFT_SIZE, 
                                &argf[0], &spectrumBuffer[0]);
                xtract_bark_coefficients(&spectrumBuffer[0], 
                                        AudioManager::FFT_SIZE/2, 
                                        audioManager->getBarkBands(), 
                                        &tempBarkBuffer[0]);
                for(unsigned int i=0;i<XTRACT_BARK_BANDS;++i) {
                    barkBuffer[i]+=tempBarkBuffer[i];
                }
                spectrumBuffersCount++;

                //get max rms
                xtract_rms_amplitude(&(tempBuffer[0]), tempBuffer.size(),NULL, 
                                    &tempRms);
                if(tempRms>rms) {
                    rms=tempRms;
                }
            }
        }
    }

    //maximize volume of sample
    m_baseVolume*=0.9/rms;

    m_staticAnalysisValues.clear();
    for(unsigned int i=0;i<XTRACT_BARK_BANDS;++i) {
        m_staticAnalysisValues.push_back(barkBuffer[i]/spectrumBuffersCount);
    }

    if(m_baseSynth) {
        m_staticAnalysisValues[0]=1;
    }


    m_staticAnalysisSent=false;
}

void HllLeafLayeredGrain::loopStop(const int& param) {
    list<HllLeafLayeredGrainEvent*>::iterator itEvent 
        = m_layeredGrainNodeEvents.begin();
    for(;itEvent!=m_layeredGrainNodeEvents.end();++itEvent) {
        (*itEvent)->fading=true;
    }
}


