/***************************************************************************
 *  HllNodeLoop.cpp
 *  Part of Drile
 *  2010- Florent Berthaut
 *  hitmuri.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "HllNodeLoop.hpp"

using namespace std;

HllNodeLoop::HllNodeLoop(const int& id): HllNode(id),  
                                        m_excitationLoopLength(0),
                                        m_excitationLoopCurrentPosition(0),
                                        m_excitationLoopElapsedTime(0),
                                        m_loopSpeed(1.0),
                                        m_sentStopPosition(false) {

    gettimeofday(&m_loopPreviousTime, NULL);
    m_parametersValues[NODE_OFFSET]=0;
    m_parametersValues[NODE_SPEED]=0.5;


    for(unsigned int i=0;i<NODE_PARAMETERS_NB;++i) {
        m_modulationLoopsPlaying[i]=false;
    }
    m_excitationLoopPlaying=false;

}

void HllNodeLoop::duplicate(std::vector<HllElement*>& createdElements, 
                                                int& elementsIdCounter) {
    //duplicate ourself
    HllNodeLoop* duplicatedNode = new HllNodeLoop(elementsIdCounter);
    elementsIdCounter++;

    //copy parameters and loop data
    duplicatedNode->m_excitationLoopEvents = m_excitationLoopEvents;
    duplicatedNode->m_excitationLoopIndex = m_excitationLoopIndex;
    duplicatedNode->m_excitationLoopLength = m_excitationLoopLength;
    duplicatedNode->m_excitationLoopCurrentPosition=0;

    for(unsigned int p=0;p<NODE_PARAMETERS_NB;++p) {
        duplicatedNode->m_parametersValues[p]=m_parametersValues[p];
        
        duplicatedNode->m_modulationLoopsEvents[p]=m_modulationLoopsEvents[p];
        duplicatedNode->m_modulationLoopsIndex[p]=m_modulationLoopsIndex[p];
        duplicatedNode->m_modulationLoopsCurrentPosition[p]=0;
        duplicatedNode->m_modulationLoopsLength[p]=m_modulationLoopsLength[p];
        duplicatedNode->m_loopPreviousTime=m_loopPreviousTime;
        duplicatedNode->m_loopSpeed=m_loopSpeed;
        duplicatedNode->m_sentStopPosition=m_sentStopPosition;
    }

    //duplicate and add the child
    m_children[0]->duplicate(createdElements, elementsIdCounter);
    duplicatedNode->addChild(createdElements.back());

    //add the duplicate to the vector
    createdElements.push_back(duplicatedNode);
}

HllNodeLoop::~HllNodeLoop() {

}

void HllNodeLoop::parse(xmlNodePtr) {

}

void HllNodeLoop::addChild(HllElement* child) {
    if(m_children.size()==0) {
        m_children.push_back(child);
    }
    else {
        m_children[0]=child;
    }
    m_children[0]->setParent(this);
    m_parametersChanged=true;
}

void HllNodeLoop::removeChild(HllElement* child) {
    if(m_children.size()>0) {
        if(m_children[0]) {
            if(m_children[0]->getID()==child->getID()) {
                m_children.clear();
            }
        }
    }
}

void HllNodeLoop::setExcitationLoop(std::vector<Event>& loop) {
    m_excitationLoopEvents=loop;
    m_excitationLoopLength=double(m_excitationLoopEvents.back().timeD);
    //synchronize all non empty modulation loops 
    for(unsigned int p=0;p<NODE_PARAMETERS_NB;++p) {
        if(m_modulationLoopsEvents[p].size()>0) {
            synchronizeModulationLoop(p);
        }
    }
}

void HllNodeLoop::setModulationLoop(const unsigned int& parameter, 
                                    std::vector<Event>& loop) {
    m_modulationLoopsEvents[parameter]=loop;
    //sync to excitation loop
    if(m_modulationLoopsEvents[parameter].size()>0) {
        synchronizeModulationLoop(parameter);
    }
}

void HllNodeLoop::synchronizeModulationLoop(const unsigned int& param) {
    double loopLength = double(m_modulationLoopsEvents[param].back().timeD);
    double pulseLength = m_excitationLoopLength;
    double epsilon=0.5; //500ms error epsilon

    if(loopLength>pulseLength-epsilon) {
        double mult = loopLength/pulseLength;
        if(mult<1.0) {
            m_modulationLoopsLength[param]=pulseLength;
        } else if(mult - floor(mult)>0.5) {
            m_modulationLoopsLength[param]=pulseLength*ceil(mult);
        }
        else {
            m_modulationLoopsLength[param]=pulseLength*floor(mult);
        }
    }
    else {
        double div = pulseLength/loopLength;
        if(div - floor(div)>0.5) {
            m_modulationLoopsLength[param]=pulseLength/ceil(div);
        }
        else {
            m_modulationLoopsLength[param]=pulseLength/floor(div);
        }
    }
}

void HllNodeLoop::process(const double& timeDiff, 
    std::vector<std::vector<double> >& tempBuffer, 
        std::vector<std::vector<double> >& outputBuffer, 
            std::vector<Event>& outputEvents) {

    //process events
    if(m_newEventsWaiting) {
        m_events = m_newEvents;
        m_newEventsWaiting=false;
        for(unsigned int evID=0; evID<m_events.size(); evID++) { 
            //play event with a non null velocity
            if(m_events[evID].parameter==NODE_EXCITE_INST 
                    && m_events[evID].values[0]>0) {
                loopPlay(m_events[evID].offset);
            }
            else if(m_events[evID].parameter==NODE_STOP){ //stop event
                loopStop();
            }
        }
        m_events.clear();
    }

    //play loops
    if(m_isPlaying) {
        double playDiffWithSpeed = 
                timeDiff*pow(2,m_parametersValues[NODE_SPEED]*2.0-1.0);
        m_excitationLoopElapsedTime += playDiffWithSpeed;
        if(!m_parent) {
            m_excitationLoopElapsedTime=fmod(m_excitationLoopElapsedTime,
                                                m_excitationLoopLength);
        }
        if(m_excitationLoopElapsedTime<m_excitationLoopLength) {
            if(m_excitationLoopPlaying) {
                processLoop(m_excitationLoopEvents, m_excitationLoopIndex, 
                            m_excitationLoopCurrentPosition, playDiffWithSpeed);
            }
            for(unsigned int p=0;p<NODE_PARAMETERS_NB;++p) {
                if(m_modulationLoopsPlaying[p]) {
                    processLoop(m_modulationLoopsEvents[p], 
                        m_modulationLoopsIndex[p], 
                        m_modulationLoopsCurrentPosition[p], playDiffWithSpeed);
                }
            }
        }
    }

    
    if(m_dynamicAnalysisCounter>=m_dynamicAnalysisPeriod) {
        if(m_isPlaying) {
            //send loop position
            outputEvents.push_back(Event());
            Event& newEvent = outputEvents.back();
            newEvent.type=EVENT_NODE;
            newEvent.node=m_id;
            newEvent.parameter=NODE_LOOP_POSITION;
            newEvent.values[0]=
                    m_excitationLoopCurrentPosition/m_excitationLoopLength;
        }
        else {
            outputEvents.push_back(Event());
            Event& newEvent = outputEvents.back();
            newEvent.type=EVENT_NODE;
            newEvent.node=m_id;
            newEvent.parameter=NODE_LOOP_POSITION;
            newEvent.values[0]=m_parametersValues[NODE_OFFSET];
            m_sentStopPosition=true;
        }
    }
    //normal process
    HllNode::process(timeDiff, tempBuffer, outputBuffer, outputEvents);
}

void HllNodeLoop::loopStop(const int& param) {
    HllNode::loopStop(param);
    switch(param) {
        case -1: { //stop everything
            m_excitationLoopPlaying=false;
            for(unsigned int i=0;i<NODE_PARAMETERS_NB;++i) {
                m_modulationLoopsPlaying[i]=false;
            }
            m_isPlaying=false;
        }break;
        case NODE_EXCITE_INST: {
            m_excitationLoopPlaying=false;
        }break;
        default: {
            m_modulationLoopsPlaying[param]=false;
        }
    }
}

void HllNodeLoop::loopPlay(double timeOffset) {
    //get start position from offset parameter
    Event& lastExcitEvent = m_excitationLoopEvents.back();
    m_excitationLoopCurrentPosition = m_parametersValues[NODE_OFFSET] * 
                                        double(lastExcitEvent.timeD);
    m_excitationLoopElapsedTime=timeOffset;
    Event prevEvent;
    bool prevEventFound=false;
    for(unsigned int i=0; i<m_excitationLoopEvents.size(); ++i) {
        if(m_excitationLoopEvents[i].parameter==NODE_EXCITE_INST) {
            prevEvent=m_excitationLoopEvents[i];
            prevEventFound=true;
        }
        if(m_excitationLoopEvents[i].timeD>=
                (unsigned long)(m_excitationLoopCurrentPosition)) {
            m_excitationLoopIndex=i;
            break;
        }
    }
    m_excitationLoopCurrentPosition += timeOffset;

    //process previous inst excit event with time offset
    if(prevEventFound) {
        prevEvent.offset
                = m_excitationLoopCurrentPosition-(double)prevEvent.timeD;
        m_children[0]->processEvent(prevEvent, true);
    }

    for(unsigned int p=0;p<NODE_PARAMETERS_NB;++p) {
        if(m_modulationLoopsEvents[p].size()>0) {
            Event& lastModulEvent = m_modulationLoopsEvents[p].back();
            m_modulationLoopsCurrentPosition[p] = 
                m_parametersValues[NODE_OFFSET] * double(lastModulEvent.timeD);
            for(unsigned int i=0; i<m_modulationLoopsEvents[p].size(); ++i) {
                if(m_modulationLoopsEvents[p][i].timeD>=
                        (unsigned long)(m_modulationLoopsCurrentPosition[p])) {
                    m_modulationLoopsIndex[p]=i;
                    break;
                }
            }
            m_modulationLoopsCurrentPosition[p] += timeOffset;
            //process previous event with time offset
            prevEvent=m_modulationLoopsEvents[p][(m_modulationLoopsIndex[p]-1)
                                            %m_modulationLoopsEvents[p].size()];
            prevEvent.offset=m_modulationLoopsCurrentPosition[p]
                                - (double)prevEvent.timeD;
            m_children[0]->processEvent(prevEvent, true);
        }
    }
    m_isPlaying=true;
    m_sentStopPosition=false;

    for(unsigned int i=0;i<NODE_PARAMETERS_NB;++i) {
        m_modulationLoopsPlaying[i]=true;
    }
    m_excitationLoopPlaying=true;
}


void HllNodeLoop::processLoop(std::vector<Event>& loopEvents, 
                            unsigned int& loopIndex, 
                            double& loopCurrentPosition, 
                            double& timeDiffWithSpeed) {
    if(loopEvents.size()>0) {
        loopCurrentPosition+=timeDiffWithSpeed;

        //go through events which should have already been sent
        bool eventBeforeCurTime=true;
        while(eventBeforeCurTime && loopIndex<loopEvents.size()) {
            if(loopEvents[loopIndex].timeD < 
                                        (unsigned long)(loopCurrentPosition)) {
                m_children[0]->processEvent(loopEvents[loopIndex], true);
                loopIndex++;
            }
            else {
                eventBeforeCurTime=false;
            }
        }
        //if reached the end of the loop
        if(loopIndex>=loopEvents.size()) {
            loopCurrentPosition = 0;
            loopIndex = 0;
        }
    }
}


