/***************************************************************************
 *  HllNodeSync.cpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "HllNodeSync.hpp"
#include "AudioManager.hpp"

using namespace std;

HllNodeSync::HllNodeSync(const int& id): HllNode(id) {}

void HllNodeSync::duplicate(std::vector<HllElement*>& createdElements, 
                                                int& elementsIdCounter) {

    //cout<<"duplicate sync node "<<m_id<<endl;

    //duplicate ourself
    HllNodeSync* duplicatedNode = new HllNodeSync(elementsIdCounter);
    elementsIdCounter++;
    
    //duplicate and add the children
    vector<HllElement*>::reverse_iterator itEl;
    for(itEl=m_children.rbegin();
            itEl!=m_children.rend(); ++itEl) {
        (*itEl)->duplicate(createdElements, elementsIdCounter);
        duplicatedNode->addChild(createdElements.back());
    }

    duplicatedNode->loopStop();

    //add the duplicate to the vector
    createdElements.push_back(duplicatedNode);
}

HllNodeSync::~HllNodeSync() {}

void HllNodeSync::parse(xmlNodePtr) {}

void HllNodeSync::processEvent(const Event& event, bool eventFromLoop) {
    switch(event.parameter) {
        case NODE_SPEED: 
        case NODE_OFFSET: { 
            for(unsigned int c=0; c<m_children.size(); ++c) {    
                static_cast<HllNodeLoop*>(
                    m_children[c])->processEvent(event, eventFromLoop);
            }
        }break;
        default:break;
    }
    HllNode::processEvent(event, eventFromLoop);
}

void HllNodeSync::process(const double& timeDiff, 
    std::vector<std::vector<double> >& tempBuffer, 
        std::vector<std::vector<double> >& outputBuffer, 
            std::vector<Event>& outputEvents) {

    //process events
    if(m_newEventsWaiting) {
        m_events = m_newEvents;
        m_newEventsWaiting=false;
        for(unsigned int evID=0; evID<m_events.size(); evID++) { 
            if(m_events[evID].parameter==NODE_EXCITE_INST 
                                    && m_events[evID].values[0]>0) {
                for(unsigned int l=0; l<m_children.size(); ++l) {    
                    static_cast<HllNodeLoop*>(
                        m_children[l])->loopPlay();
                    m_childrenData[l].loopCounter=0;
                }
                m_isPlaying=true;
            }
            else if(m_events[evID].parameter==NODE_STOP){ //stop event
                loopStop();
            }
        }
        m_events.clear();
    }

    if(m_dynamicAnalysisCounter>=m_dynamicAnalysisPeriod) {
        if(m_isPlaying) {
            outputEvents.push_back(Event());
            Event& newEvent = outputEvents.back();
            newEvent.type=EVENT_NODE;
            newEvent.node=m_id;
            newEvent.parameter=NODE_LOOP_POSITION;
            newEvent.values[0]=m_pulseCounter/m_pulseDryLength;
        }
    }

    //normal process
    HllNode::process(timeDiff, tempBuffer, outputBuffer, outputEvents);

    //advance counter for visualization
    m_pulseCounter=fmod(m_pulseCounter+timeDiff*m_pulseRatio, m_pulseDryLength);

    //test restarting each child in order to keep them synchronized
    for(unsigned int c=0; c<m_children.size(); ++c) {
        SyncedChildData &data = m_childrenData[c];
        if(static_cast<HllNodeLoop*>(m_children[c])->getLoopElapsedTime()
                >= data.loopPulses[data.loopCounter]) {

            double diff = 
                static_cast<HllNodeLoop*>(m_children[c])->getLoopElapsedTime()
                - data.loopPulses[data.loopCounter];

            static_cast<HllNodeLoop*>(m_children[c])->loopPlay(diff);
            data.loopCounter=(data.loopCounter+1)%data.loopPulses.size();


            //cout<<"restarting child "<<c<<" with diff "<<diff<<endl;
        }
    }

}

void HllNodeSync::removeChild(HllElement* el) {
    for(unsigned int c=0; c<m_children.size();) {
        if(m_children[c]->getID() == el->getID()) {
            m_children.erase(m_children.begin()+c);
            m_childrenData.erase(m_childrenData.begin()+c);
            break;
        }
        else {
             ++c;
        }
    }
}

void HllNodeSync::loopStop(const int& param) {
    //if loopStop not coming from a child
    if(param==-1) {
        HllNode::loopStop();
    }
}

void HllNodeSync::setPulseNode(HllNodeLoop* pulseNode) {
    m_pulseDryLength = pulseNode->getLoopLength();
    m_pulseRatio=1.0;
    m_pulseLength=m_pulseDryLength;
    //resync all children to this pulse
    for(unsigned int c=0;c<m_children.size();++c) {
        synchronizeChild(c);
    }
}

void HllNodeSync::addChild(HllElement* el) {
    m_children.push_back((HllNodeLoop*)el);
    m_childrenData.push_back(SyncedChildData());
    el->setParent(this);

    //sync new child
    if(m_children.size()>1) {
        synchronizeChild(m_children.size()-1);
    }
    else { //or set it as the new pulse
        setPulseNode(static_cast<HllNodeLoop*>(el));
    }
    m_parametersChanged=true;
}

void HllNodeSync::synchronizeChild(const unsigned int& child) {
    
    m_childrenData[child].loopPulses.clear();
    const double& loopLength = 
            static_cast<HllNodeLoop*>(m_children[child])->getLoopLength();
    double epsilon=100000; //100ms error epsilon
    //cout<<"sync loop "<<loopLength<<" with pulse "<<m_pulseLength<<endl;
    if(loopLength>m_pulseLength-epsilon) {
    //    cout<<"child "<<child<<" longer than pulse and ";
        double mult = loopLength/m_pulseLength;
        if(mult<1.0) {
    //        cout<<"synced to pulse"<<endl;
            m_childrenData[child].loopPulses.push_back(m_pulseLength);
        } else if(mult - floor(mult)>0.5) {
    //        cout<<"synced to pulse*"<<ceil(mult)<<endl;
            m_childrenData[child].loopPulses.push_back(
                                        m_pulseLength*ceil(mult));
        }
        else {
    //        cout<<"synced to pulse*"<<floor(mult)<<endl;
            m_childrenData[child].loopPulses.push_back(
                                            m_pulseLength*floor(mult));
        }
    }
    else {
        //cout<<"child "<<child<<" shorter than pulse and ";
        double div = m_pulseLength/loopLength;
        if(div - floor(div)>0.5) {
            //cout<<"synced to pulse/"<<ceil(div)<<endl;
            for(double i=0;i<ceil(div);++i) {
                m_childrenData[child].loopPulses.push_back(
                                    floor(m_pulseLength/ceil(div)));
            }
            m_childrenData[child].loopPulses.back()+=
                                        fmod(m_pulseLength,ceil(div));
        }
        else {
            //cout<<"synced to pulse/"<<floor(div)<<endl;
            //m_childrenData[child].loopPulse=m_pulseLength/floor(div);
            for(double i=0;i<floor(div);++i) {
                m_childrenData[child].loopPulses.push_back(
                                    floor(m_pulseLength/floor(div)));
            }
            m_childrenData[child].loopPulses.back()+=
                                        fmod(m_pulseLength,floor(div));
        }
    }
    m_childrenData[child].loopCounter=0;
}

