/***************************************************************************
 *  Sample.cpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "Sample.hpp"

using namespace std;

Sample::Sample(const std::string& name): m_name(name), m_loaded(false), 
                                        m_nbUsers(0), m_defaultFrame(0) {
}

Sample::~Sample()
{
    unload();
}

const std::vector<double>& Sample::getBuffer() const
{
    return m_buffer;
}

void Sample::load()
{
    //get the file
    SF_INFO sfInfo;
    sfInfo.format=0;
    SNDFILE* sndFile = sf_open(m_name.c_str(), SFM_READ, &sfInfo);

    if(!sndFile) {
        throw std::logic_error("Unable to load sample "+m_name);
    }
    
    //test the Samplerate
    m_sampleRate = sfInfo.samplerate;
    
    //channels
    m_channelsCount = sfInfo.channels;
    
    //frames number
    m_framesCount = (long)(sfInfo.frames);
    m_doubleFramesCount = (double)(sfInfo.frames);

    //create the buffers
    m_buffer.resize(m_framesCount);
	int maxNbFramesToRead=1000;
	float cf[sfInfo.channels * maxNbFramesToRead];
	int nbFrames;
	int pos = 0;
	while ((nbFrames = sf_readf_float(sndFile, cf, maxNbFramesToRead)) > 0) {
		for (int i = 0; i < nbFrames*sfInfo.channels; ++pos) {
            m_buffer[pos]=0;
			for(int j=0; j<sfInfo.channels; ++j) {
				m_buffer[pos] += cf[i];
				++i;
			}
            m_buffer[pos]/=float(sfInfo.channels);
		}
	}

    m_loaded=true;

    //close the file
    if(sf_close(sndFile)) {
        throw std::logic_error("Unable to close sample file"+m_name);
    }


    //compute the interpolation coefficients
    double x=0;
    for (int i = 0; i < INTERP_SIZE; i++) {
        x = (double) i / (double) INTERP_SIZE;
        m_interpCoeffs[i][0] = (x * (-0.5 + x * (1 - 0.5 * x)));
        m_interpCoeffs[i][1] = (1.0 + x * x * (1.5 * x - 2.5));
        m_interpCoeffs[i][2] = (x * (0.5 + x * (2.0 - 1.5 * x)));
        m_interpCoeffs[i][3] = (0.5 * x * x * (x - 1.0));
    }

}

void Sample::setSampleRate(int sr)
{
/*
    //if Samplerate different from jack's Samplerate -> reSample
    if(sr != m_sampleRate && m_loaded) {
        m_loaded=false;
        long newFramesCount = (long)((double)m_framesCount / (double) m_sampleRate * (double)sr); 
        double** convs = new double*[m_channelsCount];
        SRC_DATA *datas = new SRC_DATA[m_channelsCount];
        for(int i=0;i<m_channelsCount;++i) {
            convs[i] = new double[newFramesCount];
            datas[i].data_in = m_buffers[i];
            datas[i].data_out = convs[i];
            datas[i].input_frames = m_framesCount;
            datas[i].output_frames = newFramesCount;
            datas[i].src_ratio = (double)sr / (double)m_sampleRate;
            src_simple(datas+i, 0, 1);
        }

        for(int i=0;i<m_channelsCount;++i) {
            delete [] m_buffers[i];    
            m_buffers[i] = new double[datas[i].output_frames_gen];
            for(int j=0;j<datas[i].output_frames_gen;++j) {
                m_buffers[i][j] = convs[i][j];
            }
        }

        m_framesCount = newFramesCount;
        m_sampleRate = sr;

        for(int i=0;i<m_channelsCount;++i) {
            delete [] convs[i];
        }
        delete [] convs;
        delete [] datas;
        
        m_loaded=true;
    }
    */
}

void Sample::unload() {
    //unload the file
    m_buffer.clear();
    m_loaded=false;
}

const double& Sample::getFrame(const long& offset) {
    if(m_loaded) {
        return m_buffer[long(offset)];
    }
    else {
        return m_defaultFrame;
    }
}

const double& Sample::getLinInterpFrame(const double& offset) {
    if(m_loaded) {
        double tmpOffset=fmod(offset, m_framesCount);
        double tmpFlOffset=floor(tmpOffset);
        m_computedFrame =   m_buffer[long(tmpOffset)%m_framesCount]
                            +(tmpOffset-tmpFlOffset)
                                *(m_buffer[(long(tmpFlOffset)+1)]
                                    -m_buffer[long(tmpFlOffset)]);
    }
    else {
        m_computedFrame=0;
    }
    return m_computedFrame;
}


