/***************************************************************************
 *  PiivertFingersFSRMidi.cpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "PiivertFingersFSRMidi.hpp"
#include "../../env/Piivert.hpp"
#include "PiivertDevice.hpp"

using namespace std;

PiivertFingersFSRMidi::PiivertFingersFSRMidi(const int& id, 
                                const std::string& controls, const HAND& hand, 
                                PiivertDevice* piivertDevice): 
                                Mapping(id, hand), 
                                m_piivertDevice(piivertDevice), 
                                m_buttonSensorOffset(20),
                                m_ccSensorOffset(10),
                                m_hitValueThreshold(500), 
                                m_hitTimeThreshold(3000), 
                                m_releaseTimeThreshold(50000), 
                                m_releaseValueThreshold(300) {

    std::string controlsStr = controls;    
    parseFinger(controlsStr, 0); //thumb
    parseFinger(controlsStr, 1); //index
    parseFinger(controlsStr, 2); //middle
    parseFinger(controlsStr, 3); //ring
}

PiivertFingersFSRMidi::~PiivertFingersFSRMidi() {}

void PiivertFingersFSRMidi::parseFinger(std::string& controlsStr, 
                                                const int& fingerCounter) {
    size_t pos = controlsStr.find_first_of(";");
    string sensors = controlsStr.substr(0,pos);
    istringstream is1,is2;
    size_t slashPos = sensors.find_first_of("/");
    if(slashPos==string::npos) { // single sensor -> note
        is1.str(sensors);
        int val1;
        is1>>val1;
        m_fingers[val1]=fingerCounter+m_buttonSensorOffset;
        m_ids.push_back(MappingID(INPUT_MIDI_NOTE,this,val1));
    }
    else { // two sensors -> cc
        is1.str(sensors.substr(0,slashPos));
        is2.str(sensors.substr(slashPos+1));
        int val1,val2;
        is1>>val1;
        is2>>val2;
        m_fingers[val1]=fingerCounter;
        m_ids.push_back(MappingID(INPUT_MIDI_CC,this,val1));
        m_fingers[val2]=fingerCounter+m_ccSensorOffset;
        m_ids.push_back(MappingID(INPUT_MIDI_CC,this,val2));
    }
    m_firstSensorValues.push_back(127);
    m_fingersLastValues.push_back(0);
    m_fingerValues.push_back(vector<float>());
    m_fingersStates.push_back(FINGER_RELEASED);
    m_fingersPressMax.push_back(0);
    timeval curTime;
    gettimeofday(&curTime,NULL);
    m_fingersLastTimes.push_back(curTime);
    m_newSelectedWorms.push_back(-1);
    m_selectedWorms.push_back(-1);
    controlsStr = controlsStr.substr(pos+1);
}

void PiivertFingersFSRMidi::process(InputEvent& event, 
                                    std::vector<Event>& outputAudioEvents, 
                                    std::vector<Event>& outputEnvEvents) {

    if(m_selectedWormsHaveChanged) {
        //check all fingers for pressing state and send stop to corresponding 
        //selected nodes
        if(m_newSelectedWorms[1]!=m_selectedWorms[1] && m_selectedWorms[1]>=0) {
            for(unsigned int f=0; f<m_fingersStates.size(); ++f) {
                if(m_fingersStates[f]==FINGER_PRESSED) {
                    outputAudioEvents.push_back(Event());
                    Event &aEv = outputAudioEvents.back();
                    aEv.type=EVENT_NODE;
                    aEv.parameter=NODE_EXCITE_CONT_RELEASE;
                    aEv.node=m_selectedWorms[1];
                    aEv.finger=1;
                    aEv.values[0]=1;
                    outputAudioEvents.push_back(Event());
                    Event &aEv2 = outputAudioEvents.back();
                    aEv2.type=EVENT_NODE;
                    aEv2.parameter=NODE_EXCITE_CONT_RELEASE;
                    aEv2.finger=2;
                    aEv2.node=m_selectedWorms[1];
                    aEv2.values[0]=1;
                }
            }
        }
        m_selectedWorms = m_newSelectedWorms;
        m_selectedWormsHaveChanged=false;
    }

    int finger = 0;
    //GET FINGER AND VALUE
    if(m_fingers[event.parameter]>=m_buttonSensorOffset) { //note/button sensor
        finger = m_fingers[event.parameter]-m_buttonSensorOffset;
        m_fingerValues[finger].push_back((event.values[0]>0)?4096:0);
    }
    else if(m_fingers[event.parameter]>=m_ccSensorOffset){ //first CC/FSR sensor
        finger = m_fingers[event.parameter]-m_ccSensorOffset;
        m_fingerValues[finger].push_back(
                4096 - ((int(event.values[0])<<7)+m_firstSensorValues[finger]));

    }
    else {    //second CC/FSR sensor
        m_firstSensorValues[m_fingers[event.parameter]]=event.values[0];
    }
}

void PiivertFingersFSRMidi::update(vector<Event>& outputAudioEvents, 
                                        vector<Event>& outputEnvEvents) {
    
    for(int finger=0; finger<4; ++finger) {
        bool gotHit=false;
        double hitValue=0;
        if(m_fingerValues[finger].size()==0) {
            m_fingerValues[finger].push_back(m_fingersLastValues[finger]);
        }
        vector<float>::iterator itVal=m_fingerValues[finger].begin();
        for(; itVal!=m_fingerValues[finger].end(); ++itVal) {
            int resultValue=(*itVal);
            //TEST PRESSING/HIT
            if(m_fingersStates[finger] == FINGER_RELEASED) {
                if(resultValue>m_releaseValueThreshold) {
                    m_fingersStates[finger]=FINGER_PRESSING;
                    gettimeofday(&m_fingersLastTimes[finger], NULL);
                    m_fingersPressMax[finger]=resultValue;
                    if(finger==0) { //toggle grab
                        m_grabToggle=!m_grabToggle;
                        outputEnvEvents.push_back(Event());
                        Event& eEv = outputEnvEvents.back();
                        eEv.type=EVENT_PIIVERT;
                        eEv.parameter=PIIVERT_GRAB;
                        eEv.piivert=m_id;
                        eEv.hand=m_hand;
                        eEv.finger=finger;
                        eEv.values[0]=m_grabToggle;
                        if(m_loopState==LOOP_RECORDING 
                                && m_selectedWorms[1]>=0) {
                            startRecordingWorm(m_selectedWorms[1], 
                                outputAudioEvents, outputEnvEvents);
                        }
                    }
                }
                else {
                    resultValue=0;
                }
            }
            if(m_fingersStates[finger] == FINGER_PRESSING) {
                timeval curTime;
                gettimeofday(&curTime,NULL);
                timeval timeDiff;
                timersub(&curTime,&m_fingersLastTimes[finger],&timeDiff);
                //if time is over for a hit
                if(timeDiff.tv_sec>=0 && timeDiff.tv_usec>m_hitTimeThreshold) {
                    hitValue = double(m_fingersPressMax[finger])/4096.0;
                    //if value high enough, hit
                    if(m_fingersPressMax[finger]>m_hitValueThreshold) {
                        gotHit=true;
                        
                    } //else press
                    else {
                        if(m_selectedWorms[1]>=0 && (finger==1 || finger==2)) {
                            outputAudioEvents.push_back(Event());
                            Event &aEv = outputAudioEvents.back();
                            aEv.type=EVENT_NODE;
                            aEv.parameter=NODE_EXCITE_CONT;
                            aEv.node=m_selectedWorms[1];
                            aEv.finger=finger;
                            aEv.values[0]=hitValue;
                        }
                    }
                    outputEnvEvents.push_back(Event());
                    Event &eEv = outputEnvEvents.back();
                    eEv.type=EVENT_PIIVERT;
                    eEv.parameter=PIIVERT_PRESS;
                    eEv.piivert=m_id;
                    eEv.hand=m_hand;
                    eEv.finger=finger;
                    eEv.values[0]=hitValue;
                    m_fingersStates[finger]=FINGER_PRESSED;
                    gettimeofday(&m_fingersLastTimes[finger], NULL);
                }
                else if(resultValue>m_fingersPressMax[finger]) {
                    m_fingersPressMax[finger]=resultValue;
                }
            }
            if(m_fingersStates[finger] == FINGER_PRESSED) {
                timeval curTime;
                gettimeofday(&curTime,NULL);
                timeval timeDiff;
                timersub(&curTime,&m_fingersLastTimes[finger],&timeDiff);
                if(timeDiff.tv_sec>=0 
                        && timeDiff.tv_usec>m_releaseTimeThreshold
                        && resultValue<m_releaseValueThreshold) {
                    hitValue = 0; 
                    resultValue=0;
                    m_fingersStates[finger]=FINGER_RELEASED;
                    if(m_selectedWorms[1]>=0 && (finger==1 || finger==2)) {
                        outputAudioEvents.push_back(Event());
                        Event &aEv = outputAudioEvents.back();
                        aEv.type=EVENT_NODE;
                        aEv.parameter=NODE_EXCITE_CONT_RELEASE;
                        aEv.node=m_selectedWorms[1];
                        aEv.finger=finger;
                        aEv.values[0]=hitValue;
                    }
                }
                else {
                    hitValue = double(resultValue)/4096.0;
                    if(m_selectedWorms[1]>=0 && (finger==1 || finger==2)) {
                        outputAudioEvents.push_back(Event());
                        Event &aEv = outputAudioEvents.back();
                        aEv.type=EVENT_NODE;
                        aEv.parameter=NODE_EXCITE_CONT;
                        aEv.node=m_selectedWorms[1];
                        aEv.finger=finger;
                        aEv.values[0]=hitValue;
                    }
                }
                outputEnvEvents.push_back(Event());
                Event &eEv = outputEnvEvents.back();
                eEv.type=EVENT_PIIVERT;
                eEv.parameter=PIIVERT_PRESS;
                eEv.piivert=m_id;
                eEv.hand=m_hand;
                eEv.finger=finger;
                eEv.values[0]=hitValue;
            }
            m_fingersLastValues[finger]=resultValue;
        }



        //PROCESS GESTURES
        if(gotHit) {
            int gesture = (m_hand+1)*10 + finger;
            m_piivertDevice->processGesture(gesture);

            //according to gesture, send audio and env events
            switch(gesture) {
                case 1321:
                case 1322:
                case 2312:
                case 2311:{ //two hands flam little to middle 
                    switch(m_loopState) {
                        case LOOP_NONE: {
                            m_piivertDevice->startRecordingBothHands(
                                outputAudioEvents, outputEnvEvents);
                            outputEnvEvents.push_back(Event());
                            Event &eEv1 = outputEnvEvents.back();
                            eEv1.type=EVENT_PIIVERT;
                            eEv1.parameter=PIIVERT_RECORD_TWO_HANDS;
                            eEv1.piivert=m_id;
                            eEv1.hand=m_hand;
                            eEv1.values[0]=1;
                        }break;
                        case LOOP_RECORDING: {
                            m_piivertDevice->stopRecordingBothHands(
                                outputAudioEvents, outputEnvEvents);
                            m_piivertDevice->eraseRecordingBothHands();
                            outputEnvEvents.push_back(Event());
                            Event &eEv1 = outputEnvEvents.back();
                            eEv1.type=EVENT_PIIVERT;
                            eEv1.parameter=PIIVERT_RECORD_TWO_HANDS;
                            eEv1.piivert=m_id;
                            eEv1.hand=m_hand;
                            eEv1.values[0]=0;
                            m_piivertDevice->releaseBothHands(
                                outputAudioEvents, outputEnvEvents);
                        }break;
                        default:break;
                    }
                    if(m_selectedWorms[1]>=0) {
                        //excite
                        outputAudioEvents.push_back(Event());
                        Event &aEv = outputAudioEvents.back();
                        aEv.type=EVENT_NODE;
                        aEv.parameter=NODE_EXCITE_INST;
                        aEv.node=m_selectedWorms[1];
                        aEv.finger=finger;
                        aEv.values[0]=hitValue;
                    }
                    outputEnvEvents.push_back(Event());
                    Event &eEv = outputEnvEvents.back();
                    eEv.type=EVENT_PIIVERT;
                    eEv.parameter=PIIVERT_HIT;
                    eEv.piivert=m_id;
                    eEv.hand=m_hand;
                    eEv.finger=finger;
                    eEv.values[0]=hitValue;
                }break;
                case 1123:
                case 1223:
                case 2113:
                case 2213:{ //two hands flam middle to little  
                    if(m_loopState==LOOP_RECORDING) {
                        m_piivertDevice->cancelRecordingBothHands(
                            outputAudioEvents, outputEnvEvents);
                        m_piivertDevice->eraseRecordingBothHands();
                    }
                    outputEnvEvents.push_back(Event());
                    Event &eEv1 = outputEnvEvents.back();
                    eEv1.type=EVENT_PIIVERT;
                    eEv1.parameter=PIIVERT_RECORD_TWO_HANDS;
                    eEv1.piivert=m_id;
                    eEv1.hand=m_hand;
                    eEv1.values[0]=0;
                }break;
                case 1311:
                case 1312:
                case 2322:
                case 2321: { //one hand flam 
                    switch(m_loopState) {
                        case LOOP_RECORDING: {
                            timeval recordingStopTime;
                            gettimeofday(&recordingStopTime, NULL);
                            stopRecordingWorms(recordingStopTime, 
                                        outputAudioEvents, outputEnvEvents);
                            outputEnvEvents.push_back(Event());
                            Event &eEv1 = outputEnvEvents.back();
                            eEv1.type=EVENT_PIIVERT;
                            eEv1.parameter=PIIVERT_RECORD_ONE_HAND;
                            eEv1.piivert=m_id;
                            eEv1.hand=m_hand;
                            eEv1.values[0]=0;

                            //release
                            release(outputAudioEvents, outputEnvEvents);
                        }break;
                        case LOOP_NONE: {
                            timeval recordingStartTime;
                            gettimeofday(&recordingStartTime, NULL);
                            startRecordingSelectedWorms(recordingStartTime, 
                                outputAudioEvents, outputEnvEvents);
                            outputEnvEvents.push_back(Event());
                            Event &eEv1 = outputEnvEvents.back();
                            eEv1.type=EVENT_PIIVERT;
                            eEv1.parameter=PIIVERT_RECORD_ONE_HAND;
                            eEv1.piivert=m_id;
                            eEv1.hand=m_hand;
                            eEv1.values[0]=1;
                        }break;
                        default:break;
                    }
                    if(m_selectedWorms[1]>=0) {
                        outputAudioEvents.push_back(Event());
                        Event &aEv = outputAudioEvents.back();
                        aEv.type=EVENT_NODE;
                        aEv.parameter=NODE_EXCITE_INST;
                        aEv.node=m_selectedWorms[1];
                        aEv.finger=finger;
                        aEv.values[0]=hitValue;
                    }
                    outputEnvEvents.push_back(Event());
                    Event &eEv = outputEnvEvents.back();
                    eEv.type=EVENT_PIIVERT;
                    eEv.parameter=PIIVERT_HIT;
                    eEv.piivert=m_id;
                    eEv.hand=m_hand;
                    eEv.finger=finger;
                    eEv.values[0]=hitValue;
                }break;
                case 1113:
                case 1213:
                case 2123:
                case 2223:{ //one hand reverse flam 
                    if(m_loopState==LOOP_RECORDING) {
                        cancelRecordingWorms(outputAudioEvents,outputEnvEvents);
                        eraseRecordingWorms();
                    }
                    else {
                        if(m_selectedWorms[1]>=0) {
                            outputEnvEvents.push_back(Event());
                            Event &eEv = outputEnvEvents.back();
                            eEv.type=EVENT_TREE;
                            eEv.parameter=TREE_EXTRACT;
                            eEv.node=m_selectedWorms[1];
                            eEv.values[0]=m_selectedWorms[1];
                            eEv.time=m_recordingStartTime;    

                            outputAudioEvents.push_back(Event());
                            Event &aEv = outputAudioEvents.back();
                            aEv.type=EVENT_TREE;
                            aEv.parameter=TREE_EXTRACT;
                            aEv.node=m_selectedWorms[1];
                            aEv.values[0]=m_selectedWorms[1];
                            aEv.time=m_recordingStartTime;    
                        }
                    }
                    outputEnvEvents.push_back(Event());
                    Event &eEv1 = outputEnvEvents.back();
                    eEv1.type=EVENT_PIIVERT;
                    eEv1.parameter=PIIVERT_RECORD_ONE_HAND;
                    eEv1.piivert=m_id;
                    eEv1.hand=m_hand;
                    eEv1.values[0]=0;
                }break;
                case 2122:
                case 2221:
                case 21:
                case 22:
                case 1112:
                case 1211:
                case 11:
                case 12: 
                case 2112:
                case 2211:
                case 1122:
                case 1221:{ //one finger hit
                    if(m_selectedWorms[1]>=0) {
                        //start recording
                        if(m_loopState==LOOP_RECORDING) {
                            startRecordingWorm(m_selectedWorms[1], 
                                outputAudioEvents, outputEnvEvents);
                        }
                        //excite
                        outputAudioEvents.push_back(Event());
                        Event &aEv = outputAudioEvents.back();
                        aEv.type=EVENT_NODE;
                        aEv.parameter=NODE_EXCITE_INST;
                        aEv.finger=finger;
                        aEv.node=m_selectedWorms[1];
                        aEv.values[0]=hitValue;
                    }
                    outputEnvEvents.push_back(Event());
                    Event &eEv = outputEnvEvents.back();
                    eEv.type=EVENT_PIIVERT;
                    eEv.parameter=PIIVERT_HIT;
                    eEv.piivert=m_id;
                    eEv.hand=m_hand;
                    eEv.finger=finger;
                    eEv.values[0]=hitValue;
                }break;
                case 1013:
                case 2023:{ //thumb little flam 
                    if(m_selectedWorms[1]>=0) {
                        outputAudioEvents.push_back(Event());
                        Event& newEvent = outputAudioEvents.back();
                        newEvent.type = EVENT_TREE;
                        newEvent.values[0] = m_selectedWorms[1];
                        newEvent.parameter = TREE_DUPLICATE;
                    }
                }break;
                case 13:
                case 23: { //last finger -> stop
                    if(m_selectedWorms[1]>=0) {
                        if(m_loopState==LOOP_RECORDING) {
                            startRecordingWorm(m_selectedWorms[1], 
                                outputAudioEvents, outputEnvEvents);
                        }

                        outputAudioEvents.push_back(Event());
                        Event &aEv = outputAudioEvents.back();
                        aEv.type=EVENT_NODE;
                        aEv.parameter=NODE_STOP;
                        aEv.node=m_selectedWorms[1];
                        aEv.values[0]=1;

                    }
                }break;
                default:break;
            }
        }
        //clear recorded finger values
        m_fingerValues[finger].clear();
    }
}

