/***************************************************************************
 *  Camera.cpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net 
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "Camera.hpp"
#include "Piivert.hpp"
#include "../devices/head/HeadTrackingVRPN.hpp"
#include "../devices/head/HeadTrackingVRPNGameTrak.hpp"
#include "../devices/head/HeadTrackingVRPNHydra.hpp"
#include "../devices/head/HeadTrackingOscPos.hpp"
#include "Reconstruct.hpp"
#include "Zone.hpp"
#include "Reconstruct3D.hpp"
#include "Room.hpp"
#include "Passage.hpp"
#include "Tunnel.hpp"
#include "Sign.hpp"
#include "EnvManager.hpp"

using namespace std;

Camera::Camera(int id, OSG::NodeRecPtr root, 
                OSG::PassiveWindowRecPtr window):
                                        m_id(id),
                                        m_reconstruct(NULL), m_name("camera"), 
                                        m_stereo(STEREO_NONE), m_xOverlap(0), 
                                        m_eyeSeparation(0.03), m_nbDisplays(1),
                                        m_adress("127.0.0.1"), m_port(7010),  
                                        m_fakeHeadMovement(0), 
                                        m_fakeHeadMovementStep(0.0002), 
                                        m_fakeHeadMovementsActivated(false),
                                        m_reconstHTActivated(false)
                                        {


    for(int param=0; param<CAMERA_PARAMETERS_NB; ++param) {
        m_parametersChanged[param]=false;
    }

    {
        m_cam = OSG::PerspectiveCamera::create();

        m_camBeacon = OSG::Node::create();
        m_camXform = OSG::ComponentTransform::create();
        m_camXform->setTranslation(m_cameraBasePosition);
        OSG::Quaternion &rotCam = m_camXform->editRotation();
        rotCam.setValue(m_cameraBaseRotation[0]/180.0*M_PI, 
                        m_cameraBaseRotation[1]/180.0*M_PI, 
                        m_cameraBaseRotation[2]/180.0*M_PI);
        m_camBeacon->setCore(m_camXform);
        m_cam->setBeacon(m_camBeacon);
        m_cam->setFov(OSG::osgDegree2Rad(90));
        m_cam->setNear(0.01);
        m_cam->setFar(300);

        //HEAD
        m_head = OSG::Node::create();
        m_headXform = OSG::ComponentTransform::create();
        m_head->setCore(m_headXform);
        m_camBeacon->addChild(m_head);
        m_headBasePosition=OSG::Vec3f(0,1.6,2);
        m_headXform->setTranslation(m_headBasePosition);

        //add the external light
        OSG::NodeRecPtr lightBeacon = OSG::Node::create();
        m_lightTrans = OSG::ComponentTransform::create();
        lightBeacon->setCore(m_lightTrans);    
        m_externalLight= OSG::PointLight::create();
        m_externalLight->setDiffuse(OSG::Color4f(1, 1, 1, 1));
        m_externalLight->setAmbient(OSG::Color4f(0, 0, 0, 1));
        m_externalLight->setConstantAttenuation(0.01);
        m_externalLight->setLinearAttenuation(0.5);
        m_externalLight->setBeacon(lightBeacon);
        m_light = OSG::Node::create();
        m_light->setCore(m_externalLight);
        m_head->addChild(lightBeacon);

        //add the internal light
        OSG::NodeRecPtr internLightBeacon = OSG::Node::create();
        m_internalLightTrans = OSG::ComponentTransform::create();
        internLightBeacon->setCore(m_lightTrans);    
        m_internalLight = OSG::PointLight::create();
        m_internalLight->setDiffuse(OSG::Color4f(1, 1, 1, 1));
        m_internalLight->setAmbient(OSG::Color4f(0, 0, 0, 1));
        m_internalLight->setConstantAttenuation(0.01);
        m_internalLight->setLinearAttenuation(0.5);
        m_internalLight->setBeacon(lightBeacon);
        m_internalLightNode = OSG::Node::create();
        m_internalLightNode->setCore(m_internalLight);
        m_head->addChild(internLightBeacon);
        m_camBeacon->addChild(m_internalLightNode);

        //decorators
        m_leftCameraDecorator=OSG::ProjectionCameraDecorator::create();
        m_rightCameraDecorator=OSG::ProjectionCameraDecorator::create();
        m_leftCameraDecorator->setLeftEye(true);
        m_leftCameraDecorator->setEyeSeparation(m_eyeSeparation);
        m_leftCameraDecorator->setDecoratee(m_cam);
        m_leftCameraDecorator->setUser(m_head);
        m_rightCameraDecorator->setLeftEye(false);
        m_rightCameraDecorator->setEyeSeparation(m_eyeSeparation);
        m_rightCameraDecorator->setDecoratee(m_cam);
        m_rightCameraDecorator->setUser(m_head);

        //default screen coordinates
        vector<float> coords(3,0);
        coords[0]=-1.5;
        setScreenCorners(0, coords);
        coords[0]=1.5;
        setScreenCorners(1, coords);
        coords[1]=2;
        setScreenCorners(2, coords);
        setScreenCorners(3, coords);
        coords[0]=-1.5;

        //Viewports
        m_multiWindow = OSG::MultiDisplayWindow::create();

        m_leftViewportStereo = OSG::StereoBufferViewport::create();
        m_rightViewportStereo = OSG::StereoBufferViewport::create();
        m_leftViewportStereo->setCamera(m_leftCameraDecorator);
        m_leftViewportStereo->setRoot(root);
        m_leftViewportStereo->setSize(0,0,1,1);
        m_rightViewportStereo->setCamera(m_rightCameraDecorator);
        m_rightViewportStereo->setRoot(root);
        m_rightViewportStereo->setSize(0,0,1,1);
        m_leftViewportColor = OSG::ColorBufferViewport::create();
        m_rightViewportColor = OSG::ColorBufferViewport::create();
        m_leftViewportColor->setCamera(m_leftCameraDecorator);
        m_leftViewportColor->setRoot(root);
        m_leftViewportColor->setSize(0,0,1,1);
        m_rightViewportColor->setCamera(m_rightCameraDecorator);
        m_rightViewportColor->setRoot(root);
        m_rightViewportColor->setSize(0,0,1,1);

        //render action
        m_renderAction = OSG::RenderAction::create();

        //cluster
        m_multiWindow->setClientWindow(window);
        m_multiWindow->addPort(m_leftViewport);
        setAdress("local");
        //stereo default to none
        setStereo("none");
        //default background
        setBackgroundVideo(-1);

        OSG::commitChanges();
    }
}

Camera::~Camera() {}

void Camera::parse(xmlNodePtr node) {
    char* value=NULL;
    value = (char*)xmlGetProp(node,(xmlChar*)"name");
    if(value!=NULL) {
        setName(std::string(value));
    }

    value=NULL;
    value = (char*)xmlGetProp(node,(xmlChar*)"adress");
    if(value!=NULL) {
        setAdress(std::string(value));
    }

    value=NULL;
    value = (char*)xmlGetProp(node,(xmlChar*)"port");
    if(value!=NULL) {
        setPort(atoi(value));
    }
        
    value=NULL;
    value = (char*)xmlGetProp(node,(xmlChar*)"stereo");
    if(value!=NULL) {
        std::string valueStr=std::string(value);
        
        setStereo(valueStr);
    }

    value=NULL;
    value = (char*)xmlGetProp(node,(xmlChar*)"screen_corners_bl_br_tr_tl");
    if(value!=NULL) {
        std::string screenStr=std::string(value);
        size_t pos;

        for(unsigned int i=0;i<4;++i) {
            std::vector<float> coords(3,0);
            std::istringstream is,is2,is3;
            pos = screenStr.find_first_of("/");
            is.str(screenStr.substr(0,pos));
            screenStr = screenStr.substr(pos+1);
            is>>coords[0];
            pos = screenStr.find_first_of("/");
            is2.str(screenStr.substr(0,pos));
            screenStr = screenStr.substr(pos+1);
            is2>>coords[1];
            pos = screenStr.find_first_of(";");
            is3.str(screenStr.substr(0,pos));
            screenStr = screenStr.substr(pos+1);
            is3>>coords[2];
            setScreenCorners(i,coords);
        }
    }

    //head-tracking
    value=NULL;
    value = (char*)xmlGetProp(node,(xmlChar*)"head_tracking");
    if(value!=NULL) {
        std::string headStr = value;
        size_t pos = headStr.find_first_of(';');    
        std::string type = headStr.substr(0, pos);
        headStr = headStr.substr(pos+1);
        setHeadTracking(type, headStr);
    }
    //get base translation of head if any
    value=NULL;
    value = (char*)xmlGetProp(node,(xmlChar*)"head_initial_position");
    if(value!=NULL) {
        std::string valStr = value;
        std::vector<float> coords(3,0);
        parseCoords(valStr,coords);
        setHeadInitialPosition(coords);
    }

    //get camera position if any
    value=NULL;
    value = (char*)xmlGetProp(node,(xmlChar*)"camera_initial_position");
    if(value!=NULL) {
        std::string valStr = value;
        std::vector<float> coords(3,0);
        parseCoords(valStr,coords);
        setCameraBasePosition(coords);
    }

    //get camera rotation if any
    value=NULL;
    value = (char*)xmlGetProp(node,(xmlChar*)"camera_initial_rotation");
    if(value!=NULL) {
        std::string valStr = value;
        std::vector<float> coords(3,0);
        parseCoords(valStr,coords);
        setCameraBaseRotation(coords);
    }

    //get reconstructed background
    value=NULL;
    value = (char*)xmlGetProp(node,(xmlChar*)"background_video");
    if(value!=NULL) {
        setBackgroundVideo(atof(value));
    }

    //get light attenuations
    value=NULL;
    value = (char*)xmlGetProp(node,(xmlChar*)"light_attenuation");
    if(value!=NULL) {
        std::string valStr = value;
        std::vector<float> coords(3,0);
        parseCoords(valStr,coords);
        m_externalLight->setConstantAttenuation(coords[0]);
        m_internalLight->setConstantAttenuation(coords[0]);
        m_externalLight->setLinearAttenuation(coords[1]);
        m_internalLight->setLinearAttenuation(coords[1]);
        m_externalLight->setQuadraticAttenuation(coords[2]);
        m_internalLight->setQuadraticAttenuation(coords[2]);
    }
}

void Camera::parseCoords(std::string &strCoords, std::vector<float> &coords) {
    istringstream is, is2, is3;
    size_t pos = strCoords.find_first_of('/');    
    if(pos!=strCoords.npos) {
        is.str(strCoords.substr(0,pos));
        is>>coords[0];
    }
    else {
        coords[0]=0;
    }
    strCoords = strCoords.substr(pos+1);
    pos = strCoords.find_first_of('/');    
    if(pos!=strCoords.npos) {
        is2.str(strCoords.substr(0,pos));
        is2>>coords[1];
    }
    else {
        coords[1]=0;
    }
    strCoords = strCoords.substr(pos+1);
    pos = strCoords.find_first_of(';');    
    if(pos!=strCoords.npos) {
        is3.str(strCoords.substr(0,pos));
        is3>>coords[2];
    }
    else {
        coords[2]=0;
    }
}



void Camera::reset() {
    //remove all mappings ...
    vector<Mapping*>::iterator itMap;
    for(itMap=m_headMappings.begin(); itMap!=m_headMappings.end(); ++itMap) {
        delete (*itMap);
    }
}

void Camera::getMappingsIDs(std::vector<MappingID>& mappingsIDs){
    vector<Mapping*>::iterator itMap;
    for(itMap=m_headMappings.begin(); itMap!=m_headMappings.end(); ++itMap) {
        (*itMap)->getMappingsIDs(mappingsIDs);
    }
}

void Camera::setParameter(const CAMERA_PARAMETER& param, double* values, 
                                unsigned int nbValues) {
    m_parametersValues[param].assign(values, values+nbValues);
    m_parametersChanged[param]=true;
}

void Camera::setName(const std::string& name) {
    m_name=name;
    resetDisplay();
}

void Camera::setBackgroundVideo(float video) {
    if(video<0) { //no video, set gradient background
        OSG::GradientBackgroundRecPtr back = OSG::GradientBackground::create();
        back->addLine(OSG::Color3f(0,0,0),0);
        back->addLine(OSG::Color3f(0,0,0),1);
        m_background=back;

        if(m_reconstruct) {
            delete m_reconstruct;
            m_reconstruct=NULL;
        }
    }
    else {//start video background with the camera number
        m_reconstruct = new Reconstruct(video);
        m_reconstruct->load();
        m_background=m_reconstruct->getBackground();
    }
    m_leftViewportStereo->setBackground(m_background);
    m_rightViewportStereo->setBackground(m_background);
    m_leftViewportColor->setBackground(m_background);
    m_rightViewportColor->setBackground(m_background);
}

void Camera::setAdress(const std::string& adress) {
    m_adress=std::string(adress);
    if(m_adress.compare("multicast")==0) {
        m_multiWindow->setConnectionType("Multicast");
    }
    else {
        m_multiWindow->setConnectionType("StreamSock");
        m_multiWindow->setServiceAddress("127.0.0.1");
    }
}

void Camera::setPort(const int& port) {
    m_port=port;
    m_multiWindow->setServicePort(m_port);
}

void Camera::setStereo(const std::string& stereo) {
    if(stereo.compare("none")==0) {
        m_leftViewport = m_leftViewportColor;
        m_rightViewport = m_rightViewportColor;
        m_stereo=STEREO_NONE;
        m_leftViewport->setSize(0,0,1,1);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_leftViewport)->setRed(true);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_leftViewport)->setGreen(true);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_leftViewport)->setBlue(true);
    }
    else if(stereo.compare("anaglyph")==0) {
        m_leftViewport = m_leftViewportColor;
        m_rightViewport = m_rightViewportColor;
        m_stereo=STEREO_ANAGLYPH;
        bool redcyan = false;
        m_leftViewport->setSize(0,0,1,1);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_leftViewport)->setRed(!redcyan);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_leftViewport)->setGreen(redcyan);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_leftViewport)->setBlue(redcyan);
        m_rightViewport->setSize(0,0,1,1);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_rightViewport)->setRed(redcyan);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_rightViewport)->setGreen(!redcyan);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_rightViewport)->setBlue(!redcyan);
    }
    else if(stereo.compare("windowed")==0) {
        m_leftViewport = m_leftViewportColor;
        m_rightViewport = m_rightViewportColor;
        m_stereo=STEREO_WINDOWED;
        m_leftViewport->setSize(0.,0.5,1.,1.);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_leftViewport)->setRed(true);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_leftViewport)->setGreen(true);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_leftViewport)->setBlue(true);
        m_rightViewport->setSize(0.,0.,1.,0.5);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_rightViewport)->setRed(true);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_rightViewport)->setGreen(true);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_rightViewport)->setBlue(true);
    }
    else if(stereo.compare("sidebyside")==0) {
        m_leftViewport = m_leftViewportColor;
        m_rightViewport = m_rightViewportColor;
        m_stereo=STEREO_SIDEBYSIDE;
        m_leftViewport->setSize(0.,0.,0.5,1.);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_leftViewport)->setRed(true);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_leftViewport)->setGreen(true);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_leftViewport)->setBlue(true);
        m_rightViewport->setSize(0.5,0.,1.,1.);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_rightViewport)->setRed(true);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_rightViewport)->setGreen(true);
        OSG::static_pointer_cast<OSG::ColorBufferViewport>
            (m_rightViewport)->setBlue(true);
    }
    else if(stereo.compare("quad")==0) {
        m_leftViewport = m_leftViewportStereo;
        m_rightViewport = m_rightViewportStereo;
        m_stereo=STEREO_QUAD;
        m_leftViewport->setSize(0,0,1,1);
        OSG::static_pointer_cast<OSG::StereoBufferViewport>
            (m_leftViewport)->setLeftBuffer(true);
        OSG::static_pointer_cast<OSG::StereoBufferViewport>
            (m_leftViewport)->setRightBuffer(false);
        m_rightViewport->setSize(0,0,1,1);
        OSG::static_pointer_cast<OSG::StereoBufferViewport>
            (m_rightViewport)->setLeftBuffer(false);
        OSG::static_pointer_cast<OSG::StereoBufferViewport>
            (m_rightViewport)->setRightBuffer(true);
    }


    resetDisplay();
}

void Camera::setScreenCorners(const unsigned int& corner, 
                    const std::vector<float>& coords) {
    m_screenCorners[corner]=OSG::Pnt3f(coords[0],coords[1],coords[2]);
    m_leftCameraDecorator->editMFSurface()->clear();
    for(unsigned int i=0;i<4;++i) {
        m_leftCameraDecorator->editMFSurface()->push_back(m_screenCorners[i]);
    }
    m_rightCameraDecorator->editMFSurface()->clear();
    for(unsigned int i=0;i<4;++i) {
        m_rightCameraDecorator->editMFSurface()->push_back(m_screenCorners[i]);
    }
}

void Camera::setHeadTracking(const std::string& type, 
                                        const std::string& params) {
    std::string headStr=params;
    if(type.compare("fake")==0) { //fake head-tracking
        m_fakeHeadMovementsActivated=true;
    }
    else if(type.compare("kinect")==0) { //KINECT
        m_reconstHTActivated=true;
    }    
    else if(type.compare("vrpn")==0) { //VRPN
        if(m_headMappings.size()>0) {
            delete m_headMappings[0];
            m_headMappings.clear();
        }
        m_headMappings.push_back(new HeadTrackingVRPN(m_id, headStr));
    }    
    else if(type.compare("vrpn_gametrak")==0) { //VRPN
        if(m_headMappings.size()>0) {
            delete m_headMappings[0];
            m_headMappings.clear();
        }
        m_headMappings.push_back(new HeadTrackingVRPNGameTrak(m_id, headStr));
    }    
    else if(type.compare("vrpn_hydra")==0) { //VRPN
        if(m_headMappings.size()>0) {
            delete m_headMappings[0];
            m_headMappings.clear();
        }
        m_headMappings.push_back(new HeadTrackingVRPNHydra(m_id, headStr));
    }    
    else if(type.compare("osc")==0) { //OSC 
        size_t pos=headStr.find_first_of(";");
        while(pos!=headStr.npos) {
            std::string tracking = headStr.substr(0, pos);
            headStr = headStr.substr(pos+1);
            pos=headStr.find_first_of(";");
            if(tracking.compare("pos")==0) {
                if(m_headMappings.size()>0) {
                    delete m_headMappings[0];
                    m_headMappings.clear();
                }
                m_headMappings.push_back(
                    new HeadTrackingOscPos(m_id, headStr.substr(0, pos)));
            }
            headStr = headStr.substr(pos+1);
            pos=headStr.find_first_of(";");
        }
    }    
}

void Camera::setHeadInitialPosition(const std::vector<float>& coords) {
    m_headBasePosition[0]=coords[0];
    m_headBasePosition[1]=coords[1];
    m_headBasePosition[2]=coords[2];
    m_headXform->setTranslation(m_headBasePosition);
}

void Camera::setCameraBasePosition(const std::vector<float>& coords) {
    m_cameraBasePosition[0]=coords[0];
    m_cameraBasePosition[1]=coords[1];
    m_cameraBasePosition[2]=coords[2];
    m_camXform->setTranslation(m_cameraBasePosition);
}

void Camera::setCameraBaseRotation(const std::vector<float>& coords) {
    m_cameraBaseRotation[0]=coords[0];
    m_cameraBaseRotation[1]=coords[1];
    m_cameraBaseRotation[2]=coords[2];
    OSG::Quaternion &rotCam = m_camXform->editRotation();
    rotCam.setValue(m_cameraBaseRotation[0]/180.0*M_PI, 
                    m_cameraBaseRotation[1]/180.0*M_PI, 
                    m_cameraBaseRotation[2]/180.0*M_PI);
}

void Camera::resetDisplay() {
    int dis = (m_stereo==STEREO_WINDOWED)?m_nbDisplays*2:m_nbDisplays;
    m_multiWindow->editMFServers()->clear();
    for(int i=1;i<=dis;++i) {
        std::ostringstream oss;
        oss<<i;
        m_multiWindow->editMFServers()->push_back(
            (m_name+"_"+oss.str()).c_str());

        cout<<"Please start display for camera "
            <<m_name<<" using command: drile-display "
            <<m_name<<"_"<<oss.str()<<" "<<m_port<<endl;
    }
    m_multiWindow->setHServers(m_nbDisplays);
    int serv = (m_stereo==STEREO_WINDOWED)?2:1;
    m_multiWindow->setVServers(serv);
    m_multiWindow->setXOverlap(m_xOverlap);
    m_multiWindow->clearPorts();
    m_multiWindow->addPort(m_leftViewport);
    if(m_stereo!=STEREO_NONE) {
        m_multiWindow->addPort(m_rightViewport);
    }
}


#ifdef WITH_FREENECT
void Camera::addReconstruction(xmlNodePtr recNode) {
    Reconstruct3D* reconst = new Reconstruct3D();
    reconst->parse(recNode);
    reconst->load();
    m_reconstructs.push_back(reconst);
    m_camBeacon->addChild(reconst->getNode());
}
#endif

void Camera::initWindow() {
    m_multiWindow->init();
}

void Camera::connect(std::string adress, int port) {
}

void Camera::update() {
    //update parameters
    for(int param=0; param<CAMERA_PARAMETERS_NB; ++param) {
        if(m_parametersChanged[param]) {
            switch(CAMERA_PARAMETER(param)) {
                case CAMERA_HEAD_POSQUAT: {
                    {
                        OSG::Vec3f &pos = m_headXform->editTranslation();
                        pos[0]=m_parametersValues[param][0];
                        pos[1]=m_parametersValues[param][1];
                        pos[2]=m_parametersValues[param][2];
                    }
                }break;
                case CAMERA_HEAD_POS: {
                    {
                        OSG::Vec3f &pos = m_headXform->editTranslation();
                        pos[0]=m_parametersValues[param][0];
                        pos[1]=m_parametersValues[param][1];
                        pos[2]=m_parametersValues[param][2];
                    }
                }break;
                default:break;
            }
            m_parametersChanged[param]=false;
        }
    }

    //fake head movements to increase immersion when in mono
    if(m_fakeHeadMovementsActivated) {
        m_fakeHeadMovement+=m_fakeHeadMovementStep;
        OSG::Vec3f &pos = m_headXform->editTranslation();
        pos[0]=m_headBasePosition[0]+m_fakeHeadMovement;
        pos[1]=m_headBasePosition[1]+m_fakeHeadMovement;
        pos[2]=m_headBasePosition[2]+m_fakeHeadMovement;
        if(fabs(m_fakeHeadMovement)>0.002) {
            m_fakeHeadMovementStep*=-1;
        }
    }


    #ifdef WITH_FREENECT
    //update reconstruct if needed
    vector<Reconstruct3D*>::iterator it = m_reconstructs.begin();
    for(;it!=m_reconstructs.end();it++) {
        (*it)->update();
        if(m_reconstHTActivated) {
            OSG::Vec3f &pos = m_headXform->editTranslation();
            OSG::Vec3f &newpos = (*it)->getHeadPosition();
            pos[0]=newpos[0];
            pos[1]=newpos[1];
            pos[2]=newpos[2];
        }
    }
    #endif

    //update video background
    if(m_reconstruct) {
        m_reconstruct->update();
    }

    //render
    m_multiWindow->render(m_renderAction);
}

void Camera::resize(int w,int h) {
    m_multiWindow->resize(w, h);
}


