/***************************************************************************
 *  Camera.hpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net 
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef _CAMERA_H
#define _CAMERA_H

#include <iostream>
#include <cstring>
#include <sstream>
#include <cmath>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <vector>
#include <map>

#include <OpenSG/OSGGLUT.h>
#include <OpenSG/OSGConfig.h>
#include <OpenSG/OSGSimpleGeometry.h>
#include <OpenSG/OSGGLUTWindow.h>
#include <OpenSG/OSGSimpleSceneManager.h>
#include <OpenSG/OSGSceneFileHandler.h>
#include <OpenSG/OSGBaseFunctions.h>
#include <OpenSG/OSGTransform.h>
#include <OpenSG/OSGComponentTransform.h>
#include <OpenSG/OSGPassiveWindow.h>
#include <OpenSG/OSGTriangleIterator.h>
#include <OpenSG/OSGMaterialGroup.h>
#include <OpenSG/OSGShearedStereoCameraDecorator.h>
#include <OpenSG/OSGProjectionCameraDecorator.h>
#include <OpenSG/OSGPerspectiveCamera.h>
#include <OpenSG/OSGMultiDisplayWindow.h>
#include <OpenSG/OSGImage.h>
#include <OpenSG/OSGImageFileHandler.h>
#include <OpenSG/OSGMaterialGroup.h>
#include <OpenSG/OSGThreadManager.h>
#include <OpenSG/OSGColorBufferViewport.h>
#include <OpenSG/OSGStereoBufferViewport.h>
#include <OpenSG/OSGPointChunk.h>
#include <OpenSG/OSGTypedGeoIntegralProperty.h>
#include <OpenSG/OSGTypedGeoVectorProperty.h>
#include <OpenSG/OSGSimpleStage.h>
#include <OpenSG/OSGPointLight.h>
#include <OpenSG/OSGFrameBufferObject.h>
#include <OpenSG/OSGTextureBuffer.h>
#include <OpenSG/OSGRenderBuffer.h>
#include <OpenSG/OSGFrameBufferObject.h>
#include <OpenSG/OSGTextureBuffer.h>
#include <OpenSG/OSGRenderBuffer.h>
#include <OpenSG/OSGTextureObjChunk.h>
#include <OpenSG/OSGTextureEnvChunk.h>
#include <libxml/parser.h>
#include <libxml/tree.h>


enum STEREO_TYPE {STEREO_NONE, STEREO_ANAGLYPH, STEREO_WINDOWED, 
                    STEREO_QUAD, STEREO_SIDEBYSIDE};
enum CAMERA_PARAMETER {CAMERA_HEAD_POS, CAMERA_HEAD_QUAT, CAMERA_HEAD_POSQUAT, 
                        CAMERA_PARAMETERS_NB};
class Reconstruct;
class Reconstruct3D;
class Mapping;
class MappingID;
class Event;

class Camera {
    public:
        Camera(int, OSG::NodeRecPtr, OSG::PassiveWindowRecPtr);
        ~Camera();
        void parse(xmlNodePtr);
        void reset();

        void getMappingsIDs(std::vector<MappingID>&);


        void setParameter(const CAMERA_PARAMETER& param, double* values, 
                                unsigned int nbValues);

        void setName(const std::string&);
        void setAdress(const std::string&);
        void setPort(const int&);
        void setStereo(const std::string&);
        void setScreenCorners(const unsigned int& corner, 
                            const std::vector<float>& coords);
        void setHeadInitialPosition(const std::vector<float>& coords);
        void setCameraBasePosition(const std::vector<float>& coords);
        void setCameraBaseRotation(const std::vector<float>& coords);
        void setHeadTracking(const std::string&, const std::string&);
        void setBackgroundVideo(float video);
        void setFollow(const std::string& camName);

        #ifdef WITH_FREENECT
        void addReconstruction(xmlNodePtr);
        #endif

        void initWindow();

        void connect(std::string, int);

        inline const std::string& getName(){return m_name;}

        void update();
        void resize(int,int);

        inline OSG::NodeRecPtr getNode(){return m_camBeacon;}
        inline OSG::NodeRecPtr getLightNode(){return m_light;}
        inline OSG::NodeRecPtr getInternalLightNode() { 
            return m_internalLightNode;
        }
        inline OSG::ComponentTransformRecPtr getHeadTrans(){return m_headXform;}
        inline const OSG::Vec3f& getPosition(){
                        return m_camXform->getTranslation();}
        inline const OSG::Quaternion& getOrientation(){
                        return m_camXform->getRotation();}

        inline const int& getID(){return m_id;}

    protected:
        int m_id;

        OSG::PerspectiveCameraRecPtr m_cam;
        OSG::NodeRecPtr m_camBeacon;
        OSG::ComponentTransformRecPtr m_camXform;
        OSG::NodeRecPtr m_head;
        OSG::ComponentTransformRecPtr m_headXform;
        OSG::Vec3f m_headBasePosition;
        OSG::Vec3f m_cameraBasePosition;
        OSG::Vec3f m_cameraBaseRotation;
        OSG::Pnt3f m_screenCorners[4];

        OSG::BackgroundRecPtr m_background;
        Reconstruct* m_reconstruct;

        OSG::NodeRecPtr m_light;
        OSG::PointLightRecPtr m_externalLight;
        OSG::ComponentTransformRecPtr m_lightTrans;
        OSG::NodeRecPtr m_internalLightNode;
        OSG::PointLightRecPtr m_internalLight;
        OSG::ComponentTransformRecPtr m_internalLightTrans;

        std::string m_name;
        STEREO_TYPE m_stereo;
        int m_xOverlap;
        double m_eyeSeparation;
        int m_nbDisplays;
        std::string m_adress;
        int m_port;

        OSG::ViewportRecPtr m_leftViewport;
        OSG::ViewportRecPtr m_rightViewport;
        OSG::StereoBufferViewportRecPtr m_leftViewportStereo;
        OSG::StereoBufferViewportRecPtr m_rightViewportStereo;
        OSG::ColorBufferViewportRecPtr m_leftViewportColor;
        OSG::ColorBufferViewportRecPtr m_rightViewportColor;
        
        OSG::ProjectionCameraDecoratorRecPtr m_leftCameraDecorator;
        OSG::ProjectionCameraDecoratorRecPtr m_rightCameraDecorator;
        OSG::MultiDisplayWindowRecPtr m_multiWindow;
        OSG::RenderActionRefPtr m_renderAction;
        
        std::vector<Mapping*> m_headMappings;    

        std::vector<double> m_parametersValues[CAMERA_PARAMETERS_NB];
        bool m_parametersChanged[CAMERA_PARAMETERS_NB];

        double m_fakeHeadMovement;
        double m_fakeHeadMovementStep;
        bool m_fakeHeadMovementsActivated;
        bool m_reconstHTActivated;

        #ifdef WITH_FREENECT
        std::vector<Reconstruct3D*> m_reconstructs;    
        #endif

    protected:
        void resetDisplay();
        void parseCoords(std::string &strCoords, std::vector<float>&);
};



#endif
