/***************************************************************************
 *  EnvManager.hpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "EnvManager.hpp"
#include "../Drile.hpp"

#include "Passage.hpp"
#include "Tunnel.hpp"
#include "Worm.hpp"
#include "Sign.hpp"
#include "Grid.hpp"
#include "Zone.hpp"
#include "Room.hpp"
#include "Rope.hpp"


using namespace std;


EnvManager::EnvManager(): m_started(false), m_currentRoom(0), 
                            m_roomIdCounter(0), m_tunnelIDCounter(0), 
                            m_gridIDCounter(0), m_wormIdCounter(0) {}

EnvManager::~EnvManager(){}

EnvManager* EnvManager::getInstance() {
    static EnvManager instance;
    return &instance;
}

void EnvManager::start() {
    {
        m_passiveWindow = OSG::PassiveWindow::create();

        setWormSize(0.1);

        //root node
        m_root = OSG::Node::create();
        m_root->setCore(OSG::Group::create());
        m_litRoot=m_root;

        //musicians
        OSG::NodeRecPtr tempRoot=m_root;
        m_globalCamera = OSG::Node::create();
        m_globalCameraTrans = OSG::ComponentTransform::create();
        m_globalCamera->setCore(m_globalCameraTrans);
        m_root->addChild(m_globalCamera);

        
        //add a first room
        m_roomsRoot = OSG::makeCoredNode<OSG::Group>();
        m_litRoot->addChild(m_roomsRoot);
        getCreateRoom(0);

        //make sure there is at least one musician
        addMusician();

        m_intersectAction = OSG::IntersectAction::create();

        OSG::commitChanges();
    }

    m_started=true;
    m_replaceCamera=false;
    m_replaceRoom=false;
}

void EnvManager::stop() {
    m_started=false;
}

void EnvManager::reset() {

    //remove the rest
    m_roomIdCounter=0;
    m_currentRoom=0;
    vector<Room*>::iterator itRoom;
    for(itRoom=m_rooms.begin(); itRoom!=m_rooms.end(); ++itRoom) {
        delete (*itRoom);
    }
    m_rooms.clear();

    //make sure we have at least one room
    getCreateRoom(0);
    //make sure we have at least one musician by replacing the first
    m_replaceCamera=true;
    vector<Musician*>::iterator itCam=m_musicians.begin();
    //reset the first musician
    (*itCam)->reset();
    //remove all other musicians
    itCam++;
    for(;itCam!=m_musicians.end(); ++itCam) {
        delete (*itCam);
    }
    m_musicians.erase(m_musicians.begin()+1,m_musicians.end());
    
    //remove all spectators
    vector<Spectator*>::iterator itSpect;
    for(itSpect=m_spectators.begin(); itSpect!=m_spectators.end(); ++itSpect) {
        delete (*itSpect);
    }
    m_spectators.clear();
    

    m_piiverts.clear();
    m_piivertsMap.clear();
    m_tunnels.clear();
    m_tunnelsMap.clear();
    m_tunnelIDCounter=0;
    m_wormsTrees.clear();
    m_wormsMap.clear();
    m_wormIdCounter=0;
    m_grids.clear();
    m_gridsMap.clear();
    m_gridIDCounter=0;


    //update lit root
    updateLitRoot();
}

Spectator* EnvManager::addSpectator() {
    Spectator* newCam 
        = new Spectator(m_spectators.size(), m_root, m_passiveWindow);
    m_spectators.push_back(newCam);
    m_globalCamera->addChild(newCam->getNode());

    //get light and add as child of root
    updateLitRoot();

    return newCam;
}

void EnvManager::moveToRoom(Room* room) {
    vector<Piivert*>::iterator itPiiv = m_piiverts.begin();
    for(; itPiiv!=m_piiverts.end(); ++itPiiv) {
        (*itPiiv)->setRoom(room);
    }
    m_currentRoom = room->getID();
}

Musician* EnvManager::addMusician() {
    Musician* newCam;
    if(m_replaceCamera) {
        newCam=m_musicians[0];
        m_replaceCamera=false; 
    }
    else {
        newCam = new Musician(m_musicians.size(), m_root, m_passiveWindow);
        m_musicians.push_back(newCam);
        newCam->setCurrentRoom(m_rooms[0]);
        m_globalCamera->addChild(newCam->getNode());

        updateLitRoot();


        for(unsigned int i=0; i<m_musicians.size()-1; ++i) {
            //create one pair of passages per other musician
            Passage * firstPassage = new Passage();
            Passage * secondPassage = new Passage();
            firstPassage->setOtherPassage(secondPassage);
            secondPassage->setOtherPassage(firstPassage);
            //firstPassage->setOtherZoneFboMaterial(
            //                            m_zones[i]->getFboMaterial());
            //secondPassage->setOtherZoneFboMaterial(
            //                           newZone->getFboMaterial());
            newCam->addPassage(firstPassage);
            firstPassage->setColor(m_musicians[i]->getColor());
            m_musicians[i]->addPassage(secondPassage);
            secondPassage->setColor(newCam->getColor());
        }


    }

    //update musicians orientations and positions list
    updateMusiciansZones();
    
    return newCam;
}

Camera* EnvManager::getCamera(const std::string& camStr) {
    Camera* res=NULL;
    for(unsigned int i=0; i<m_musicians.size() && !res; ++i) {
        if(m_musicians[i]->getName().compare(camStr)==0) {
            res=m_musicians[i];
        }
    }
    for(unsigned int i=0; i<m_spectators.size() && !res; ++i) {
        if(m_spectators[i]->getName().compare(camStr)==0) {
            res=m_spectators[i];
        }
    }
    return res;
}

void EnvManager::updateLitRoot() {
    //reconstruct the light graph
    m_litRoot=m_root;
    vector<Spectator*>::iterator itSpect;
    for(itSpect=m_spectators.begin(); itSpect!=m_spectators.end(); ++itSpect) {
        OSG::NodeRecPtr lightNode = (*itSpect)->getLightNode();
        if(lightNode->getParent()) {
            lightNode->getParent()->subChild(lightNode);
        }
        lightNode->clearChildren();;
        m_litRoot->addChild(lightNode);
        m_litRoot=lightNode;
    }
    vector<Musician*>::iterator itMus;
    for(itMus=m_musicians.begin(); itMus!=m_musicians.end(); ++itMus) {
        OSG::NodeRecPtr lightNode = (*itMus)->getLightNode();
        if(lightNode->getParent()) {
            lightNode->getParent()->subChild(lightNode);
        }
        lightNode->clearChildren();;
        m_litRoot->addChild(lightNode);
        m_litRoot=lightNode;
    }
    //put all rooms under the new light root
    if(m_roomsRoot->getParent()) {
        m_roomsRoot->getParent()->subChild(m_roomsRoot);
    }
    m_litRoot->addChild(m_roomsRoot);
}


Room* EnvManager::addRoom() {
    m_rooms.push_back(new Room(m_roomIdCounter));
    m_roomsRoot->addChild(m_rooms.back()->getNode());
    updateMusiciansZones();
    m_roomIdCounter++;
    return m_rooms.back();
}

Room* EnvManager::getCreateRoom(const unsigned int& id) {
    while(id>=m_rooms.size()) {
        addRoom();
    }
    return m_rooms[id];
}

void EnvManager::updateMusiciansZones() {
    m_musiciansPositions.clear();
    m_musiciansOrientations.clear();

    //add position and orientation of musicians with piivert attached 
    vector<Musician*>::iterator itCam;
    for(itCam=m_musicians.begin(); itCam!=m_musicians.end(); ++itCam) {
        m_musiciansPositions.push_back((*itCam)->getPosition());
        m_musiciansOrientations.push_back((*itCam)->getOrientation());
    }

    //send to all existing rooms
    vector<Room*>::iterator itRoom;
    for(itRoom=m_rooms.begin(); itRoom!=m_rooms.end(); ++itRoom) {
        (*itRoom)->setZonesParams(m_musiciansPositions,m_musiciansOrientations);
    }
}

Worm* EnvManager::addWorm(const NODE_TYPE& type, const int& roomID, 
                            const unsigned int& zone, const int& parentWorm, 
                            xmlNodePtr xmlNode) {
    Worm* newWorm; 
    switch(type) {
        case NODE_LOOP:{
            newWorm = new LoopWorm(m_wormIdCounter);
        }break;
        case NODE_SYNC:{
            newWorm = new SyncWorm(m_wormIdCounter);
        }break;
        case LEAF_SAMPLE: 
        case LEAF_LAYERED_GRAIN:
        case LEAF_SAS:{
            newWorm = new LeafWorm(m_wormIdCounter);
        }break;
        default:{
            newWorm = new LeafWorm(m_wormIdCounter);
        }break;
    }

    m_wormsTrees.push_back(newWorm);
    m_wormsMap[m_wormIdCounter]=newWorm;

    bool found=false;
    vector<Room*>::iterator itRoom;
    for(itRoom=m_rooms.begin(); itRoom!=m_rooms.end() && !found; ++itRoom) {
        if((*itRoom)->getID()==roomID) {
            (*itRoom)->getCreateZone(zone)->addWorm(newWorm);
            found=true;
        }
    }

    if(xmlNode) {
        newWorm->parse(xmlNode);
    }

    m_wormIdCounter++;

    return newWorm;
}

void EnvManager::setWormSize(const float& size) {
    Worm::WORM_SIZE = size;
    Worm::WORM_DISTANCE = Worm::WORM_SIZE*3.0;
}

Piivert* EnvManager::addPiivert(const int& id, Musician* cam) {
    Piivert* newPiiv = new Piivert(id);
    m_piivertsMap[id] = newPiiv;
    m_piiverts.push_back(newPiiv);
    newPiiv->setMusician(cam);
    newPiiv->setRoom(m_rooms[m_currentRoom]);
    cam->addPiivert(newPiiv);
    return newPiiv;
}

Tunnel* EnvManager::addTunnel(const int& roomID, const unsigned int& zone) {
    Tunnel* newTunnel = new Tunnel(m_tunnelIDCounter);
    m_tunnels.push_back(newTunnel);
    m_tunnelsMap[m_tunnelIDCounter]=newTunnel;
    bool found=false;        
    for(vector<Room*>::iterator it=m_rooms.begin(); 
                it!=m_rooms.end() && !found; ++it) {
        if((*it)->getID()==roomID) {
            (*it)->getCreateZone(zone)->addTunnel(newTunnel);
            found=true;
        }
    }
    m_tunnelIDCounter++;
    return newTunnel;
}

Grid* EnvManager::addGrid(const int& roomID, const unsigned int& zone) {
    Grid* newGrid = new Grid(m_gridIDCounter);
    newGrid->setZoneID(zone);
    m_grids.push_back(newGrid);
    m_gridsMap[m_gridIDCounter]=newGrid;
    bool found=false;        
    for(vector<Room*>::iterator it=m_rooms.begin(); 
                        it!=m_rooms.end() && !found; ++it) {
        if((*it)->getID()==roomID) {
            (*it)->getCreateZone(zone)->addGrid(newGrid);
            found=true;
        }
    }
    m_gridIDCounter++;
    return newGrid;
}

void EnvManager::addRope(Room* room1, Room* room2) {
    Rope* newRope = new Rope();
    newRope->setFirstRoom(room1);
    newRope->setSecondRoom(room2);
}

void EnvManager::addLight(OSG::Vec3f pos) {
    OSG::NodeRecPtr lightBeacon = OSG::Node::create();
    OSG::ComponentTransformRecPtr lightBeaconTrans 
                                        = OSG::ComponentTransform::create();
    lightBeaconTrans->setTranslation(pos);
    lightBeacon->setCore(lightBeaconTrans);    
    OSG::PointLightRecPtr pLight = OSG::PointLight::create();
    pLight->setDiffuse(OSG::Color4f(1,1,1,1));
    pLight->setAmbient(OSG::Color4f(0.2,0.2,0.2,1));
    pLight->setSpecular(OSG::Color4f(1,1,1,1));
    pLight->setBeacon(lightBeacon);
    OSG::NodeRecPtr lightNode = OSG::Node::create();
    lightNode->setCore(pLight);
    m_root->addChild(lightNode);
}

void EnvManager::initWindows() {
    for(unsigned int i=0;i<m_musicians.size();++i) {
        m_musicians[i]->initWindow();
    }
    for(unsigned int i=0;i<m_spectators.size();++i) {
        m_spectators[i]->initWindow();
    }
    m_passiveWindow->init();
}

void EnvManager::update(std::vector<Event>& inputEvents, 
                    std::vector<Event>& outputEvents) {

    //get current time for animations and loops
    gettimeofday(&m_currentTime, NULL);

    vector<Event>::iterator itEvent;
    for(itEvent=inputEvents.begin(); itEvent!=inputEvents.end();++itEvent) {
        switch((*itEvent).type) {
            case EVENT_NODE: {
                if(m_wormsMap.find((*itEvent).node)!=m_wormsMap.end()) {

                    m_wormsMap[(*itEvent).node]->setParameter(    
                        Worm::nodeWormMap[NODE_PARAMETER((*itEvent).parameter)],
                        (*itEvent).values, Event::nbValues);
                }    
            }break;
            case EVENT_PIIVERT: {
                if(m_piivertsMap.find((*itEvent).piivert)
                        !=m_piivertsMap.end()) {

                    m_piivertsMap[(*itEvent).piivert]->setParameter(
                                (*itEvent).hand, 
                                PIIVERT_PARAMETER((*itEvent).parameter), 
                                (*itEvent).values, 
                                Event::nbValues, outputEvents);
                }
            }break;
            case EVENT_CAMERA: {
                m_musicians[(*itEvent).camera]->setParameter(
                            CAMERA_PARAMETER((*itEvent).parameter), 
                            (*itEvent).values, Event::nbValues);
            }break;
            case EVENT_TREE: {
                switch((*itEvent).parameter) {
                    case TREE_NODE_ADD: {
                        Worm* newWorm=NULL;
                        switch(int((*itEvent).values[0])) {
                            case NODE_LOOP: {
                                newWorm = new LoopWorm(
                                                    (*itEvent).values[1]);
                            }break;
                            case NODE_SYNC: {
                                newWorm = new SyncWorm(
                                                    (*itEvent).values[1]);
                            }break;
                            default:{
                                newWorm = new LeafWorm(
                                                    (*itEvent).values[1]);
                            }break;
                        }
                        //add the new worm to the map
                        m_wormsMap[newWorm->getID()]=newWorm;
                        //create a new tree root
                        m_wormsTrees.push_back(newWorm);

                        //if given a corresponding worm (child or ...)
                        unsigned int zone=0;
                        if((*itEvent).values[2]!=-1) {
                            Worm* corresWorm = 
                                m_wormsMap[(*itEvent).values[2]];
                            zone=corresWorm->getZoneID();
                            //put at the same position + size on x
                            newWorm->getTrans()->setTranslation(
                                corresWorm->getTrans()->getTranslation());
                            OSG::Vec3f& pos = 
                                newWorm->getTrans()->editTranslation();
                            pos[0]+=Worm::WORM_SIZE/2.0;
                            pos[2]=0;
                        }
                        newWorm->setZoneID(zone);
                        //add to current room
                        m_rooms[m_currentRoom]->getCreateZone(zone)
                                                    ->addWorm(newWorm);
                        m_rooms[m_currentRoom]
                            ->getCreateZone(zone)->getWormsRoot()
                                                ->addChild(newWorm->getNode());
                    }break;
                    case TREE_NODE_REMOVE: {
                        Worm* removedWorm=
                            m_wormsMap[(*itEvent).values[0]]; 
                        //if it has a parent, remove it from its parent
                        if(removedWorm->getParent()) {
                            removedWorm->getParent()->removeChild(removedWorm);
                        }

                        //remove the worm with the given id
                        m_wormsMap.erase((*itEvent).values[0]);
                        vector<Worm*>::iterator itWorm;
                        for(itWorm=m_wormsTrees.begin();
                                itWorm!=m_wormsTrees.end();++itWorm) {
                            if((*itWorm)->getID()==
                                    (*itEvent).values[0]) {
                                m_wormsTrees.erase(itWorm);
                                break;
                            }
                        }
                        //remove it from the current Room
                        m_rooms[m_currentRoom]->
                                        getCreateZone(removedWorm->getZoneID())
                                            ->removeWorm(removedWorm);

                        //delete the node
                        delete removedWorm;

                    }break;
                    case TREE_NODE_MOVE: {
                        //set second worm as parent of the first worm 
                        //move to root if parentid==-1
                        Worm* movedWorm = 
                            m_wormsMap[(*itEvent).values[0]]; 
                        
                        //move from root
                        if((*itEvent).values[1]==-1) {
                                //remove from root
                                vector<Worm*>::iterator itWorm;
                                for(itWorm=m_wormsTrees.begin(); 
                                        itWorm!=m_wormsTrees.end();++itWorm) {
                                    if((*itWorm)->getID()==movedWorm->getID()) {
                                        m_wormsTrees.erase(itWorm);
                                        break;
                                    }
                                }

                                if((*itEvent).values[2]>=0) {
                                    //add to new parent
                                    Worm *newParent =
                                        m_wormsMap[(*itEvent).values[2]]; 
                                    newParent->addChild(movedWorm);
                                    newParent->updateChildrenPositions();
                                }
                                else {//put back to root
                                    m_wormsTrees.push_back(movedWorm);
                                    OSG::Vec3f& pos = 
                                    movedWorm->getTrans()->editTranslation();
                                    pos[2]=0;
                                }
                        }
                        else {
                            Worm* oldParent = 
                                m_wormsMap[(*itEvent).values[1]]; 
                            oldParent->removeChild(movedWorm);
                            //move to root
                            if((*itEvent).values[2]==-1) {
                                m_wormsTrees.push_back(movedWorm);
                                OSG::Vec3f& pos = 
                                    movedWorm->getTrans()->editTranslation();
                                pos[2]=0;
                            }
                            else {
                                //add to new parent
                                Worm *newParent =
                                    m_wormsMap[(*itEvent).values[2]]; 
                                newParent->addChild(movedWorm);
                                newParent->updateChildrenPositions();
                            }
                        }
                        movedWorm->updateChildrenPositions();
                    }break;
                }
            }break;
            default:break;
        }
    }


    //clear input events
    inputEvents.clear();

    //update piiverts
    for(unsigned int p=0; p<m_piiverts.size(); ++p) {
        //update parameters
        m_piiverts[p]->update(m_currentTime, outputEvents, 
                                m_intersectAction, m_rooms[m_currentRoom]);
    }
    
    //update room
    m_rooms[m_currentRoom]->update(m_intersectAction, m_currentTime, 
                                    outputEvents);
}

void EnvManager::updateWindows() {
    try {
        //check the musicians and spectators for changes and display
        for(unsigned int i=0;i<m_musicians.size();++i) {
            m_musicians[i]->update();
        }
        for(unsigned int i=0;i<m_spectators.size();++i) {
            m_spectators[i]->update();
        }
        OSG::Thread::getCurrentChangeList()->clear();
    }
    catch(OSG_STDEXCEPTION_NAMESPACE::exception &e) {
        OSG::Thread::getCurrentChangeList()->clear();
    }
}


void EnvManager::reshape(int w,int h) {
    for(unsigned int i=0;i<m_musicians.size();++i) {
        m_musicians[i]->resize(w, h);
    }
    for(unsigned int i=0;i<m_musicians.size();++i) {
        m_spectators[i]->resize(w, h);
    }
}


Room* EnvManager::getRoom(const std::string& roomStr) {
    Room* room=NULL;
    for(unsigned int r=0; r<m_rooms.size() && !room; ++r) {
        if(m_rooms[r]->getName().compare(roomStr)==0) {
            room=m_rooms[r];
        }
    }
    return room;
}

