/***************************************************************************
 *  EnvManager.hpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef ENV_MANAGER_H
#define ENV_MANAGER_H

#include <OpenSG/OSGConfig.h>
#include <OpenSG/OSGGLUT.h>
#include <OpenSG/OSGSimpleGeometry.h>
#include <OpenSG/OSGPassiveWindow.h>
#include <OpenSG/OSGSimpleSceneManager.h>
#include <OpenSG/OSGSceneFileHandler.h>
#include <OpenSG/OSGGradientBackground.h>
#include <OpenSG/OSGBaseFunctions.h>
#include <OpenSG/OSGTransform.h>
#include <OpenSG/OSGComponentTransform.h>
#include <OpenSG/OSGTriangleIterator.h>
#include <OpenSG/OSGMaterialGroup.h>
#include <OpenSG/OSGShearedStereoCameraDecorator.h>
#include <OpenSG/OSGProjectionCameraDecorator.h>
#include <OpenSG/OSGMultiDisplayWindow.h>
#include <OpenSG/OSGImage.h>
#include <OpenSG/OSGImageFileHandler.h>
#include <OpenSG/OSGChunkMaterial.h>
#include <OpenSG/OSGMaterialGroup.h>
#include <OpenSG/OSGPointLight.h>

#include <iostream>
#include <math.h>
#include <vector>
#include <sys/time.h>

#include "Musician.hpp"
#include "Spectator.hpp"

#include "Room.hpp"

#include "Worm.hpp"
#include "LoopWorm.hpp"
#include "LeafWorm.hpp"
#include "SyncWorm.hpp"

#include "Piivert.hpp"

#include "../HllEnums.hpp"
#include "../utils/RingBuffer.hpp"

class EnvManager {
    public:
        static EnvManager* getInstance();
        ~EnvManager();
        void start();
        void stop();
        void reset();

        void reshape(int,int);

        void addLight(OSG::Vec3f);

        void initWindows();
        void updateWindows();
        void update(std::vector<Event>& inputEvents, 
                        std::vector<Event>& outputEvents);

        //root
        inline OSG::NodeRecPtr getRoot(){return m_root;}

        inline OSG::PassiveWindowRecPtr getWindow(){return m_passiveWindow;}

        void setWormSize(const float&);
        Worm* addWorm(const NODE_TYPE& type, const int& room, 
                        const unsigned int& zone, const int& parent=-1, 
                        xmlNodePtr xmlNode=NULL);
        inline Worm* getWorm(const int& id) {
            if(m_wormsMap.find(id)!=m_wormsMap.end()) { 
                return m_wormsMap[id];
            } else { 
                return NULL;
            }
        }
        inline int getNbWorms(){return m_wormsMap.size();}

        void moveToRoom(Room*);
        OSG::Vec3f& editGlobalCameraPosition() { 
            return m_globalCameraTrans->editTranslation();
        }
        Musician* addMusician();
        inline Musician* getMusician(const int& cam){return m_musicians[cam];}
        inline unsigned int getNbMusicians(){return m_musicians.size();}

        Spectator* addSpectator();
        inline Spectator* getSpectator(const int& cam){ 
                                                return m_spectators[cam];}
        inline unsigned int getNbSpectators(){return m_spectators.size();}

        Camera* getCamera(const std::string& camStr);

        Piivert* addPiivert(const int&, Musician*);
        inline Piivert* getPiivert(const int& piiv){return m_piiverts[piiv];}
        inline int getNbPiiverts(){return m_piiverts.size();}
        inline Piivert* getPiivertFromID(const int& piivID) {    
            if(m_piivertsMap.find(piivID)!=m_piivertsMap.end()) {
                return m_piivertsMap[piivID];
            }
            else {
                std::cout<<"Piivert "<<piivID<<" not found"<<std::endl;
                return NULL;
            }
        }

        Room* addRoom();
        Room* getCreateRoom(const unsigned int&);
        inline Room* getRoom(int room){return m_rooms[room];}
        inline int getNbRooms(){return m_rooms.size();}
        inline Room* getCurrentRoom(){return m_rooms[m_currentRoom];}
        Room* getRoom(const std::string&);

        Tunnel* addTunnel(const int&, const unsigned int&);
        inline Tunnel* getTunnel(int tunnel){return m_tunnels[tunnel];}
        inline Tunnel* getTunnelFromMap(int tunnelID) {
            if(m_tunnelsMap.find(tunnelID)!=m_tunnelsMap.end()) {
                return m_tunnelsMap[tunnelID];
            }
            else {
                return NULL;
            }
        }
        inline int getNbTunnels(){return m_tunnels.size();}

        Grid* addGrid(const int&, const unsigned int&);
        inline Grid* getGrid(int grid){return m_grids[grid];}
        inline Grid* getGridFromMap(int gridID) {
            if(m_gridsMap.find(gridID)!=m_gridsMap.end()) {
                return m_gridsMap[gridID];
            } else {
                return NULL;
            }
        }
        inline int getNbGrids(){return m_grids.size();}
        
        void addRope(Room* room1, Room* room2);

        inline bool isStarted(){return m_started;}


    private:
        EnvManager();
        EnvManager( const EnvManager & ); 
        EnvManager & operator =( const EnvManager & ); 
       
        void updateLitRoot(); 

        
    private:    
        OSG::NodeRecPtr m_root;
        OSG::NodeRecPtr m_litRoot;
        OSG::PassiveWindowRecPtr m_passiveWindow;

        timeval m_currentTime;

        bool m_started;

        std::vector<Room*> m_rooms;
        int m_currentRoom;
        int m_roomIdCounter;
        bool m_replaceRoom;
        OSG::NodeRecPtr m_roomsRoot;

        void updateMusiciansZones();
        std::vector<OSG::Vec3f> m_musiciansPositions;
        std::vector<OSG::Quaternion> m_musiciansOrientations;
        std::vector<double> m_musiciansColors;

        std::vector<Tunnel*> m_tunnels;
        std::map<int, Tunnel*> m_tunnelsMap;
        int m_tunnelIDCounter;

        std::vector<Grid*> m_grids;
        std::map<int, Grid*> m_gridsMap;
        int m_gridIDCounter;

        std::vector<Worm*> m_wormsTrees;
        std::map<int,Worm*> m_wormsMap;
        int m_wormIdCounter;

        OSG::NodeRecPtr m_globalCamera;
        OSG::ComponentTransformRecPtr m_globalCameraTrans;
        std::vector<Musician*> m_musicians;
        std::vector<Spectator*> m_spectators;
        bool m_replaceCamera;

        std::map<int, Piivert*> m_piivertsMap;
        std::vector<Piivert*> m_piiverts;

        OSG::IntersectActionRefPtr m_intersectAction;

};

#endif
