/***************************************************************************
 *  Piivert.hpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef _PIIVERT_H
#define _PIIVERT_H

#include <iostream>
#include <cstring>
#include <cmath>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <vector>
#include <map>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <OpenSG/OSGSimpleGeometry.h>
#include <OpenSG/OSGComponentTransform.h>
#include <OpenSG/OSGMaterialGroup.h>
#include <OpenSG/OSGSimpleMaterial.h>
#include <OpenSG/OSGIntersectAction.h>
#include <OpenSG/OSGNameAttachment.h>
#include <OpenSG/OSGCylinderVolume.h>

#include "../Event.hpp"
#include "../utils/OneEuroFilter.hpp"

enum PIIVERT_PARAMETER { PIIVERT_HIT, PIIVERT_PRESS, PIIVERT_GRAB, 
                        PIIVERT_GRAB_SELECT, PIIVERT_SLIP, 
                        PIIVERT_ERASE_ONE_HAND, PIIVERT_ERASE_TWO_HANDS, 
                        PIIVERT_RECORD_ONE_HAND, PIIVERT_RECORD_TWO_HANDS, 
                        PIIVERT_STORE, PIIVERT_POSX, PIIVERT_POSY, PIIVERT_POSZ,
                        PIIVERT_POS, PIIVERT_ROTX, PIIVERT_ROTY, PIIVERT_ROTZ, 
                        PIIVERT_ROT, PIIVERT_QUAT, PIIVERT_POSQUAT, 
                        PIIVERT_SELECT_WORM, PIIVERT_SPEED, 
                        PIIVERT_PARAMETERS_NB};

class Musician;
class Worm;
class Tunnel;
class Room;
class Sign;
class Passage;
class Manipulable;

class Piivert {
    public:
        Piivert(const int&);
        virtual ~Piivert();
        void parse(xmlNodePtr);

        inline std::string getName(){return m_name;}
        inline const int& getID(){return m_id;}

        inline OSG::NodeRecPtr getNode(){return m_node;}

        void setColor(const float&);
        void setParameter(std::string, std::string);
        void setParameter(const int&, const PIIVERT_PARAMETER&, 
                            double*, unsigned int, std::vector<Event>&);

        inline OSG::NodeRecPtr getHandNode(const int& h ){return m_piivert[h];}

        virtual void update(const timeval&, std::vector<Event>&,
                            OSG::IntersectAction*, Room*);

        void setMusician(Musician* cam);

        void release(const int& hand, std::vector<Event>& outputEvents);

        void storeTunnel(Tunnel*,int);
        void removeTunnel(Tunnel*,int);

        void attachGrabbedManipulable(const int& hand);
        void detachGrabbedManipulable(const int& hand);

        inline void setRoom(Room* room){m_currentRoom=room;}

    protected:

        std::string m_name;
        int m_id;
        double m_color;

        unsigned int m_zoneID;
        Room* m_currentRoom;

        OSG::NodeRecPtr m_node;
        Musician* m_parentMusician;

        timeval m_timeDiff;
        timeval m_previousTime;
        LowPassFilter* m_grabbingSpeedFilter[2][3];
        OSG::Vec3f m_grabbingSpeed[2];

        OneEuroFilter* m_positionFilter[2][3];
    
        OSG::NodeRecPtr m_piivert[2];
        OSG::ComponentTransformRecPtr m_piivertTrans[2];
        OSG::Vec3f m_piivertDir[2];

        bool m_recording[2];
        float m_recordingRotation[2];

        std::vector<OSG::NodeRecPtr> m_elements[2];
        std::vector<OSG::ComponentTransformRecPtr> m_elementsTrans[2];
        OSG::SimpleMaterialRecPtr m_material;

        unsigned int m_piivertElementsNb;
        double m_piivertLength;
        double m_piivertWidth;

        enum GRAB_STATE{GRAB_NONE, GRAB_ASKED, GRAB_SELECT, 
                        GRAB_SELECTION_MADE, GRAB_SLIPPING, GRAB_GRABBING};
        GRAB_STATE m_grabState[2];

        bool m_parametersChanged[2][PIIVERT_PARAMETERS_NB];
        std::vector<double> m_parametersValues[2][PIIVERT_PARAMETERS_NB];


        OSG::Quaternion m_handRotation[2];
        OSG::NodeRecPtr m_multiSelecCenterNode[2];
        OSG::ComponentTransformRecPtr m_multiSelecCenterTrans[2];
        OSG::SimpleMaterialRecPtr m_multiSelecCenterMat[2];
        OSG::ComponentTransformRecPtr m_multiSelecTrans[2][2];
        OSG::NodeRecPtr m_multiSelecNodes[2][2];
        int m_nbFingers;

        double m_grabThreshold;
        bool m_initialGrabDirection;
        bool m_grabToggle[2];
        bool m_doneDuplicating;

        Manipulable* m_selectedManipulable[2];
        Manipulable* m_grabbedManipulable[2];
        std::vector<Manipulable*> m_multiSelectedManipulable[2];
        std::vector<Manipulable*> m_multiGrabbedManipulable[2];

        double m_vibrationCounter[2];
        double m_vibrationStep[2];
        double m_vibrationAmplitude[2];

        OSG::Vec3f m_previousTrackedPosition[2];

        void setVisibleLength(const unsigned int& hand, double length);

        void testMultipleSelection(const unsigned int& hand,  
                                    Room*,std::vector<Event>&);
        void testRay(const unsigned int& hand, 
                        OSG::IntersectAction*, Room*,std::vector<Event>&);

        bool testIntersectManipulable(const int& h, OSG::Line& line, 
                                    OSG::IntersectAction*, Room* currentRoom,
                                    std::vector<Event>& outputEvents);

        Worm* m_selectedWorm[2];
        Worm *m_grabbedWorm[2];
        OSG::Vec3f m_initialGrabPosition[2];
        Tunnel* m_selectedTunnel[2];
        Tunnel *m_grabbedTunnel[2];
        std::vector<Tunnel*> m_packedTunnels[2];
        Passage* m_selectedPassage[2]; 
        Passage* m_grabbedPassage[2]; 
        Sign* m_selectedSign[2]; 
        Sign* m_grabbedSign[2]; 
        std::vector<Worm*> m_multiSelectedWorms[2];
        std::vector<Worm*> m_multiGrabbedWorms[2];

};



#endif
