/***************************************************************************
 *  Rope.cpp
 *  Part of Drile
 *  2010-2013 Florent Berthaut
 *  hitmuri.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "Rope.hpp"

#include <OpenSG/OSGVoidPAttachment.h>
#include "Room.hpp"
#include "EnvManager.hpp"

using namespace std;

Rope::Rope(): Manipulable(0), m_thickness(0.01) {
    m_rooms[0]=NULL;
    m_rooms[1]=NULL;
    {
        m_material = OSG::SimpleMaterial::create();
        m_material->setLit(false);
        m_matChunk = OSG::PolygonChunk::create();
        m_material->addChunk(m_matChunk);
        OSG::Color3f &color = m_material->editDiffuse();
        color.setValuesHSV(360.0, 0, 0.8); 
        OSG::VoidPAttachmentRefPtr ptr = OSG::VoidPAttachment::create();
        ptr->editFieldPtr()->setValue(static_cast<Manipulable*>(this));
        for(unsigned int r=0; r<2; ++r) {
            OSG::GeometryRefPtr geo=OSG::makeBoxGeo(1, 1, 1, true, true, true);
            geo->setMaterial(m_material);
            OSG::NodeRefPtr ropeGeom = OSG::Node::create();
            ropeGeom->setCore(geo);
            ropeGeom->addAttachment(ptr);
            m_rope[r] = OSG::Node::create();
            m_ropeTrans[r] = OSG::ComponentTransform::create();
            m_rope[r]->setCore(m_ropeTrans[r]);
            m_rope[r]->addChild(ropeGeom);
            m_rope[r]->addAttachment(ptr);
        }
        OSG::commitChanges();
    }
}

Rope::~Rope() {}

void Rope::setFirstRoom(Room* room) { 
    m_rooms[0]=room;
    m_rooms[0]->addRopeNode(m_rope[0]);
    updatePosition();
}

void Rope::setSecondRoom(Room* room) { 
    m_rooms[1]=room;
    m_rooms[1]->addRopeNode(m_rope[1]);
    updatePosition();
}

void Rope::updatePosition() {
    if(m_rooms[0] && m_rooms[1]) {
        OSG::Vec3f dir[2];
        dir[0] = m_rooms[1]->getPosition() - m_rooms[0]->getPosition();
        dir[1] = m_rooms[0]->getPosition() - m_rooms[1]->getPosition();
        dir[0].normalize();
        dir[1].normalize();
        for(unsigned int r=0; r<2; ++r) {
            m_ropeTrans[r]->setTranslation(OSG::Vec3f(0, 1.5, 0) + dir[r]);
            m_ropeTrans[r]->setScale(OSG::Vec3f(m_thickness, 1.0, m_thickness));
            OSG::Quaternion& quat = m_ropeTrans[r]->editRotation();
            quat.setValue(OSG::Vec3f(0, 1, 0), dir[r]);
        }
    }
}

void Rope::grab(Piivert* tool, 
                    const int& hand, 
                        std::vector<Event>& outputEvents) {
    Manipulable::grab(tool, hand, outputEvents);
}

void Rope::release(Piivert* tool, 
                        const int& hand, 
                            std::vector<Event>& outputEvents) {
    Manipulable::release(tool, hand, outputEvents);
    m_grabInitalised=false;

    OSG::Vec3f& camPos = EnvManager::getInstance()->editGlobalCameraPosition();
    if((camPos-m_rooms[0]->getPosition()).length() 
            < (camPos-m_rooms[1]->getPosition()).length()) {
        camPos = m_rooms[0]->getPosition();
        EnvManager::getInstance()->moveToRoom(m_rooms[0]);
    }
    else {
        camPos = m_rooms[1]->getPosition();
        EnvManager::getInstance()->moveToRoom(m_rooms[1]);
    }
    //detach manipulable grabbed with the other hand
    tool->detachGrabbedManipulable(1-hand);
}

void Rope::moveTo(Piivert* tool,
                        const int& hand, 
                            const float& depth, 
                                const OSG::Line& ray, 
                                    std::vector<Event>& outputEvents) {
    OSG::Vec3f pos;
    OSG::Plane plane(OSG::Vec3f(0,0,-1), OSG::Pnt3f(0, 0, depth));
    plane.intersect(ray, pos);
    if(!m_grabInitalised) {
        m_initGrabPos=pos;
        m_grabInitalised=true;
        m_initCamPos=EnvManager::getInstance()->editGlobalCameraPosition();
        if((m_initCamPos-m_rooms[0]->getPosition()).length() 
                < (m_initCamPos-m_rooms[1]->getPosition()).length()) {
            m_roomFrom=m_rooms[0];
            m_roomTo=m_rooms[1];
        } 
        else {
            m_roomFrom=m_rooms[1];
            m_roomTo=m_rooms[0];
        }
        m_ropeDir=m_roomTo->getPosition()-m_roomFrom->getPosition();

        //attach objects grabbed with the other hand
        tool->attachGrabbedManipulable(1-hand);
    }
    else {
        pos=pos-m_initGrabPos;
        OSG::Vec3f& camPos
                        = EnvManager::getInstance()->editGlobalCameraPosition();
        camPos=m_roomFrom->getPosition()
                    - m_ropeDir*(m_ropeDir.dot(pos)/(Worm::WORM_SIZE*20.0));
    }
}

void Rope::hit(std::vector<Event>& outputEvents) {

}

void Rope::setHighlight(const bool& highlight, 
                            const double& hue, const double& scale) {
    if(!highlight) { 
        OSG::Color3f &color = m_material->editDiffuse();
        color.setValuesHSV(0, 0, 0.8); 
        m_matChunk->setFrontMode(GL_FILL);
        m_matChunk->setBackMode(GL_FILL);
    }
    else if(highlight) {
        OSG::Color3f &color = m_material->editDiffuse();
        color.setValuesHSV(hue*360.0, 0.6, 0.8); 
        m_matChunk->setFrontMode(GL_LINE);
        m_matChunk->setBackMode(GL_LINE);
    }
}

