/***************************************************************************
 *  Tunnel.cpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "Tunnel.hpp"
#include "Zone.hpp"
#include "Musician.hpp"

using namespace std;

Tunnel::Tunnel(const int& id): Ghostable(), Manipulable(id), m_currentScale(0), 
                                m_sides(4), m_nbElements(50) {

    for(unsigned int p=0; p<Worm::WORM_PARAMETERS_NB; ++p) {
        m_parameterActivated[p]=false;
    }

    {
        OSG::VoidPAttachmentRefPtr ptr = OSG::VoidPAttachment::create();
        ptr->editFieldPtr()->setValue(static_cast<Manipulable*>(this));

        double tunnelElementsWidth = (Worm::WORM_SIZE*3.0)/double(m_nbElements);

        m_tunnel = OSG::Node::create();
        m_tunnel->addAttachment(ptr);
        m_tunnelTrans = OSG::ComponentTransform::create();
        m_tunnelTrans->setTranslation(m_position);
        m_tunnel->setCore(m_tunnelTrans);

        OSG::GeometryRefPtr geo = 
                OSG::makeBoxGeo(tunnelElementsWidth*double(m_nbElements), 
                                Worm::WORM_SIZE/2.0, Worm::WORM_SIZE/2.0, 
                                1, 1, 1);
        m_highlightMaterial = OSG::SimpleMaterial::create();
        m_highlightMaterial->setLit(false);
        m_highlightMaterial->setColorMaterial(GL_NONE);
        m_highlightMaterial->setDiffuse(OSG::Color3f(0.5, 0.5, 0.5));
        m_highlightMaterial->setTransparency(1);
        OSG::PolygonChunkRefPtr chunk = OSG::PolygonChunk::create();
        chunk->setFrontMode(GL_LINE);
        chunk->setBackMode(GL_LINE);
        m_highlightMaterial->addChunk(chunk);
        geo->setMaterial(m_highlightMaterial);
        m_highlight = OSG::Node::create();
        m_highlight->addAttachment(ptr);
        m_highlight->setCore(geo);
        m_tunnel->addChild(m_highlight);

        m_elementsRoot = OSG::makeCoredNode<OSG::Group>(); 
        m_tunnel->addChild(m_elementsRoot);
        for(int i=0; i<m_nbElements; ++i) {
            OSG::GeometryRefPtr geo2 = 
                OSG::makeCylinderGeo(tunnelElementsWidth, 
                            Worm::WORM_SIZE/4.0, m_sides, true, false, false);

            OSG::SimpleMaterialRefPtr material = OSG::SimpleMaterial::create();
            material->setLit(true);
            material->setColorMaterial(GL_AMBIENT_AND_DIFFUSE);
            material->setDiffuse(OSG::Color3f(0.5,0.5,0.5));
            material->setAmbient(OSG::Color3f(0.5,0.5,0.5));
            material->setSpecular(OSG::Color3f(0.2,0.2,0.2));
            material->setTransparency(0.3);
            OSG::PolygonChunkRefPtr matchunk = OSG::PolygonChunk::create();
            matchunk->setCullFace(GL_BACK);
            material->addChunk(matchunk);

            geo2->setMaterial(material);

            OSG::NodeRefPtr element = OSG::Node::create();
            element->setCore(geo2);

            OSG::NodeRefPtr tempTransNode = OSG::Node::create();
            OSG::ComponentTransformRecPtr tempTrans 
                    = OSG::ComponentTransform::create();
            OSG::Quaternion quat2;
            quat2.setValue(0,0,M_PI/2.0);
            tempTrans->setRotation(quat2);
            tempTransNode->addChild(element);
            tempTransNode->setCore(tempTrans);

            OSG::NodeRecPtr transNode = OSG::Node::create();
            OSG::ComponentTransformRecPtr trans 
                    = OSG::ComponentTransform::create();
            trans->setTranslation(OSG::Vec3f(double(i - m_nbElements/2)
                                                *tunnelElementsWidth, 0, 0));
            trans->setRotation(OSG::Quaternion(1,0,0,0));
            transNode->addChild(tempTransNode);
            transNode->setCore(trans);
            m_elementsRoot->addChild(transNode);

            m_elements.push_back(element);
            m_elementsTrans.push_back(trans);
            m_elementsMaterial.push_back(material);
            m_elementsSpeed.push_back(0);
            m_elementsRot.push_back(0);

        }
        setScale(0);
        m_length = tunnelElementsWidth*double(m_nbElements);
        OSG::commitChanges();

    }
}

Tunnel::~Tunnel() {
    if(m_tunnel->getParent()) {
        m_tunnel->getParent()->subChild(m_tunnel);
    }
}

void Tunnel::parse(xmlNodePtr node) {
    char* value=NULL;
    value = (char*)xmlGetProp(node,(xmlChar*)"position");
    if(value!=NULL) {
        string valueStr=value;
        istringstream is,is2,is3;
        size_t pos;
        std::vector<float> coords(2,0);
        pos=valueStr.find_first_of("/");
        if(pos!=std::string::npos) {
            is.str(valueStr.substr(0,pos));
            is>>coords[0];
        }
        is2.str(valueStr.substr(pos+1));
        is2>>coords[1];
        setPosition(coords);
    }

    value=NULL;
    value = (char*)xmlGetProp(node,(xmlChar*)"rotation");
    if(value!=NULL) {
        std::vector<float> coords(2,0);
        string valueStr=value;
        istringstream is,is2,is3;
        size_t pos;
        pos=valueStr.find_first_of("/");
        if(pos!=std::string::npos) {
            is.str(valueStr.substr(0,pos));
            is>>coords[0];
        }
        is2.str(valueStr.substr(pos+1));
        is2>>coords[1];
        setRotation(coords);
    }
    
    xmlNodePtr paramNode;
    for (paramNode = node->children; paramNode; paramNode = paramNode->next) {
        if (paramNode->type == XML_ELEMENT_NODE && 
            !xmlStrcmp(paramNode->name, (const xmlChar *) "TunnelParameter")) {

            char* value=NULL;
            value = (char*)xmlGetProp(paramNode,(xmlChar*)"parameter");
            if(value!=NULL) {
                std::string valueStr(value);
                addParameter(valueStr);
                TunnelParameter &newParameter = m_tunnelParameters.back();
                m_parameterActivated[newParameter.m_param]=true;

                xmlNodePtr scaleNode;
                for (scaleNode = paramNode->children; scaleNode; 
                        scaleNode = scaleNode->next) {

                    if (scaleNode->type == XML_ELEMENT_NODE && 
                                !xmlStrcmp(scaleNode->name, 
                                (const xmlChar *) "TunnelScale")) {

                        newParameter.m_scales.push_back(TunnelScale());
                        TunnelScale &newScale=
                            newParameter.m_scales[
                                newParameter.m_scales.size()-1];
                        value=NULL;
                        value = (char*)xmlGetProp(scaleNode,(xmlChar*)"slices");
                        if(value!=NULL) {
                            std::string valuesStr = std::string(value);
                            parseSlice(valuesStr, newScale);
                        }
                    }
                }
            }
        }
    }
    setScale(0);
}

void Tunnel::parseSlice(std::string& valuesStr, TunnelScale& scale) {
    size_t pos = valuesStr.find_first_of(";");
    size_t pos2;
    while(pos!=string::npos) {//while there is another slice
        scale.m_slices.push_back(TunnelSlice());
        TunnelSlice &newSlice = scale.m_slices.back();
        istringstream is,is2;
        pos2 = valuesStr.find_first_of("-");
        if(pos2<pos) { //if continuous slice
            newSlice.m_continuous=true;
            is.str(valuesStr.substr(0,pos2));
            is2.str(valuesStr.substr(pos2+1,pos));
            is>>newSlice.m_minValue;
            is2>>newSlice.m_maxValue;
        }
        else { // if discrete slice
            newSlice.m_continuous=false;
            is.str(valuesStr.substr(0,pos));
            is>>newSlice.m_minValue;
            newSlice.m_maxValue = newSlice.m_minValue;
        }
        valuesStr = valuesStr.substr(pos+1);    
        pos = valuesStr.find_first_of(";");
    }

}

void Tunnel::setPosition(const std::vector<float>& pos) {
    m_position[0]=pos[0];
    m_position[1]=pos[1];
    m_tunnelTrans->setTranslation(m_position);
}

void Tunnel::setRotation(const std::vector<float>& rot) {
    m_rotation[0]=rot[0];
    m_rotation[1]=rot[1];
    m_rotation[2]=rot[2];
    OSG::Quaternion quat;
    quat.setValue(m_rotation[0]/180*M_PI, m_rotation[1]/180*M_PI, 
                    m_rotation[2]/180*M_PI);
    m_tunnelTrans->setRotation(quat);
}

void Tunnel::setParameter(const string& param, const string& value) {}

void Tunnel::addParameter(const string& param) {
    TunnelParameter newParameter;
    if(param.compare("color")==0) {
        newParameter.m_param=Worm::WORM_COLOR;
    }
    else if(param.compare("transparency")==0) {
        newParameter.m_param=Worm::WORM_TRANSPARENCY;
    }
    else if(param.compare("distortion")==0) {
        newParameter.m_param=Worm::WORM_DISTORTION;
    }
    else if(param.compare("scattering")==0) {
        newParameter.m_param=Worm::WORM_SCATTERING;
    }
    else if(param.compare("speed")==0) {
        newParameter.m_param=Worm::WORM_SPEED;
    }
    else if(param.compare("rotation")==0) {
        newParameter.m_param=Worm::WORM_ROTATION_OFFSET;
    }
    else if(param.compare("alignment")==0) {
        newParameter.m_param=Worm::WORM_ALIGNEMENT;
    }
    else {
        newParameter.m_param=Worm::WORM_HEIGHT;
    }
    m_tunnelParameters.push_back(newParameter);
}

void Tunnel::setPublic(const bool& pub) {
    {
        //if public make the tunnel vertical
        if(pub) {
            OSG::Quaternion& quat = m_tunnelTrans->editRotation();
            quat.setValue(0, 0, M_PI/2.0);
        }
        else {
            OSG::Quaternion& quat = m_tunnelTrans->editRotation();
            quat.setValue(0, 0, 0);
        }
    }
}

void Tunnel::pack(unsigned int index) {
    //align tunnels to piivert
    OSG::Quaternion quat;
    quat.setValue(0,M_PI/2.0,0);
    m_tunnelTrans->setRotation(quat);

    //set position
    m_tunnelTrans->setTranslation(OSG::Vec3f(0,0,-double(index)*0.2-1));

    //resize
    m_tunnelTrans->setScale(OSG::Vec3f(0.2,0.2,0.2));
}

void Tunnel::unpack() {
    //align tunnels to plane
    OSG::Quaternion quat;
    quat.setValue(0,M_PI/2.0,0);
    m_tunnelTrans->setRotation(quat);

    //resize
    m_tunnelTrans->setScale(OSG::Vec3f(1,1,1));
}

void Tunnel::setScale(const int& scaleID) {
    double value;
    vector<TunnelParameter>::iterator itParam=m_tunnelParameters.begin();
    for(;itParam!=m_tunnelParameters.end();++itParam) {
        if(scaleID<int((*itParam).m_scales.size())) {
            (*itParam).m_currentScale=scaleID;
            for(int el=0; el<m_nbElements; ++el) {
                computeValueFromRatio(value, *itParam, 
                                        double(el)/double(m_nbElements));
                switch((*itParam).m_param) {
                    case Worm::WORM_COLOR: {
                        OSG::Color3f &color = 
                            m_elementsMaterial[el]->editDiffuse();
                        color.setValuesHSV(fmod(value*5.0,1.0)*360.0, 
                                        0.6*(1.0-abs(value*2.0-1.0))+0.1, 
                                        value*0.9+0.1); 
                    }break;
                    case Worm::WORM_TRANSPARENCY: {
                        m_elementsMaterial[el]->setTransparency(0.2+value*0.7);
                    }break;
                    case Worm::WORM_HEIGHT: {
                        OSG::Vec3f &scale = m_elementsTrans[el]->editScale();
                        scale[1]=value+0.02;
                        scale[2]=value+0.02;
                    }break;
                    case Worm::WORM_DISTORTION: {

                    }break;
                    case Worm::WORM_ROTATION_OFFSET: {
                        OSG::Quaternion &quat = 
                            m_elementsTrans[el]->editRotation();
                        quat.setValue(value*M_PI/2.0,0,0);
                    }break;
                    case Worm::WORM_SCATTERING: {
                        OSG::Vec3f &scale = m_elementsTrans[el]->editScale();
                        scale[0]=1.0-(value*0.9);
                    }break;
                    case Worm::WORM_SPEED: {
                        m_elementsRot[el]=0;
                        m_elementsSpeed[el]=value*1.9+0.1;
                    }break;
                    case Worm::WORM_ALIGNEMENT: {
                        OSG::Quaternion &rot = 
                            m_elementsTrans[el]->editRotation();
                        rot[2]=(value*M_PI/16.0)
                                *(float(el%2)*2.0-1.0);
                    }break;
                    default:break;
                }
            }
        }    
    }    

    vector<Ghostable*>::iterator itGhost;
    for(itGhost=m_ghosts.begin();itGhost!=m_ghosts.end();++itGhost) {
        static_cast<Tunnel*>(*itGhost)->setScale(m_currentScale);
    }
}

void Tunnel::setStretch(const double& stretch) {
    OSG::Vec3f& tunnelScale = m_tunnelTrans->editScale();
    if(stretch>0.01) {
        tunnelScale[0]=stretch;
    }
}

void Tunnel::setHeight(const double& height) {
    OSG::Vec3f& tunnelScale = m_tunnelTrans->editScale();
    if(height>1) {
        tunnelScale[1]=height;
    }
}

void Tunnel::computeValueFromRatio(double & value, TunnelParameter& parameter, 
                                    const double& ratio) {

    TunnelScale &scale = parameter.m_scales[parameter.m_currentScale];
    //find which slice we are on    
    double nbSlices = scale.m_slices.size();
    double sliceInd = (ratio>=1.0)? nbSlices-1:
                                      floor(ratio*nbSlices);
    TunnelSlice& slice = scale.m_slices[sliceInd];
    
    //test if continuous or not
    if(slice.m_continuous) {
        //get ratio inside slice 
        double ratioInSlice = (ratio - (sliceInd/nbSlices))*nbSlices;
        value = slice.m_minValue + 
                ratioInSlice*(slice.m_maxValue-slice.m_minValue);

        if(value>=2) { //pitch value
            value/=128.0;
        }
    }
    else {
        value=slice.m_minValue;

        if(value>=2) { //pitch value
            value/=128.0;
        }
    }
}

void Tunnel::testCollisionWithWorm(Worm* worm, bool keepValue) {

    OSG::Vec3f posTunnel, scale, posWorm, beginTunnel;
    OSG::Quaternion quat, scaleQuat;
    
    OSG::BoxVolume tunnelVolume;
    m_elementsRoot->getWorldVolume(tunnelVolume);
    m_elementsRoot->getToWorld().getTransform(posTunnel, quat, 
                                                scale, scaleQuat);
    
    //test collision
    if(worm->testCollisionWithVolume(tunnelVolume)) {
        double stretch = m_tunnelTrans->getScale()[0]; 
        stretch*=m_length;
        
        OSG::Vec3f dir(1, 0, 0);
        quat.multVec(dir, dir);

        beginTunnel = posTunnel - dir*(stretch/2.0);
        worm->getNode()->getToWorld().getTransform(posWorm, quat, 
                                                    scale, scaleQuat);
        //test if worm already in the tunnel
        int prevWorm=-1;
        for(unsigned int w=0; w<m_insideWormsIDs.size(); ++w) {
            if(m_insideWormsIDs[w]==worm->getID()+1) {
                prevWorm=w;
            }
        }
        if(prevWorm==-1) { 
            //if no, tell it to remember the values of controled parameters 
            for(unsigned int p=0;p<m_tunnelParameters.size();++p) {
                worm->storeParameterValue(m_tunnelParameters[p].m_param);
            }
            //and keep the entering position (x,y,z)
            OSG::Vec3f prevPos(1,1,1);
            for(int i=0;i<3;++i) {
                prevPos[i]=posWorm[i]-posTunnel[i];
            }
            //add the worm
            m_insideWormsPositions.push_back(prevPos);
            m_insideWormsIDs.push_back(worm->getID()+1);
            //bypass its parameters
            vector<TunnelParameter>::iterator itParam=
                                                   m_tunnelParameters.begin();
            for(;itParam<m_tunnelParameters.end(); ++itParam) {
                worm->bypassAudioParameter((*itParam).m_param, true);
            }
        }

        //compute ratio according to position
        double ratio=(cos(dir.enclosedAngle(posWorm-beginTunnel))
                        *(posWorm.addToZero().dist(beginTunnel.addToZero()) ))
                    /stretch;
        if(ratio>1) {
            ratio=1;
        }
        else if(cos(dir.enclosedAngle(posWorm-beginTunnel))<0) {
            ratio=0;
        }
        //for each parameter, get value from ratio and set the value
        double value=0;
        vector<TunnelParameter>::iterator itParam=m_tunnelParameters.begin();
        for(; itParam<m_tunnelParameters.end(); ++itParam) {
            computeValueFromRatio(value, *itParam, ratio);
            worm->setParameter((*itParam).m_param, &value, 1);
        }

    }
    else {
        //if not in tunnel but was before 
        int prevWorm = -1;
        for(unsigned int w=0; w<m_insideWormsIDs.size(); ++w) {
            if(m_insideWormsIDs[w]==worm->getID()+1) {
                prevWorm=w;
            }
        }
        if(prevWorm>=0) {
            worm->getNode()->getToWorld().getTransform(posWorm, quat, 
                                                    scale, scaleQuat);
            OSG::Vec3f aftPos(1,1,1);
            for(int i=0;i<3;++i) {
                aftPos[i]=posWorm[i]-posTunnel[i];
            }
/*
            //if did not cross vertically, restore parameters
            if(aftPos[1]*m_insideWormsPositions[prevWorm][1]>0 && !keepValue) {
                for(unsigned int p=0;p<m_tunnelParameters.size();++p) {
                    worm->restoreParameterValue(m_tunnelParameters[p].m_param);
                }
            }
*/
            m_insideWormsIDs.erase(m_insideWormsIDs.begin()+prevWorm);
            m_insideWormsPositions.erase(
                        m_insideWormsPositions.begin()+prevWorm);
        }
    }
}

void Tunnel::update() {
    if(m_parameterActivated[Worm::WORM_SPEED]) {
        //update rotation of the elements
        for(unsigned int i=0; i<m_elementsTrans.size(); ++i) {
            OSG::Quaternion& quat = m_elementsTrans[i]->editRotation(); 
            m_elementsRot[i]=double(fmod(double(m_elementsRot[i]
                                            +m_elementsSpeed[i]/10.0), 
                                        double(M_PI*2.0)));
            quat.setValue(m_elementsRot[i], 0 , 0);
        }
    }

}

void Tunnel::setHighlight(const bool& highlight, 
                            const double& hue, 
                                const double& scale) {
    if((!highlight && m_nbUsers==0) || (highlight && m_nbUsers>0)) {
        {
            m_highlightMaterial->setTransparency(!highlight);
            OSG::Color3f &color = m_highlightMaterial->editDiffuse();
            color.setValuesHSV(hue*360.0, 0.6, 0.8); 
        }
    }
}

Ghostable* Tunnel::createGhost() {
    Tunnel* newGhost = new Tunnel(m_id);

    //fill the scales 
    newGhost->m_tunnelParameters=m_tunnelParameters;
    memcpy(newGhost->m_parameterActivated, m_parameterActivated, 
            Worm::WORM_PARAMETERS_NB);
    newGhost->setScale(m_currentScale);
    newGhost->setStretch(m_tunnelTrans->getScale()[0]);

    m_ghosts.push_back(newGhost);
    return newGhost;
}

void Tunnel::grab(Piivert* tool, 
                    const int& hand, 
                        std::vector<Event>& outputEvents) {
    Manipulable::grab(tool, hand, outputEvents);
}

void Tunnel::release(Piivert* tool, 
                        const int& hand, 
                            std::vector<Event>& outputEvents) {
    Manipulable::release(tool, hand, outputEvents);

    m_tunnelTrans->editTranslation()[2]=-Worm::WORM_DISTANCE/2.0;

/*
        if(m_grabbedTunnel[hand]) {
            //if not within a portal (no ghosts), move back to depth
            if(m_grabbedTunnel[hand]->getNbGhosts()==0) {
                OSG::Vec3f& tunnelPos  = 
                m_grabbedTunnel[hand]->getTrans()
                                            ->editTranslation();
                tunnelPos[2]=-Worm::WORM_DISTANCE/2.0;
            }
            m_grabbedTunnel[hand]->setHighlight(false);
            m_grabbedTunnel[hand]->removeGrabbingTool(m_id);
            m_parentMusician->testTunnelPassagesCollision(
                                            m_grabbedTunnel[hand], 
                                            outputEvents, true);
        }
*/
}

void Tunnel::moveTo(Piivert* tool,
                        const int& hand, 
                            const float& depth, 
                                const OSG::Line& ray, 
                                    std::vector<Event>& outputEvents) {

    OSG::Vec3f pos=m_tunnelTrans->editTranslation();
    float tunnelDepth=depth*2.0;
    float div = tunnelDepth/(-Worm::WORM_DISTANCE/2.0);
    if(tunnelDepth>0) {
         tunnelDepth=0;
    }
    else if(div-floor(div)>0.5) {
        tunnelDepth=(-Worm::WORM_DISTANCE/2.0)*floor(div+1);
    }
    else {
        tunnelDepth=(-Worm::WORM_DISTANCE/2.0)*(floor(div));
    }
    OSG::Plane tunnelPlane(OSG::Vec3f(0,0,-1), OSG::Pnt3f(0, 0, tunnelDepth));
    tunnelPlane.intersect(ray, pos);
    pos[2] += m_zone->getDepth();

    if(!m_grabInitalised) {
        m_grabOffset=m_tunnelTrans->getTranslation()-pos;
        m_storedGrabPos=pos;
        m_grabInitalised=true;
    }

    if(m_isGrabbedWithBothHands) {
        if(hand==0) { //first hand, stretch
            setStretch((m_storedGrabPos-pos).length());
        } 
        else { //second hand, store
            m_storedGrabPos=pos;
        }
    }
    else {
        m_tunnelTrans->setTranslation(pos);
        m_zone->testCollisionsForTunnel(this, outputEvents);
    }

}

void Tunnel::hit(std::vector<Event>& outputEvents) {
    bool reset=true;    
    for(unsigned int p=0;p<m_tunnelParameters.size();++p) {
        TunnelParameter &parameter=m_tunnelParameters[p];
        if(m_currentScale+1<int(parameter.m_scales.size())) {
            reset=false;
        }
    }
    if(reset) {
        m_currentScale=0;
    }
    else {
        m_currentScale++;
    }
    setScale(m_currentScale);
    m_zone->testCollisionsForTunnel(this, outputEvents);
}

void Tunnel::removeFromZone(Zone* zone) {
    zone->removeTunnel(this);
}

void Tunnel::moveToZone(Zone* zone) {
    setZone(zone);
    zone->addTunnel(this);
}

void Tunnel::attachToMusician(Musician* mus) {
    mus->getInternalLightNode()->addChild(m_tunnel);
}

void Tunnel::detachFromMusician(Musician* mus) {
    mus->getInternalLightNode()->subChild(m_tunnel);
}

