/***************************************************************************
 *  Tunnel.hpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef _TUNNEL_H
#define _TUNNEL_H

#include <iostream>
#include <cstring>
#include <cmath>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <vector>
#include <map>

#include <OpenSG/OSGSimpleGeometry.h>
#include <OpenSG/OSGComponentTransform.h>
#include <OpenSG/OSGMaterialGroup.h>
#include <OpenSG/OSGVoidPAttachment.h>
#include <OpenSG/OSGSimpleMaterial.h>
#include <OpenSG/OSGIntersectAction.h>
#include <OpenSG/OSGBoxVolume.h>

#include "Worm.hpp"

enum TUNNEL_SCALE_TYPE {TUNNEL_SCALE_LINEAR, TUNNEL_SCALE_ARRAY};

struct TunnelSlice {
    bool m_continuous;
    double m_minValue;    
    double m_maxValue;
};

struct TunnelScale {
    TUNNEL_SCALE_TYPE m_type;
    std::vector<double> m_values;
    std::vector<TunnelSlice> m_slices;
};

struct TunnelParameter {
    Worm::WORM_PARAMETER m_param;
    int m_currentScale;
    std::vector<TunnelScale> m_scales;
};

class Tunnel: public Ghostable, public Manipulable {
    public:
        Tunnel(const int&);
        ~Tunnel();

        void parse(xmlNodePtr);
        void parseSlice(std::string&, TunnelScale&);

        void setParameter(const std::string&, const std::string&);
        void setPosition(const std::vector<float>&);
        void setRotation(const std::vector<float>&);

        void update();
        
        void setPublic(const bool& pub);
        void pack(unsigned int index);
        void unpack();

        void addParameter(const std::string&);

        inline OSG::NodeRecPtr getNode(){return m_tunnel;}
        inline OSG::NodeRecPtr getCollisionNode(){return m_tunnel;}
        inline OSG::ComponentTransformRecPtr getTrans(){return m_tunnelTrans;}
        inline const OSG::Vec3f& getPosition() { 
            return m_tunnelTrans->getTranslation();
        }

        void setScale(const int&);
        void computeValueFromRatio(double &, TunnelParameter&, const double&);
        void setStretch(const double&);
        void setHeight(const double&);
        inline void setDepth(const double& depth) {
            m_position[2]=depth;
            m_tunnelTrans->setTranslation(m_position);
        }

        void testCollisionWithWorm(Worm*, bool keepValue=false); 
        inline void testCollisionWithWormKeepValue(Worm* w){
            testCollisionWithWorm(w,true);
        }
        //void testCollision(OSG::IntersectAction*, OSG::NodeRefPtr); 

        void setHighlight(const bool&, const double& hue=0, 
                                        const double& scale=1.0);

        void grab(Piivert* tool, 
                    const int& hand, 
                        std::vector<Event>& outputEvents);
        void release(Piivert* tool, 
                        const int& hand, 
                            std::vector<Event>& outputEvents);
        void moveTo(Piivert* tool,
                        const int& hand, 
                            const float& depth, 
                                const OSG::Line& ray, 
                                    std::vector<Event>& outputEvents);
        void hit(std::vector<Event>& outputEvents);
        void removeFromZone(Zone*);
        void moveToZone(Zone*);
        void attachToMusician(Musician* mus);
        void detachFromMusician(Musician* mus);

        Ghostable* createGhost();

    private:

        std::vector<TunnelParameter> m_tunnelParameters;
        int m_currentScale;
        bool m_parameterActivated[Worm::WORM_PARAMETERS_NB];

        OSG::NodeRecPtr m_tunnel;
        OSG::ComponentTransformRecPtr m_tunnelTrans;
        OSG::Vec3f m_position;
        OSG::Vec3f m_rotation;

        OSG::NodeRefPtr m_highlight;
        OSG::SimpleMaterialRefPtr m_highlightMaterial;

        OSG::NodeRecPtr m_elementsRoot;
        std::vector<OSG::NodeRecPtr> m_elements;
        std::vector<OSG::ComponentTransformRecPtr> m_elementsTrans;
        std::vector<OSG::SimpleMaterialRefPtr> m_elementsMaterial;
        std::vector<double> m_elementsSpeed;
        std::vector<double> m_elementsRot;

        std::vector<int> m_insideWormsIDs;
        std::vector<OSG::Vec3f> m_insideWormsPositions;

        double m_length;
        unsigned int m_sides;
        int m_nbElements;

};



#endif 
