/***************************************************************************
 *  Worm.hpp
 *  Part of Drile
 *  2010-  Florent Berthaut
 *  hitmuri.net 
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#ifndef _WORM_H
#define _WORM_H

#include <iostream>
#include <cstring>
#include <cmath>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <vector>
#include <map>
#include <libxml/parser.h>
#include <libxml/tree.h>

#include <OpenSG/OSGSimpleGeometry.h>
#include <OpenSG/OSGComponentTransform.h>
#include <OpenSG/OSGMaterialGroup.h>
#include <OpenSG/OSGSimpleMaterial.h>
#include <OpenSG/OSGNameAttachment.h>
#include <OpenSG/OSGVoidPAttachment.h>
#include <OpenSG/OSGPolygonChunk.h>

#include "../HllEnums.hpp"
#include "Ghostable.hpp"
#include "Manipulable.hpp"

class Worm: public Ghostable, public Manipulable {

    public:
        Worm(const int&);
        virtual ~Worm();
        virtual void parse(xmlNodePtr);
        
        inline const std::string& getName(){return m_name;}

        inline OSG::NodeRecPtr getNode(){return m_worm;}
        inline OSG::ComponentTransformRecPtr getTrans(){return m_wormTrans;}
        inline OSG::NodeRecPtr getCollisionNode(){
                                    return m_elementsGroupTransNode;}
        inline const OSG::Vec3f& getPosition() { 
            return m_wormTrans->getTranslation();
        }
        
        enum WORM_PARAMETER{WORM_COLOR, WORM_HEIGHT,
                                        WORM_SCATTERING, WORM_TRANSPARENCY, 
                                        WORM_DISTORTION, WORM_SPEED,
                                        WORM_ALIGNEMENT, 
                                        WORM_ROTATION_OFFSET, 
                                        WORM_DYNAMIC_SHAPE, 
                                        WORM_STATIC_SHAPE, 
                                        WORM_ROTATION, 
                                        WORM_LOOP_RECORDING, 
                                        WORM_APPEARANCE, 
                                        WORM_PARAMETERS_NB};

        virtual void setParameter(const WORM_PARAMETER&, double*, 
                                    unsigned int, 
                                    bool valueFromAudio=false);

        virtual void setParameter(const std::string&, const std::string&);

        virtual void setColor(const double&);
        virtual void setHeight(const double&);
        virtual void setTransparency(const double&);
        virtual void setScattering(const double&);
        virtual void setAlignement(const double&);
        virtual void setRotationOffset(const double&);
        virtual void setSpeed(const double&);

        void sendParameterValue(const Worm::WORM_PARAMETER&);

        virtual void update(std::vector<Event>&);

        void setWidth(const float& w){m_elementsGroupTrans->editScale()[0]=w;}
        void setHighlight(const bool&, const double& hue=0, 
                            const double& scale=1.0);

        void setVisible(bool);

        void grab(Piivert* tool, 
                    const int& hand, 
                        std::vector<Event>& outputEvents);
        void release(Piivert* tool, 
                        const int& hand, 
                            std::vector<Event>& outputEvents);
        void moveTo(Piivert* tool,
                        const int& hand, 
                            const float& depth, 
                                const OSG::Line& ray, 
                                    std::vector<Event>& outputEvents);
        void hit(std::vector<Event>& outputEvents);
        void removeFromZone(Zone*);
        void moveToZone(Zone*);
        void attachToMusician(Musician* mus);
        void detachFromMusician(Musician* mus);

        inline const unsigned int& getZoneID(){return m_zoneID;}
        inline void setZoneID(const unsigned int& zone){m_zoneID=zone;}

        void storeParameterValue(const WORM_PARAMETER&);
        virtual void restoreParameterValue(const WORM_PARAMETER&);

        inline Worm* getParent(){return m_parent;}
        inline void setParent(Worm* parent);
        virtual void addChild(Worm*);
        virtual void removeChild(Worm*);
        void updateBranchesPositions();
        void updateChildrenPositions();
        inline std::vector<Worm*>& getChildren(){ return m_children;}
        inline bool hasChildren(){return m_children.size()>0;}

        void fold(const int& depth);
        void unfold(const int& depth);
        void setRadialTree(const OSG::Vec3f& treeCenter, const float& radius, 
                            const float& curveCenter, const float& curveAngle);
        void curve();
        void uncurve();

        virtual bool testCollisionWithVolume(OSG::BoxVolume&);
        virtual void testCollisionWithWorm(Worm*, std::vector<Event>&, 
                                            const bool& released);

        inline void bypassAudioParameter(const WORM_PARAMETER& param, 
                                                            bool bypass) { 
            m_parametersBypass[param]=bypass;
        }

        inline const double& getDynamicShape(){return m_dynamicShapeValues[0];}

        static double WORM_SIZE;
        static double WORM_DISTANCE;
        inline const int& getDepth(){return m_depth;};
        void setDepth(const int& depth);

        //mapping with the node parameters
        static std::map<WORM_PARAMETER,NODE_PARAMETER> wormNodeMap;
        static std::map<NODE_PARAMETER,WORM_PARAMETER> nodeWormMap;
        static std::map<NODE_PARAMETER,WORM_PARAMETER> 
                    createNodeWormMappings() {
            std::map<NODE_PARAMETER,WORM_PARAMETER> map2;
            map2[NODE_VOLUME]=WORM_HEIGHT;
            map2[NODE_MODULATION]=WORM_APPEARANCE;
            map2[NODE_PITCH]=WORM_COLOR;
            map2[NODE_NOISINESS]=WORM_TRANSPARENCY;
            map2[NODE_RESONANCE]=WORM_SCATTERING;
            map2[NODE_BRIGHTNESS]=WORM_DISTORTION;
            map2[NODE_IRREGULARITY]=WORM_ALIGNEMENT;
            map2[NODE_OFFSET]=WORM_ROTATION_OFFSET;
            map2[NODE_SPEED]=WORM_SPEED;
            map2[NODE_ANALYSIS_DYNAMIC]=WORM_DYNAMIC_SHAPE;
            map2[NODE_ANALYSIS_STATIC]=WORM_STATIC_SHAPE;
            map2[NODE_LOOP_POSITION]=WORM_ROTATION;
            map2[NODE_LOOP_RECORDING]=WORM_LOOP_RECORDING;
            return map2;
        }
        static std::map<WORM_PARAMETER,NODE_PARAMETER> 
                    createWormNodeMappings() {
            std::map<WORM_PARAMETER,NODE_PARAMETER> map1; 
            map1[WORM_HEIGHT]=NODE_VOLUME;
            map1[WORM_COLOR]=NODE_PITCH;
            map1[WORM_TRANSPARENCY]=NODE_NOISINESS;
            map1[WORM_SCATTERING]=NODE_RESONANCE;
            map1[WORM_DISTORTION]=NODE_BRIGHTNESS;
            map1[WORM_ALIGNEMENT]=NODE_IRREGULARITY;
            map1[WORM_ROTATION_OFFSET]=NODE_OFFSET;
            map1[WORM_SPEED]=NODE_SPEED;
            return map1;
        }

    protected:
        std::string m_name;

        unsigned int m_zoneID;
        int m_depth;

        OSG::NodeRecPtr m_worm;
        OSG::ComponentTransformRecPtr m_wormTrans;

        OSG::ComponentTransformRecPtr m_elementsGroupTrans;
        OSG::NodeRecPtr m_elementsGroupTransNode;
        OSG::SimpleMaterialRefPtr m_wormMaterial;
        OSG::Vec3f m_position;

        OSG::NodeRefPtr m_highlight;
        OSG::SimpleMaterialRefPtr m_highlightMaterial;
        OSG::ComponentTransformRefPtr m_highlightTrans;
        OSG::NodeRefPtr m_highlightTransNode;

        float m_nbElements;
        float m_elementHeight;
        std::vector<OSG::ComponentTransformRecPtr> m_elementsTrans;
        std::vector<OSG::NodeRecPtr> m_elementsNode;
        std::vector<OSG::Vec3f> m_elementsRotation;
        std::vector<float> m_elementsWidth;

        bool m_appearanceChanged;
        double m_appearanceValues[WORM_PARAMETERS_NB];

        bool m_staticShapeChanged;
        std::vector<double> m_staticShapeValues;
        
        bool m_dynamicShapeChanged;
        std::vector<double> m_dynamicShapeValues;

        bool m_looped;


        double m_parametersValues[WORM_PARAMETERS_NB];
        double m_parametersStoredValues[WORM_PARAMETERS_NB];
        bool m_parametersBypass[WORM_PARAMETERS_NB];

        timeval m_storedCollisionTime;
        int m_storedCollisionWorm;

        Worm* m_parent;
        std::vector<Worm*> m_children;
        std::vector<OSG::NodeRecPtr> m_branches;
        std::vector<OSG::ComponentTransformRecPtr> m_branchesScales;
        bool m_folded;
        OSG::Vec3f m_treeCenter;
        float m_treeRadius;
        float m_curveAngle;
        float m_curveCenter;
        bool m_behindTunnelPlane;

        OSG::SimpleMaterialRefPtr m_inputBranchMat;
        OSG::SimpleMaterialRefPtr m_downOutputBranchMat;
        OSG::SimpleMaterialRefPtr m_frontOutputBranchMat;
        OSG::SimpleMaterialRefPtr m_backOutputBranchMat;

        OSG::ComponentTransformRecPtr m_recTrans;    
        OSG::SimpleMaterialRefPtr m_recMaterial;

        std::vector<Event> m_outputEvents;

        double m_highlightScaleStep;

        bool m_duplicated;
};

#endif

