/***************************************************************************
 *  Zone.cpp
 *  Part of Drile
 *  2010- Florent Berthaut
 *  hitmuri.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "Zone.hpp"
#include "Worm.hpp"
#include "Tunnel.hpp"
#include "Grid.hpp"
#include "Sign.hpp"

using namespace std;

Zone::Zone(const unsigned int& id):m_id(id), m_depth(2){


    {
        m_zone=OSG::Node::create();
        m_zoneTrans=OSG::ComponentTransform::create();
        m_zone->setCore(m_zoneTrans);
        m_zoneTempTrans=OSG::ComponentTransform::create();
        m_zoneRoot=OSG::Node::create();
        m_zoneRoot->setCore(m_zoneTempTrans);
        m_zoneTempTrans->setTranslation(OSG::Vec3f(0,0,-m_depth));
        m_zone->addChild(m_zoneRoot);

        m_wormsRoot=OSG::Node::create();
        m_wormsRootTrans = OSG::ComponentTransform::create();
        m_wormsRoot->setCore(m_wormsRootTrans);
        m_zoneRoot->addChild(m_wormsRoot);
        m_toolsRoot=OSG::Node::create();
        m_toolsRootTrans = OSG::ComponentTransform::create();
        m_toolsRoot->setCore(m_toolsRootTrans);
        m_zoneRoot->addChild(m_toolsRoot);

        OSG::commitChanges();
    }
}

Zone::~Zone(){}

void Zone::parse(xmlNodePtr zoneNode) {

    char* value=NULL;
    value = (char*)xmlGetProp(zoneNode,(xmlChar*)"depth");
    if(value!=NULL) {
        setDepth(atof(value));;
    } 
}

void Zone::update(std::vector<Event>& events) {
    //update/test tunnels
    vector<Tunnel*>::iterator itTunnel;
    for(itTunnel=m_tunnels.begin(); itTunnel!=m_tunnels.end(); ++itTunnel) {
        (*itTunnel)->update();
    }
    //update worms and get total musical activity of the zone
    m_activity=0;
    m_activityCounter=1;
    vector<Worm*>::iterator itWorm;
    for(itWorm=m_worms.begin(); itWorm!=m_worms.end();++itWorm) {
        (*itWorm)->update(events);    
        if((*itWorm)->getParent()==NULL) {
            m_activity+=(*itWorm)->getDynamicShape();
            m_activityCounter++;
        }
    }
    m_activity/=m_activityCounter;
}

void Zone::testCollisionsForWorm(Worm* worm, 
                                std::vector<Event>& events, 
                                const bool& released) {

    //cout<<"Test collision for worm: "<<worm->getID()<<//endl;;

    OSG::BoxVolume worm1Volume, worm2BranchVolume, worm2Volume;
    worm->getCollisionNode()->getWorldVolume(worm1Volume);


    //test collision with other worms of the same zone
    //FIXME test release 
    vector<Worm*>::iterator itWorm;
    for(itWorm=m_worms.begin(); itWorm!=m_worms.end();++itWorm) {
        Worm* collidingWorm =(*itWorm);
        //cout<<"with worm "<<collidingWorm->getID()<<" : ";
        //get another worm to collide with
        if(worm->getID()!=collidingWorm->getID() && 
                collidingWorm->getNbUsers()==0) {
            worm->testCollisionWithWorm(collidingWorm,events,released);
        }
    }

    //test collision with tunnels
    vector<Tunnel*>::iterator itTunnel;
    for(itTunnel=m_tunnels.begin(); 
            itTunnel!=m_tunnels.end(); ++itTunnel) {
        (*itTunnel)->testCollisionWithWorm(worm);
    }
}


void Zone::testCollisionsForTunnel(Tunnel* tunnel, 
                                    std::vector<Event>& events, 
                                    const bool& released) {
    vector<Worm*>::iterator itWorm;
    for(itWorm=m_worms.begin(); itWorm!=m_worms.end();++itWorm) {
        tunnel->testCollisionWithWormKeepValue(*itWorm);
    }
}

void Zone::addWorm(Worm* worm) {
    m_worms.push_back(worm);
    worm->setZoneID(m_id);
    worm->setZone(this);
    m_wormsRoot->addChild(worm->getNode());
}

void Zone::removeWorm(Worm* worm) {
    vector<Worm*>::iterator itWorm;
    for(itWorm=m_worms.begin(); itWorm!=m_worms.end();++itWorm) {
        if((*itWorm)->getID()==worm->getID()) {
            m_worms.erase(itWorm);
            break;
        }
    }
    m_wormsRoot->subChild(worm->getNode());
}

void Zone::addTunnel(Tunnel* tunnel) {
    m_tunnels.push_back(tunnel);
    tunnel->setZone(this);
    tunnel->setDepth(-Worm::WORM_DISTANCE/2.0);
    m_toolsRoot->addChild(tunnel->getNode());
}

void Zone::removeTunnel(Tunnel* tunnel) {
    vector<Tunnel*>::iterator itTunnel;
    for(itTunnel=m_tunnels.begin(); itTunnel!=m_tunnels.end(); ++itTunnel) {
        if((*itTunnel)->getID()==tunnel->getID()) {
            m_tunnels.erase(itTunnel);
            break;
        }
    }
    m_toolsRoot->subChild(tunnel->getNode());
}

void Zone::addGrid(Grid* newGrid) {
    m_grids.push_back(newGrid);
    m_toolsRoot->addChild(newGrid->getNode());
}

void Zone::removeGrid(Grid* newGrid) {
    vector<Grid*>::iterator itGrid;
    for(itGrid=m_grids.begin(); itGrid!=m_grids.end(); ++itGrid) {
        if((*itGrid)->getID()==newGrid->getID()) {
            m_grids.erase(itGrid);
            break;
        }
    }
}

void Zone::markWormsForSelection() {
    m_zone->setTravMask(OSG::TypeTraits<OSG::UInt32>::getMax());
    m_wormsRoot->setTravMask(OSG::TypeTraits<OSG::UInt32>::getMax());
    m_toolsRoot->setTravMask(OSG::TypeTraits<OSG::UInt32>::getZeroElement());
}

void Zone::markTunnelsForSelection() {
    m_zone->setTravMask(OSG::TypeTraits<OSG::UInt32>::getMax());
    m_toolsRoot->setTravMask(OSG::TypeTraits<OSG::UInt32>::getMax());
    m_wormsRoot->setTravMask(OSG::TypeTraits<OSG::UInt32>::getZeroElement());
}

void Zone::markForSelection(){
    m_zone->setTravMask(OSG::TypeTraits<OSG::UInt32>::getMax());
    m_toolsRoot->setTravMask(OSG::TypeTraits<OSG::UInt32>::getMax());
    m_wormsRoot->setTravMask(OSG::TypeTraits<OSG::UInt32>::getMax());
}

void Zone::unmarkForSelection(){
    m_zone->setTravMask(OSG::TypeTraits<OSG::UInt32>::getZeroElement());
}

