/* -*- coding: utf-8 -*-
 *
 * OneEuroFilter.cc -
 *
 * Author: Nicolas Roussel (nicolas.roussel@inria.fr)
 *
 */

#ifndef OneEuroFilter_H
#define OneEuroFilter_H

#include <iostream>
#include <stdexcept>
#include <cmath>
#include <ctime>

class LowPassFilter {
    
  double y, a, s ;
  bool initialized ;

  void setAlpha(double alpha) {
    if (alpha<=0.0 || alpha>1.0) 
      throw std::range_error("alpha should be in (0.0., 1.0]") ;
    a = alpha ;
  }

public:

  LowPassFilter(double alpha, double initval=0.0) {
    y = s = initval ;
    setAlpha(alpha) ;
    initialized = false ;
  }

  double filter(double value) {
    double result ;
    if (initialized)
      result = a*value + (1.0-a)*s ;
    else {
      result = value ;
      initialized = true ;
    }
    y = value ;
    s = result ;
    return result ;
  }

  double filterWithAlpha(double value, double alpha) {
    setAlpha(alpha) ;
    return filter(value) ;
  }

  bool hasLastRawValue(void) {
    return initialized ;
  }

  double lastRawValue(void) {
    return y ;
  }

} ;

// -----------------------------------------------------------------

class OneEuroFilter {

  double freq ;
  double mincutoff ;
  double beta_ ;
  double dcutoff ;
  LowPassFilter *x ;
  LowPassFilter *dx ;
  double lasttime ;

  inline double alpha(double cutoff) {
    double te = 1.0 / freq ;
    double tau = 1.0 / (2*M_PI*cutoff) ;
    return 1.0 / (1.0 + tau/te) ;
  }

  inline void setFrequency(double f) {
    if (f<=0) throw std::range_error("freq should be >0") ;
    freq = f ;
  }

  inline void setMinCutoff(double mc) {
    if (mc<=0) throw std::range_error("mincutoff should be >0") ;
    mincutoff = mc ;
  }

  inline void setBeta(double b) {
    beta_ = b ;
  }

  inline void setDerivateCutoff(double dc) {
    if (dc<=0) throw std::range_error("dcutoff should be >0") ;
    dcutoff = dc ;
  }

public:

  OneEuroFilter(double freq=30, double mincutoff=1.0, 
                double beta_=0.0, double dcutoff=1.0);
  double filter(double value, double timestamp=-1.0);
  ~OneEuroFilter(void);

} ;

#endif

